#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include <iomanip>
#include <gdkmm.h>
#include <gtkmm.h>

#include "cfmuvalidateserver.hh"

CFMUValidateServer::CFMUValidateServer(int& argc, char**& argv)
	: m_portal_webview(0), m_validate_webview(0),
	  m_portal_webview_close_sighandler(0), m_portal_webview_create_sighandler(0),
	  m_portal_webview_ready_sighandler(0), m_portal_webview_loadstatus_sighandler(0),
	  m_portal_webview_navpolicy_sighandler(0), m_portal_webview_newwndpolicy_sighandler(0),
	  m_portal_webview_rsrcload_sighandler(0), m_portal_webview_consolemsg_sighandler(0),
	  m_validate_webview_close_sighandler(0), m_validate_webview_create_sighandler(0),
	  m_validate_webview_ready_sighandler(0), m_validate_webview_loadstatus_sighandler(0),
	  m_validate_webview_navpolicy_sighandler(0), m_validate_webview_newwndpolicy_sighandler(0),
	  m_validate_webview_rsrcload_sighandler(0), m_validate_webview_consolemsg_sighandler(0),
	  m_verbose(0), m_wvstate(wvstate_failed), m_quit(false)
{
        bool version(false);
	Glib::ustring fpl;
	int verbose(0);
	// option group
	{
		Glib::OptionGroup options("cfmuvalidateserver", "CFMU Flight Plan Validation Server");
		// version      
		Glib::OptionEntry entry_version;
		entry_version.set_long_name(_("version"));
		entry_version.set_description(_("Show version and exit."));
		options.add_entry(entry_version, version);
		// fpl
                Glib::OptionEntry entry_fpl;
                entry_fpl.set_long_name(_("fpl"));
                entry_fpl.set_description(_("Query flight plan given as argument and return."));
                options.add_entry(entry_fpl, fpl);
 		// verbose      
		Glib::OptionEntry entry_verbose;
		entry_verbose.set_long_name(_("verbose"));
		entry_verbose.set_description(_("Set Verbosity."));
		options.add_entry(entry_verbose, verbose);

		// option context
		Glib::OptionContext context(_("-- CFMU Flight Plan Validation Server"));
		//context.add_group(options);
		context.set_main_group(options);
		context.set_help_enabled(true);
		context.set_ignore_unknown_options(true);
		Glib::set_prgname(Glib::path_get_basename(*argv));
		if (!context.parse(argc, argv)) {
			std::cerr << "Command Line Option Error: Usage: " << Glib::get_prgname() << std::endl
				  << context.get_help(false);
			exit(1);
		}
	}
        if (version) {
                std::cout << Glib::get_prgname() << (" Version " VERSION) << std::endl;
		exit(0);
	}
	m_verbose = std::max(verbose, 0);
        m_mainloop = Glib::MainLoop::create(Glib::MainContext::get_default());	
        m_inputchan = Glib::IOChannel::create_from_fd(0);
        m_outputchan = Glib::IOChannel::create_from_fd(1);
        m_inputchan->set_encoding();
        m_outputchan->set_encoding();
 	if (!fpl.empty()) {
		m_flightplans.push_back(fpl);
		m_quit = true;
	} else {
		m_inputreaderconn = Glib::signal_io().connect(sigc::mem_fun(this, &CFMUValidateServer::input_line_handler),
							      m_inputchan, Glib::IO_IN | Glib::IO_ERR | Glib::IO_HUP);
	}
	prepare_webview();
}

CFMUValidateServer::~CFMUValidateServer()
{
	remove_portal_sighandlers();
	remove_validate_sighandlers();
	if (m_portal_webview)
		g_object_unref(m_portal_webview);
	if (m_validate_webview)
		g_object_unref(m_validate_webview);
}

void CFMUValidateServer::run(void)
{
        m_mainloop->run();
}

bool CFMUValidateServer::input_line_handler(Glib::IOCondition iocond)
{
        Glib::ustring line;
        Glib::IOStatus iostat(m_inputchan->read_line(line));
        if (iostat == Glib::IO_STATUS_ERROR ||
            iostat == Glib::IO_STATUS_EOF) {
		m_quit = true;
		switch (m_wvstate) {
		case wvstate_loadmainpage:
		case wvstate_loadquerypage:
		case wvstate_ready:
			if (m_flightplans.empty())
				m_mainloop->quit();
			break;

		default:
			break;
		}
		m_inputreaderconn.disconnect();
		return true;
        }
        while (!line.empty()) {
                char ch(line[line.size() - 1]);
                if (ch != '\r' && ch != '\n')
                        break;
                line.erase(line.size() - 1, 1);
        }
	if (line.empty())
		return true;
	if (m_verbose > 1U)
		std::cerr << "Queue Flight Plan: " << line << std::endl;
	m_flightplans.push_back(line);
	if (m_wvstate == wvstate_ready)
		set_wvstate(wvstate_ready);
	return true;
}

void CFMUValidateServer::add_portal_sighandlers(void)
{
	remove_portal_sighandlers();
	m_portal_webview_close_sighandler = g_signal_connect(m_portal_webview, "close-web-view", G_CALLBACK(&CFMUValidateServer::close_webview_1), this);
	m_portal_webview_create_sighandler = g_signal_connect(m_portal_webview, "create-web-view", G_CALLBACK(&CFMUValidateServer::create_webview_1), this);
	m_portal_webview_ready_sighandler = g_signal_connect(m_portal_webview, "web-view-ready", G_CALLBACK(&CFMUValidateServer::webview_ready_1), this);
	m_portal_webview_loadstatus_sighandler = g_signal_connect(m_portal_webview, "notify::load-status", G_CALLBACK(&CFMUValidateServer::loadstatus_webview_1), this);
	m_portal_webview_navpolicy_sighandler = g_signal_connect(m_portal_webview, "navigation-policy-decision-requested", G_CALLBACK(&CFMUValidateServer::policy_decision_required_1), this);
	m_portal_webview_newwndpolicy_sighandler = g_signal_connect(m_portal_webview, "new-window-policy-decision-requested", G_CALLBACK(&CFMUValidateServer::policy_decision_required_1), this);
	m_portal_webview_rsrcload_sighandler = g_signal_connect(m_portal_webview, "resource-load-finished", G_CALLBACK(&CFMUValidateServer::resourceload_finished_1), this);
	m_portal_webview_consolemsg_sighandler = g_signal_connect(m_portal_webview, "console-message", G_CALLBACK(&CFMUValidateServer::console_message_1), this);
}

void CFMUValidateServer::remove_portal_sighandlers(void)
{
        if (m_portal_webview_close_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_close_sighandler);
		m_portal_webview_close_sighandler = 0;
        }
        if (m_portal_webview_create_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_create_sighandler);
		m_portal_webview_create_sighandler = 0;
        }
	if (m_portal_webview_ready_sighandler) {
		g_signal_handler_disconnect(m_portal_webview, m_portal_webview_ready_sighandler);
		m_portal_webview_ready_sighandler = 0;
	}
        if (m_portal_webview_loadstatus_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_loadstatus_sighandler);
		m_portal_webview_loadstatus_sighandler = 0;
        }
        if (m_portal_webview_navpolicy_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_navpolicy_sighandler);
		m_portal_webview_navpolicy_sighandler = 0;
        }
        if (m_portal_webview_newwndpolicy_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_newwndpolicy_sighandler);
		m_portal_webview_newwndpolicy_sighandler = 0;
        }
        if (m_portal_webview_rsrcload_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_rsrcload_sighandler);
		m_portal_webview_rsrcload_sighandler = 0;
        }
        if (m_portal_webview_consolemsg_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_consolemsg_sighandler);
		m_portal_webview_consolemsg_sighandler = 0;
        }
}

void CFMUValidateServer::add_validate_sighandlers(void)
{
	remove_validate_sighandlers();
	m_validate_webview_close_sighandler = g_signal_connect(m_validate_webview, "close-web-view", G_CALLBACK(&CFMUValidateServer::close_webview_1), this);
	m_validate_webview_create_sighandler = g_signal_connect(m_validate_webview, "create-web-view", G_CALLBACK(&CFMUValidateServer::create_webview_1), this);
	m_validate_webview_ready_sighandler = g_signal_connect(m_validate_webview, "web-view-ready", G_CALLBACK(&CFMUValidateServer::webview_ready_1), this);
	m_validate_webview_loadstatus_sighandler = g_signal_connect(m_validate_webview, "notify::load-status", G_CALLBACK(&CFMUValidateServer::loadstatus_webview_1), this);
	m_validate_webview_navpolicy_sighandler = g_signal_connect(m_validate_webview, "navigation-policy-decision-requested", G_CALLBACK(&CFMUValidateServer::policy_decision_required_1), this);
	m_validate_webview_newwndpolicy_sighandler = g_signal_connect(m_validate_webview, "new-window-policy-decision-requested", G_CALLBACK(&CFMUValidateServer::policy_decision_required_1), this);
	m_validate_webview_rsrcload_sighandler = g_signal_connect(m_validate_webview, "resource-load-finished", G_CALLBACK(&CFMUValidateServer::resourceload_finished_1), this);
	m_validate_webview_consolemsg_sighandler = g_signal_connect(m_validate_webview, "console-message", G_CALLBACK(&CFMUValidateServer::console_message_1), this);
}

void CFMUValidateServer::remove_validate_sighandlers(void)
{
        if (m_validate_webview_close_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_close_sighandler);
		m_validate_webview_close_sighandler = 0;
        }
        if (m_validate_webview_create_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_create_sighandler);
		m_validate_webview_create_sighandler = 0;
        }
	if (m_validate_webview_ready_sighandler) {
		g_signal_handler_disconnect(m_validate_webview, m_validate_webview_ready_sighandler);
		m_validate_webview_ready_sighandler = 0;
	}
        if (m_validate_webview_loadstatus_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_loadstatus_sighandler);
		m_validate_webview_loadstatus_sighandler = 0;
        }
        if (m_validate_webview_navpolicy_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_navpolicy_sighandler);
		m_validate_webview_navpolicy_sighandler = 0;
        }
        if (m_validate_webview_newwndpolicy_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_newwndpolicy_sighandler);
		m_validate_webview_newwndpolicy_sighandler = 0;
        }
        if (m_validate_webview_rsrcload_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_rsrcload_sighandler);
		m_validate_webview_rsrcload_sighandler = 0;
        }
        if (m_validate_webview_consolemsg_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_consolemsg_sighandler);
		m_validate_webview_consolemsg_sighandler = 0;
        }
}

void CFMUValidateServer::prepare_webview(void)
{
	remove_portal_sighandlers();
	if (m_portal_webview)
		g_object_unref(m_portal_webview);
	m_portal_webview = GTK_WIDGET(g_object_ref_sink(webkit_web_view_new()));
	gtk_widget_show(m_portal_webview);
	add_portal_sighandlers();
	WebKitWebView *webview = WEBKIT_WEB_VIEW(m_portal_webview);
	webkit_web_view_load_uri(webview, "https://www.public.cfmu.eurocontrol.int/PUBPORTAL/gateway/spec/index.html");
	set_wvstate(wvstate_loadmainpage);
}

void CFMUValidateServer::set_wvstate(wvstate_t newstate)
{
	m_delayedstatechangeconn.disconnect();
	bool chg(m_wvstate != newstate);
	m_wvstate = newstate;
	switch (m_wvstate) {
	case wvstate_ready:
	{
		if (m_flightplans.empty()) {
			if (m_quit)
				m_mainloop->quit();
			break;
		}
		if (!m_validate_webview)
			break;
		WebKitWebView *webview(WEBKIT_WEB_VIEW(m_validate_webview));
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		if (!document)
			break;
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.GENERAL_DATA_ENTRY.INTRODUCE_FLIGHT_PLAN_FIELD"));
		if (!element || !WEBKIT_DOM_IS_HTML_TEXT_AREA_ELEMENT(element))
			break;
		webkit_dom_html_text_area_element_set_value(WEBKIT_DOM_HTML_TEXT_AREA_ELEMENT(element), m_flightplans.front().c_str());
		m_flightplans.pop_front();
		element = webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.VALIDATE_ACTION_LABEL");
		if (!element)
			break;
		zap_errors();
		webkit_web_view_execute_script(webview, "document.getElementById(\"FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.VALIDATE_ACTION_LABEL\").click();");
		set_wvstate(wvstate_query);
		break;
	}

	case wvstate_failed:
		if (chg) {
			m_outputchan->write("FAIL: WEBPAGE\n");
			m_outputchan->flush();
			break;
		}
		m_mainloop->quit();
		break;

	default:
		break;
	}
}

void CFMUValidateServer::parse_results(void)
{
	if (!m_validate_webview)
		return;
	typedef std::pair<guint,bool> lvlvisible_t;
	typedef std::pair<WebKitDOMNode *,lvlvisible_t> nodelvlvisible_t;
	std::stack<nodelvlvisible_t> subtree;
	{
		WebKitWebView *webview(WEBKIT_WEB_VIEW(m_validate_webview));
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA"));
		if (!element || !WEBKIT_DOM_IS_NODE(element))
			return;
		subtree.push(std::make_pair(WEBKIT_DOM_NODE(element), std::make_pair(0U, true)));
	}
	std::ostringstream results;
	while (!subtree.empty()) {
		WebKitDOMNode *node;
		guint level;
		bool visible;
		{
			nodelvlvisible_t nl(subtree.top());
			node = nl.first;
			level = nl.second.first;
			visible = nl.second.second;
		}
		subtree.pop();
		if (WEBKIT_DOM_IS_HTML_ELEMENT(node)) {
			WebKitDOMCSSStyleDeclaration *style(webkit_dom_element_get_style(WEBKIT_DOM_ELEMENT(node)));
			if (style) {
				gchar *disp(webkit_dom_css_style_declaration_get_property_value(style, "display"));
				if (disp) {
					visible = !!strcmp(disp, "none");
					g_free(disp);
				}
			}
		}
		if (visible) {
			WebKitDOMNodeList *list(webkit_dom_node_get_child_nodes(node));
			gulong i(webkit_dom_node_list_get_length(list));
			while (i) {
				--i;
				WebKitDOMNode *ptr(webkit_dom_node_list_item(list, i));
				if (!ptr)
					continue;
				subtree.push(std::make_pair(ptr, std::make_pair(level + 1U, visible)));
			}
		}
		if (!WEBKIT_DOM_IS_HTML_ELEMENT(node))
			continue;
		char *class_name(webkit_dom_html_element_get_class_name(WEBKIT_DOM_HTML_ELEMENT(node)));
		if (m_verbose > 2U) {
			std::cerr << std::string(2U * level, ' ') << "element:";
			gchar *lname(webkit_dom_node_get_local_name(WEBKIT_DOM_NODE(node)));
			if (lname) {
				if (*lname)
					std::cerr << " name \"" << lname << "\"";
				g_free(lname);
			}
			lname = webkit_dom_html_element_get_id(WEBKIT_DOM_HTML_ELEMENT(node));
			if (lname) {
				if (*lname)
					std::cerr << " id \"" << lname << "\"";
				g_free(lname);
			}
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(node)));
			if (typ && *typ)
				std::cerr << " type \"" << typ << "\"";
			if (class_name && *class_name)
				std::cerr << " class \"" << class_name << "\"";
			WebKitDOMCSSStyleDeclaration *style(webkit_dom_element_get_style(WEBKIT_DOM_ELEMENT(node)));
			if (style) {
				gchar *disp(webkit_dom_css_style_declaration_get_property_value(style, "display"));
				if (disp) {
					if (*disp)
						std::cerr << " display \"" << disp << "\"" << std::endl;
					g_free(disp);
				}
			}
			gchar *txt(webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (txt) {
				if (*txt)
					std::cerr << " text \"" << txt << "\"";
				g_free(txt);
			}
			std::cerr << std::endl;
		}
		if (!class_name)
			continue;
		if (!visible) {
			g_free(class_name);
			continue;
		}
		if (!strcmp(class_name, "portal_dataValue") && WEBKIT_DOM_IS_HTML_DIV_ELEMENT(node)) {
			gchar *txt(webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (txt) {
				if (*txt) {
					results << txt << std::endl;
					if (m_verbose > 2U)
						std::cerr << "*** " << txt << std::endl;
				}
				g_free(txt);
			}
		}
		if (false && !strcmp(class_name, "portal_decotableCell ") && WEBKIT_DOM_IS_HTML_TABLE_CELL_ELEMENT(node)) {
			gchar *txt(webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (txt) {
				if (*txt) {
					results << txt << std::endl;
					if (m_verbose > 2U)
						std::cerr << "*** " << txt << std::endl;
				}
				g_free(txt);
			}
		}
		if (true && (!strcmp(class_name, "portal_rowOdd") || !strcmp(class_name, "portal_rowEven")) &&
		    WEBKIT_DOM_IS_HTML_TABLE_ROW_ELEMENT(node)) {
			gchar *txt(webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (txt) {
				if (*txt) {
					results << txt << std::endl;
					if (m_verbose > 2U)
						std::cerr << "*** " << txt << std::endl;
				}
				g_free(txt);
			}
		}
		g_free(class_name);
	}
	if (!results.str().empty()) {
                m_outputchan->write(results.str() + "\n");
                m_outputchan->flush();
		set_wvstate(wvstate_ready);
	}
}

void CFMUValidateServer::zap_errors(void)
{
	if (!m_validate_webview)
		return;
	std::stack<WebKitDOMNode *> subtree;
	{
		WebKitWebView *webview(WEBKIT_WEB_VIEW(m_validate_webview));
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA"));
		if (!element || !WEBKIT_DOM_IS_NODE(element))
			return;
		subtree.push(WEBKIT_DOM_NODE(element));
	}
	std::ostringstream results;
	while (!subtree.empty()) {
		WebKitDOMNode *node(subtree.top());
		subtree.pop();
		if (WEBKIT_DOM_IS_HTML_ELEMENT(node)) {
			char *class_name(webkit_dom_html_element_get_class_name(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (class_name) {
		     		if ((!strcmp(class_name, "portal_rowOdd") || !strcmp(class_name, "portal_rowEven")) &&
				    WEBKIT_DOM_IS_HTML_TABLE_ROW_ELEMENT(node)) {
					g_free(class_name);
					while (webkit_dom_node_has_child_nodes(node)) {
						GError *error(0);
						webkit_dom_node_remove_child(node, webkit_dom_node_get_first_child(node), &error);
						if (error) {
							std::cerr << "Cannot remove child: " << error->message << std::endl;
							g_error_free(error);
						}
					}
					continue;
				}
				g_free(class_name);
			}
		}
		{
			WebKitDOMNodeList *list(webkit_dom_node_get_child_nodes(node));
			gulong i(webkit_dom_node_list_get_length(list));
			while (i) {
				--i;
				WebKitDOMNode *ptr(webkit_dom_node_list_item(list, i));
				if (!ptr)
					continue;
				subtree.push(ptr);
			}
		}
	}
}

gboolean CFMUValidateServer::close_webview_1(WebKitWebView* webview, gpointer userdata)
{
	return static_cast<CFMUValidateServer *>(userdata)->close_webview(webview);
}

gboolean CFMUValidateServer::close_webview(WebKitWebView* webview)
{
        set_wvstate(wvstate_failed);
	return true;
}

WebKitWebView *CFMUValidateServer::create_webview_1(WebKitWebView *webview, WebKitWebFrame *frame, gpointer userdata)
{
	return static_cast<CFMUValidateServer *>(userdata)->create_webview(webview, frame);
}

WebKitWebView *CFMUValidateServer::create_webview(WebKitWebView *webview, WebKitWebFrame *frame)
{
	if (m_verbose > 1U)
		std::cerr << "new webview!" << std::endl;
	if (webview == WEBKIT_WEB_VIEW(m_validate_webview))
		return 0;
        g_assert(webview == WEBKIT_WEB_VIEW(m_portal_webview));
	if (m_wvstate != wvstate_loadmainpage)
		return 0;
	remove_validate_sighandlers();
	if (m_validate_webview)
		g_object_unref(m_validate_webview);
	m_validate_webview = GTK_WIDGET(g_object_ref_sink(webkit_web_view_new()));
	gtk_widget_show(m_validate_webview);
	webkit_web_view_set_settings(WEBKIT_WEB_VIEW(m_validate_webview), webkit_web_view_get_settings(WEBKIT_WEB_VIEW(m_portal_webview)));
	add_validate_sighandlers();
	set_wvstate(wvstate_loadquerypage);
	WebKitWebView *wv(WEBKIT_WEB_VIEW(m_validate_webview));
	return wv;
}

gboolean CFMUValidateServer::webview_ready_1(WebKitWebView* webview, gpointer userdata)
{
	return static_cast<CFMUValidateServer *>(userdata)->webview_ready(webview);
}

gboolean CFMUValidateServer::webview_ready(WebKitWebView* webview)
{
	if (m_verbose > 1U)
		std::cerr << "WebView: ready" << std::endl;
        g_assert(webview == WEBKIT_WEB_VIEW(m_portal_webview) || webview == WEBKIT_WEB_VIEW(m_validate_webview));
	gtk_widget_show_all(GTK_WIDGET(webview));
	return false;
}

void CFMUValidateServer::loadstatus_webview_1(WebKitWebView *webview, GParamSpec *pspec, gpointer userdata)
{
	static_cast<CFMUValidateServer *>(userdata)->loadstatus_webview(webview, pspec);
}

void CFMUValidateServer::loadstatus_webview(WebKitWebView *webview, GParamSpec *pspec)
{
        WebKitLoadStatus status(webkit_web_view_get_load_status(webview));
        GObject *object(G_OBJECT(webview));
        g_assert(webview == WEBKIT_WEB_VIEW(m_portal_webview) || webview == WEBKIT_WEB_VIEW(m_validate_webview));
        g_object_freeze_notify(object);
	switch (status) {
        case WEBKIT_LOAD_FINISHED:
		if (m_verbose > 1U)
			std::cerr << "WebView: load finished" << std::endl;
		gtk_widget_show_all(GTK_WIDGET(webview));
		break;

        case WEBKIT_LOAD_PROVISIONAL:
		if (m_verbose > 1U)
			std::cerr << "WebView: load provisional" << std::endl;
                break;

        case WEBKIT_LOAD_COMMITTED:
		if (m_verbose > 1U)
			std::cerr << "WebView: load committed" << std::endl;
                break;

        case WEBKIT_LOAD_FIRST_VISUALLY_NON_EMPTY_LAYOUT:
		if (m_verbose > 1U)
			std::cerr << "WebView: load first visually nonempty layout" << std::endl;
                break;

        case WEBKIT_LOAD_FAILED:
		if (m_verbose > 1U)
			std::cerr << "WebView: load failed" << std::endl;
		set_wvstate(wvstate_failed);
                break;

        default:
                break;
        }
        g_object_thaw_notify(object);
}

gboolean CFMUValidateServer::policy_decision_required_1(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitNetworkRequest *request,
							      WebKitWebNavigationAction *action, WebKitWebPolicyDecision *decision, gpointer userdata)
{
	return static_cast<CFMUValidateServer *>(userdata)->policy_decision_required(webview, webframe, request, action, decision);
}

gboolean CFMUValidateServer::policy_decision_required(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitNetworkRequest *request,
							    WebKitWebNavigationAction *action, WebKitWebPolicyDecision *decision)
{
        g_assert(webview == WEBKIT_WEB_VIEW(m_portal_webview) || webview == WEBKIT_WEB_VIEW(m_validate_webview));
	WebKitWebNavigationReason reason(webkit_web_navigation_action_get_reason(action));
        gint button(webkit_web_navigation_action_get_button(action));
        gint state(webkit_web_navigation_action_get_modifier_state(action));
        const char *uri(webkit_network_request_get_uri(request));
	if (m_verbose > 1U)
		std::cerr << "policy_decision_required: reason " << reason << " button " << button << " state " << state
			  << " uri " << uri << std::endl;
	return false;
}

void CFMUValidateServer::resourceload_finished_1(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitWebResource *webresource, gpointer userdata)
{
	static_cast<CFMUValidateServer *>(userdata)->resourceload_finished(webview, webframe, webresource);
}

void CFMUValidateServer::resourceload_finished(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitWebResource *webresource)
{
        g_assert(webview == WEBKIT_WEB_VIEW(m_portal_webview) || webview == WEBKIT_WEB_VIEW(m_validate_webview));
	if (m_verbose > 1U)
		std::cerr << "resource load finished! state " << (int)m_wvstate << std::endl;
	switch (m_wvstate) {
	case wvstate_loadmainpage:
	{
		if (webview != WEBKIT_WEB_VIEW(m_portal_webview))
			break;
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "IFPUV_LAUNCH_AREA.FREE_TEXT_EDIT_LINK_LABEL"));
		if (!element)
			break;
		if (m_verbose > 2U)
			std::cerr << "Button found! Injecting script" << std::endl;
		webkit_web_view_execute_script(webview, "document.getElementById(\"IFPUV_LAUNCH_AREA.FREE_TEXT_EDIT_LINK_LABEL\").click();");
		set_wvstate(wvstate_loadquerypage);
		break;
	}

	case wvstate_loadquerypage:
	{
		if (webview != WEBKIT_WEB_VIEW(m_validate_webview))
			break;
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.VALIDATE_ACTION_LABEL"));
		if (!element)
			break;
		if (m_verbose > 1U) {
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
			if (typ)
				std::cerr << "Validate element type " << typ << std::endl;
		}
		element = webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.GENERAL_DATA_ENTRY.INTRODUCE_FLIGHT_PLAN_FIELD");
		if (!element)
			break;
		if (m_verbose > 1U) {
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
			if (typ)
				std::cerr << "FPL entry element type " << typ << std::endl;
		}
		if (!WEBKIT_DOM_IS_HTML_TEXT_AREA_ELEMENT(element))
			break;
		element = webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA.ERRORS_DATA_TABLE");
		if (!element)
			break;
		if (m_verbose > 1U) {
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
			if (typ)
				std::cerr << "Result Errors element type " << typ << std::endl;
		}
		element = webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA");
		if (!element)
			break;
		webkit_dom_event_target_add_event_listener(WEBKIT_DOM_EVENT_TARGET(element), "DOMSubtreeModified", G_CALLBACK(&result_change_1), true, this);
		if (m_verbose > 1U) {
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
			if (typ)
				std::cerr << "Results element type " << typ << std::endl;
		}
		m_delayedstatechangeconn.disconnect();
		m_delayedstatechangeconn = Glib::signal_timeout().connect(sigc::bind_return(sigc::bind(sigc::mem_fun(*this, &CFMUValidateServer::set_wvstate), wvstate_ready), false), 3000);
		break;
	}

	case wvstate_query:
	{
		if (webview != WEBKIT_WEB_VIEW(m_validate_webview))
			break;
		set_wvstate(wvstate_waitresult);
		parse_results();
		break;
	}

	default:
		break;
	}
}

gboolean CFMUValidateServer::result_change_1(WebKitDOMElement *target, WebKitDOMEvent *e, gpointer userdata)
{
	return static_cast<CFMUValidateServer *>(userdata)->result_change(target, e);
}

gboolean CFMUValidateServer::result_change(WebKitDOMElement *target, WebKitDOMEvent *e)
{
	std::cerr << "result change! state " << (int)m_wvstate << std::endl;
	switch (m_wvstate) {
	case wvstate_waitresult:
		parse_results();
		break;

	default:
		break;
	}
	return true;
}

gboolean CFMUValidateServer::console_message_1(WebKitWebView *webview, gchar *message, gint line, gchar *source_id, gpointer userdata)
{
	return static_cast<CFMUValidateServer *>(userdata)->console_message(webview, message, line, source_id);
}

gboolean CFMUValidateServer::console_message(WebKitWebView *webview, gchar *message, gint line, gchar *source_id)
{
	if (m_verbose > 0U) {
		std::cerr << message << " (" << (source_id ? "*null*" : source_id) << ':' << line << ')' << std::endl;
	}
	return true;
}

int main(int argc, char *argv[])
{
        try {
		g_type_init();
                Glib::init();
		Gio::init();
                Glib::set_application_name("CFMU Validate Server");
                Glib::thread_init();
		{
			bool found(false);
			std::string disp(Glib::getenv("DISPLAY", found));
			if (found)
				gtk_init(&argc, &argv);
		}
                CFMUValidateServer app(argc, argv);
                app.run();
                return 0;
        } catch(const Glib::FileError& ex) {
                std::cerr << "FileError: " << ex.what() << std::endl;
                return 1;
        } catch(const Gtk::BuilderError& ex) {
                std::cerr << "BuilderError: " << ex.what() << std::endl;
                return 1;
        }
        return 1;
}
