#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <getopt.h>
#include <iostream>
#include <iomanip>
#include <glibmm.h>
#include <stdexcept>

#ifdef HAVE_SYSEXITS_H
#include <sysexits.h>
#else
#define EX_OK           0
#define EX_USAGE        64
#define EX_DATAERR      65
#define EX_NOINPUT      66
#define EX_UNAVAILABLE  69
#endif  

#include "engine.h"
#include "aircraft.h"
#include "icaofpl.h"
#include "tfr.hh"
#include "tfrproxy.hh"

class Command : public FPlanCheckerProxy::Command {
public:
	Command(void) : FPlanCheckerProxy::Command() {}
	Command(const std::string& cmd) : FPlanCheckerProxy::Command(cmd) {}

	using FPlanCheckerProxy::Command::set_option;
};

static Glib::RefPtr<Glib::MainLoop> mainloop;
typedef void (*cmd_t)(const Command& cmdin, Command& cmdout);
typedef std::map<std::string,cmd_t> cmdlist_t;
cmdlist_t cmdlist;
Glib::RefPtr<Glib::IOChannel> inputchan;
Glib::RefPtr<Glib::IOChannel> outputchan;
AirportsDb airportdb;
NavaidsDb navaiddb;
WaypointsDb waypointdb;
AirwaysDb airwaydb;
AirspacesDb airspacedb;
Engine *eng(0);
Glib::ustring dir_main(""), dir_aux("");
Engine::auxdb_mode_t auxdbmode(Engine::auxdb_prefs);
TrafficFlowRestrictions tfr;

static void machine_writecmd(const Command& cmd)
{
        outputchan->write(cmd.get_cmdstring() + "\n");
}

static bool machine_input(Glib::IOCondition iocond)
{
        Glib::ustring line;
        Glib::IOStatus iostat(inputchan->read_line(line));
        if (iostat == Glib::IO_STATUS_ERROR ||
            iostat == Glib::IO_STATUS_EOF) {
                mainloop->quit();
                return true;
        }
        while (!line.empty()) {
                char ch(line[line.size() - 1]);
                if (ch != '\r' && ch != '\n')
                        break;
                line.erase(line.size() - 1, 1);
        }
        Command cmdin(line);
        Command cmdout;
        {
                std::string cmdname("checkfplan");
                cmdname = cmdin.get_cmdname();
                if (cmdname.empty())
                        cmdname = "checkfplan";
                cmdout.set_cmdname(cmdname);
                {
                        std::pair<const std::string&,bool> seq(cmdin.get_option("cmdseq"));
                        if (seq.second)
                                cmdout.set_option("cmdseq", seq.first);
                }
        }
        try {
                Command cmdout1(cmdout);
                cmdlist_t::const_iterator cmdi(cmdlist.find(cmdin.get_cmdname()));
                if (cmdi == cmdlist.end()) {
                        cmdout1.set_option("error", "command not found");
                } else {
                        (cmdi->second)(cmdin, cmdout1);
                }
                machine_writecmd(cmdout1);
        } catch (const std::runtime_error& e) {
                cmdout.set_option("error", e.what());
		machine_writecmd(cmdout);
        } catch (const Glib::Exception& e) {
                cmdout.set_option("error", e.what());
                machine_writecmd(cmdout);
        }
        outputchan->flush();
        return true;
}

static void cmd_nop(const Command& cmdin, Command& cmdout)
{
}

static void cmd_quit(const Command& cmdin, Command& cmdout)
{
	mainloop->quit();
}

static bool normal_input(Glib::IOCondition iocond)
{
        Glib::ustring line;
        Glib::IOStatus iostat(inputchan->read_line(line));
        if (iostat == Glib::IO_STATUS_ERROR ||
            iostat == Glib::IO_STATUS_EOF) {
                mainloop->quit();
                return true;
        }
        while (!line.empty()) {
                char ch(line[line.size() - 1]);
                if (ch != '\r' && ch != '\n')
                        break;
                line.erase(line.size() - 1, 1);
        }
	FPlanRoute fpl(*(FPlan *)0);
	char type_of_flight('G');
	std::string acft_type;
	std::string equipment;
	{
		IcaoFlightPlan icaofpl(*eng);
		std::string err(icaofpl.parse(line, false));
		if (!err.empty()) {
			outputchan->write("EFPL: " + err + "\n");
			outputchan->flush();
			return true;
		}
		icaofpl.set_route(fpl);
		type_of_flight = icaofpl.get_flighttype();
		acft_type = icaofpl.get_aircrafttype();
		equipment = icaofpl.get_equipment();
		if (false) {
			std::ostringstream oss;
			oss << "I: FPL Parsed, " << fpl.get_nrwpt() << " waypoints";
			outputchan->write(oss.str() + "\n");
		}
	}
	// print Parsed Flight Plan
	if (false) {
		for (unsigned int i = 0; i < fpl.get_nrwpt(); ++i) {
			const FPlanWaypoint& wpt(fpl[i]);
			std::ostringstream oss;
			oss << "PFPL[" << i << "]: " << wpt.get_icao() << '/' << wpt.get_name() << ' ';
			switch (wpt.get_pathcode()) {
			case FPlanWaypoint::pathcode_sid:
                        case FPlanWaypoint::pathcode_star:
			case FPlanWaypoint::pathcode_airway:
				oss << wpt.get_pathname();
				break;

                        case FPlanWaypoint::pathcode_directto:
				oss << "DCT";
				break;

			default:
				oss << '-';
				break;
			}
			oss << ' ' << wpt.get_coord().get_lat_str2() << ' ' << wpt.get_coord().get_lon_str2()
			    << ' ' << (wpt.is_standard() ? 'F' : 'A') << std::setw(3) << std::setfill('0') << ((wpt.get_altitude() + 50) / 100)
			    << ' ' << (wpt.is_ifr() ? 'I' : 'V') << "FR";
			outputchan->write(oss.str() + "\n");
		}
	}
	TrafficFlowRestrictions::Result res(tfr.check_fplan(fpl, type_of_flight, acft_type, equipment,
							    airportdb, navaiddb, waypointdb, airwaydb, airspacedb));
	// print Flight Plan
	{
		{
			std::ostringstream oss;
			oss << "FR: " << res.get_aircrafttype() << ' ' << res.get_aircrafttypeclass() << ' '
			    << res.get_equipment() << ' ' << res.get_typeofflight();
			outputchan->write(oss.str() + "\n");
		}
		const TrafficFlowRestrictions::Fpl& fpl(res.get_fplan());
		for (unsigned int i = 0; i < fpl.size(); ++i) {
			const TrafficFlowRestrictions::FplWpt& wpt(fpl[i]);
			std::ostringstream oss;
			oss << "F[" << i << '/' << wpt.get_wptnr() << "]: " << wpt.get_ident() << ' ' << to_str(wpt.get_type());
			if (wpt.get_type() == TrafficFlowRestrictions::Vertex::type_navaid)
				oss << '/' << NavaidsDb::Navaid::get_navaid_typename(wpt.get_navaid_type());
			oss << ' ';
			switch (wpt.get_pathcode()) {
			case FPlanWaypoint::pathcode_sid:
                        case FPlanWaypoint::pathcode_star:
			case FPlanWaypoint::pathcode_airway:
				oss << wpt.get_pathname();
				break;

                        case FPlanWaypoint::pathcode_directto:
				oss << "DCT";
				break;

			default:
				oss << '-';
				break;
			}
			oss << ' ' << wpt.get_coord().get_lat_str2() << ' ' << wpt.get_coord().get_lon_str2()
			    << ' ' << (wpt.is_standard() ? 'F' : 'A') << std::setw(3) << std::setfill('0') << ((wpt.get_altitude() + 50) / 100)
			    << ' ' << wpt.get_icao() << '/' << wpt.get_name() << ' ' << (wpt.is_ifr() ? 'I' : 'V') << "FR";
			outputchan->write(oss.str() + "\n");
		}
	}
	// print Messages
	for (unsigned int i = 0; i < res.messages_size(); ++i) {
		const TrafficFlowRestrictions::Message& msg(res.get_message(i));
		std::ostringstream oss;
		oss << msg.get_type_char() << ':';
		if (!msg.get_rule().empty())
			oss << " R:" << msg.get_rule();
		if (!msg.get_vertexset().empty()) {
			oss << " V:";
			bool subseq(false);
			for (TrafficFlowRestrictions::Message::set_t::const_iterator i(msg.get_vertexset().begin()), e(msg.get_vertexset().end()); i != e; ++i) {
				if (subseq)
					oss << ',';
				subseq = true;
				oss << *i;
			}
		}
		if (!msg.get_edgeset().empty()) {
			oss << " E:";
			bool subseq(false);
			for (TrafficFlowRestrictions::Message::set_t::const_iterator i(msg.get_edgeset().begin()), e(msg.get_edgeset().end()); i != e; ++i) {
				if (subseq)
					oss << ',';
				subseq = true;
				oss << *i;
			}
		}
		oss << ' ' << msg.get_text();
		outputchan->write(oss.str() + "\n");
	}
	outputchan->write(res ? "OK\n" : "FAIL\n");
	for (unsigned int i = 0; i < res.size(); ++i) {
		const TrafficFlowRestrictions::RuleResult& rule(res[i]);
		std::ostringstream oss;
		oss << "Rule: " << rule.get_rule() << ' ';
		switch (rule.get_codetype()) {
		case TrafficFlowRestrictions::RuleResult::codetype_mandatory:
			oss << "Mandatory";
			break;

		case TrafficFlowRestrictions::RuleResult::codetype_forbidden:
			oss << "Forbidden";
			break;

		case TrafficFlowRestrictions::RuleResult::codetype_closed:
			oss << "ClosedForCruising";
			break;

		default:
			oss << '?';
			break;
		}
		if (!rule.get_vertexset().empty()) {
			oss << " V:";
			bool subseq(false);
			for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(rule.get_vertexset().begin()), e(rule.get_vertexset().end()); i != e; ++i) {
				if (subseq)
					oss << ',';
				subseq = true;
				oss << *i;
			}
		}
		if (!rule.get_edgeset().empty()) {
			oss << " E:";
			bool subseq(false);
			for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(rule.get_edgeset().begin()), e(rule.get_edgeset().end()); i != e; ++i) {
				if (subseq)
					oss << ',';
				subseq = true;
				oss << *i;
			}
		}
		outputchan->write(oss.str() + "\n");
		outputchan->write("Desc: " + rule.get_desc() + "\n");
		outputchan->write("OprGoal: " + rule.get_oprgoal() + "\n");
		for (unsigned int i = 0; i < rule.size(); ++i) {
			const TrafficFlowRestrictions::RuleResult::Alternative& alt(rule[i]);
			{
				std::ostringstream oss;
				oss << "  Alternative " << i;
				if (!alt.get_vertexset().empty()) {
					oss << " V:";
					bool subseq(false);
					for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(alt.get_vertexset().begin()), e(alt.get_vertexset().end()); i != e; ++i) {
						if (subseq)
							oss << ',';
						subseq = true;
						oss << *i;
					}
				}
				if (!alt.get_edgeset().empty()) {
					oss << " E:";
					bool subseq(false);
					for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(alt.get_edgeset().begin()), e(alt.get_edgeset().end()); i != e; ++i) {
						if (subseq)
							oss << ',';
						subseq = true;
						oss << *i;
					}
				}
				outputchan->write(oss.str() + "\n");
			}
			for (unsigned int j = 0; j < alt.size(); ++j) {
				const TrafficFlowRestrictions::RuleResult::Alternative::Sequence& el(alt[j]);
				std::ostringstream oss;
				oss << "    " << el.get_type_string();
				switch (el.get_type()) {
				case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airway:
					oss << ' ' << el.get_ident();
					break;

				case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_sid:
					oss << ' ' << el.get_ident() << ' ' << el.get_airport();
					break;

				case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_star:
					oss << ' ' << el.get_ident() << ' ' << el.get_airport();
					break;

				case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airspace:
					oss << ' ' << el.get_ident() << '/'
					    << AirspacesDb::Airspace::get_class_string(el.get_bdryclass(), AirspacesDb::Airspace::typecode_ead);
					break;

				case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct:
				case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_point:
				default:
					break;
				}
				oss << " F" << std::setw(3) << std::setfill('0') << ((el.get_lower_alt() + 50) / 100) << "...F"
				    << std::setw(3) << std::setfill('0') << ((std::min((int32_t)el.get_upper_alt(), (int32_t)99900) + 50) / 100);
				outputchan->write(oss.str() + "\n");
				if (el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airway ||
				    el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct ||
				    el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_point) {
					const TrafficFlowRestrictions::RuleResult::Alternative::Sequence::RulePoint& pt(el.get_point0());
					std::ostringstream oss;
					oss << "      " << pt.get_ident() << ' ' << pt.get_coord().get_lat_str2() << ' '
					    << pt.get_coord().get_lon_str2() << ' ' << to_str(pt.get_type());
					outputchan->write(oss.str() + "\n");
				}
				if (el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airway ||
				    el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct) {
					const TrafficFlowRestrictions::RuleResult::Alternative::Sequence::RulePoint& pt(el.get_point1());
					std::ostringstream oss;
					oss << "      " << pt.get_ident() << ' ' << pt.get_coord().get_lat_str2() << ' '
					    << pt.get_coord().get_lon_str2() << ' ' << to_str(pt.get_type());
					outputchan->write(oss.str() + "\n");
				}
			}
		}
	}
        outputchan->flush();
        return true;
}

int main(int argc, char *argv[])
{
        try {
                Glib::init();
                Glib::set_application_name("Flight Plan Checker");
                Glib::thread_init();
		mainloop = Glib::MainLoop::create(Glib::MainContext::get_default());
		bool dis_aux(false), machine(false), newparser(false);

		srand(time(0));

		{
			static struct option long_options[] = {
				{ "maindir", required_argument, 0, 'm' },
				{ "auxdir", required_argument, 0, 'a' },
				{ "disableaux", no_argument, 0, 'A' },
				{ "machineinterface", no_argument, 0, 0x100 },
				{ "trace", required_argument, 0, 't' },
				{ "newparser", no_argument, 0, 'n' },
				{0, 0, 0, 0}
			};
			int c, err(0);
			while ((c = getopt_long(argc, argv, "m:a:At:n", long_options, 0)) != EOF) {
				switch (c) {
				case 'm':
					if (optarg)
						dir_main = optarg;
					break;

				case 'a':
					if (optarg)
						dir_aux = optarg;
					break;

				case 'A':
					dis_aux = true;
					break;

				case 0x100:
					machine = true;
					break;

				case 'n':
					newparser = true;
					break;

				case 't':
					for (const char *cp = optarg; cp; ) {
						char *cp1(const_cast<char *>(strchr(cp, ',')));
						if (cp1)
							*cp1++ = 0;
						tfr.tracerules_add(Glib::ustring(cp).uppercase());
						cp = cp1;
					}
					break;

				default:
					err++;
					break;
				}
			}
			if (err) {
				std::cerr << "usage: checkfplan [-m <maindir>] [-a <auxdir>] [-A] [-t <trace>]" << std::endl;
				return EX_USAGE;
			}
		}
		if (dis_aux)
			auxdbmode = Engine::auxdb_none;
		else if (!dir_aux.empty())
			auxdbmode = Engine::auxdb_override;
		eng = new Engine(dir_main, auxdbmode, dir_aux);
		dir_aux = eng->get_aux_dir(auxdbmode, dir_aux);
		airportdb.open(dir_main);
		if (!dir_aux.empty() && airportdb.is_dbfile_exists(dir_aux))
			airportdb.attach(dir_aux);
		if (airportdb.is_aux_attached())
			std::cerr << "Auxillary airports database attached" << std::endl;
		navaiddb.open(dir_main);
		if (!dir_aux.empty() && navaiddb.is_dbfile_exists(dir_aux))
			navaiddb.attach(dir_aux);
		if (navaiddb.is_aux_attached())
			std::cerr << "Auxillary navaids database attached" << std::endl;
		waypointdb.open(dir_main);
		if (!dir_aux.empty() && waypointdb.is_dbfile_exists(dir_aux))
			waypointdb.attach(dir_aux);
		if (waypointdb.is_aux_attached())
			std::cerr << "Auxillary waypoints database attached" << std::endl;
		airwaydb.open(dir_main);
		if (!dir_aux.empty() && airwaydb.is_dbfile_exists(dir_aux))
			airwaydb.attach(dir_aux);
		if (airwaydb.is_aux_attached())
			std::cerr << "Auxillary airways database attached" << std::endl;
		airspacedb.open(dir_main);
		if (!dir_aux.empty() && airspacedb.is_dbfile_exists(dir_aux))
			airspacedb.attach(dir_aux);
		if (airspacedb.is_aux_attached())
			std::cerr << "Auxillary airspaces database attached" << std::endl;
		if (newparser) {
			Glib::ustring fname(Glib::build_filename(dir_aux, "eRAD.xml"));
			std::vector<TrafficFlowRestrictions::Message> msg;
			bool ld(tfr.load_rules(msg, fname, airportdb, navaiddb, waypointdb, airwaydb, airspacedb));
			for (unsigned int i = 0; i < msg.size(); ++i) {
				const TrafficFlowRestrictions::Message& m(msg[i]);
				std::cout << m.get_type_char() << ':';
				if (!m.get_rule().empty())
					std::cout << " R:" << m.get_rule();
				std::cout << ' ' << m.get_text() << std::endl;
			}
			if (!ld) {
				std::cerr << "Cannot load Traffic Flow Restrictions file " << fname << std::cerr;
				return EX_UNAVAILABLE;
			} else if (true) {
				tfr.reset_rules();
				tfr.prepare_dct_rules();
				std::cerr << tfr.count_rules() << " rules, " << tfr.count_dct_rules() << " DCT rules" << std::endl;
			}
		} else {
			Glib::ustring fname(Glib::build_filename(dir_aux, "eRAD.xml"));
			if (!tfr.load(fname)) {
				std::cerr << "Cannot load Traffic Flow Restrictions file " << fname << std::cerr;
				return EX_UNAVAILABLE;
			}
		}
		inputchan = Glib::IOChannel::create_from_fd(0);
		outputchan = Glib::IOChannel::create_from_fd(1);
		inputchan->set_encoding();
		outputchan->set_encoding();
		if (machine) {
			delete eng;
			eng = 0;
			cmdlist["nop"] = cmd_nop;
			cmdlist["quit"] = cmd_quit;
			Glib::signal_io().connect(sigc::ptr_fun(&machine_input), inputchan, Glib::IO_IN | Glib::IO_ERR | Glib::IO_HUP);
			{
				Command cmd("checkfplan");
				cmd.set_option("version", VERSION);
				machine_writecmd(cmd);
			}
			outputchan->flush();
			mainloop->run();
			return EX_OK;
		}
		if (!tfr.tracerules_empty()) {
			std::ostringstream oss;
			for (TrafficFlowRestrictions::tracerules_const_iterator_t i(tfr.tracerules_begin()), e(tfr.tracerules_end()); i != e; ++i)
				oss << ',' << *i;
			outputchan->write("I: Trace Rules: " + oss.str().substr(1) + "\n");
			outputchan->flush();
		}
		Glib::signal_io().connect(sigc::ptr_fun(&normal_input), inputchan, Glib::IO_IN | Glib::IO_ERR | Glib::IO_HUP);
		mainloop->run();
	} catch (const Glib::FileError& ex) {
                std::cerr << "FileError: " << ex.what() << std::endl;
                return EX_DATAERR;
        } catch (const Glib::Exception& ex) {
                std::cerr << "Glib exception: " << ex.what() << std::endl;
                return EX_DATAERR;
        } catch (const std::exception& ex) {
                std::cerr << "exception: " << ex.what() << std::endl;
                return EX_DATAERR;
        }
        return EX_OK;
}
