#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <libxml++/nodes/textnode.h>

#include "tfr.hh"
#include "aircraft.h"

const Glib::ustring& TrafficFlowRestrictions::FplWpt::get_ident(void) const
{
	switch (get_type()) {
	default:
	case Vertex::type_airport:
		if (get_icao() == "ZZZZ")
			return get_name();
		/* fall through */

	case Vertex::type_undefined:
		if (get_icao().empty())
			return get_name();
		/* fall through */

	case Vertex::type_navaid:
		return get_icao();

	case Vertex::type_intersection:
		return get_name();
	}
}

bool TrafficFlowRestrictions::FplWpt::operator==(const RuleWpt& r) const
{
	return get_type() == r.get_type() &&
		get_ident() == r.get_ident() &&
		(has_vor() || !r.is_vor()) &&
		(has_ndb() || !r.is_ndb()) &&
		get_coord().spheric_distance_nmi_dbl(r.get_coord()) < 0.1;
}

bool TrafficFlowRestrictions::FplWpt::within(const AltRange& ar) const
{
	// we do not handle modes correctly; and we probably cannot reasonably, as we do not know the 
	// local QNH in advance
	return get_altitude() >= ar.get_lower_alt() && get_altitude() <= ar.get_upper_alt();
}

TrafficFlowRestrictions::Fpl& TrafficFlowRestrictions::Fpl::operator=(const FPlanRoute& rte)
{
	clear();
	for (unsigned int i = 0; i < rte.get_nrwpt(); ++i) {
		FplWpt w;
		(FPlanWaypoint&)w = rte[i];
		push_back(w);
	}
	return *this;
}

Rect TrafficFlowRestrictions::Fpl::get_bbox(void) const
{
	if (empty())
		return Rect();
	Rect r(front().get_coord(), front().get_coord());
	for (const_iterator i(begin()), e(end()); i != e; ++i)
		r = r.add(i->get_coord());
	return r;
}

const std::string& TrafficFlowRestrictions::Message::get_type_string(type_t t)
{
	switch (t) {
	case type_error:
	{
		static const std::string r("error");
		return r;
	}

	case type_warning:
	{
		static const std::string r("warning");
		return r;
	}

	case type_info:
	{
		static const std::string r("info");
		return r;
	}

	case type_tracecond:
	{
		static const std::string r("tracecond");
		return r;
	}

	case type_tracetfe:
	{
		static const std::string r("tracetfe");
		return r;
	}

	default:
	{
		static const std::string r("?");
		return r;
	}
	}
}

char TrafficFlowRestrictions::Message::get_type_char(type_t t)
{
	switch (t) {
	case type_error:
		return 'E';

	case type_warning:
		return 'W';

	case type_info:
		return 'I';

	case type_tracecond:
		return 'C';

	case type_tracetfe:
		return 'T';

	default:
		return '?';
	}
}

void TrafficFlowRestrictions::Message::add_vertex(unsigned int nr)
{
	m_vertexset.insert(nr);
}

void TrafficFlowRestrictions::Message::add_edge(unsigned int nr)
{
	m_edgeset.insert(nr);
}

void TrafficFlowRestrictions::Message::insert_vertex(unsigned int nr)
{
	{
		set_t x;
		for (set_t::const_iterator i(m_vertexset.begin()), e(m_vertexset.end()); i != e; ++i) {
			if (*i >= nr)
				x.insert(1U + *i);
			else
				x.insert(*i);
		}
		m_vertexset.swap(x);
	}
	{
		set_t x;
		for (set_t::const_iterator i(m_edgeset.begin()), e(m_edgeset.end()); i != e; ++i) {
			if (*i >= nr)
				x.insert(1U + *i);
			else
				x.insert(*i);
		}
		m_edgeset.swap(x);
	}			
}

void TrafficFlowRestrictions::message(const Glib::ustring& text, Message::type_t t, const Glib::ustring& rule)
{
	message(Message(text, t, rule));
}

void TrafficFlowRestrictions::message(const Message& msg)
{
	m_messages.push_back(msg);
}

TrafficFlowRestrictions::RuleResult::Alternative::Sequence::RulePoint::RulePoint(const Glib::ustring& id, const Point& pt, Vertex::type_t typ, bool vor, bool ndb)
	: m_ident(id), m_coord(pt), m_type(typ), m_vor(vor), m_ndb(ndb)
{
	if (m_type != Vertex::type_navaid)
		m_vor = m_ndb = false;
	m_ndb = m_ndb && !m_vor;
}

TrafficFlowRestrictions::RuleResult::Alternative::Sequence::RulePoint::RulePoint(const FplWpt& wpt, bool vor, bool ndb)
	: m_ident(wpt.get_ident()), m_coord(wpt.get_coord()), m_type(wpt.get_type()), m_vor(vor), m_ndb(ndb)
{
	if (m_type != Vertex::type_navaid)
		m_vor = m_ndb = false;
	m_ndb = m_ndb && !m_vor;
}

TrafficFlowRestrictions::RuleResult::Alternative::Sequence::RulePoint::RulePoint(const RuleWpt& wpt)
	: m_ident(wpt.get_ident()), m_coord(wpt.get_coord()), m_type(wpt.get_type()), m_vor(wpt.is_vor()), m_ndb(wpt.is_ndb())
{
}

TrafficFlowRestrictions::RuleResult::Alternative::Sequence::RulePoint::RulePoint(const Vertex& vv)
	: m_ident(vv.get_ident()), m_coord(vv.get_coord()), m_type(vv.get_type()), m_vor(false), m_ndb(false)
{
	if (m_type == Vertex::type_navaid) {
		NavaidsDb::Navaid n;
		if (vv.get(n) && n.is_valid()) {
			m_vor = n.has_vor();
			m_ndb = n.has_ndb();
		}
	}
}

const std::string& TrafficFlowRestrictions::RuleResult::Alternative::Sequence::get_type_string(type_t t)
{
	switch (t) {
	case type_airway:
	{
		static const std::string r("AWY");
		return r;
	}

	case type_dct:
	{
		static const std::string r("DCT");
		return r;
	}

	case type_point:
	{
		static const std::string r("POINT");
		return r;
	}

	case type_sid:
	{
		static const std::string r("SID");
		return r;
	}

	case type_star:
	{
		static const std::string r("STAR");
		return r;
	}

	case type_airspace:
	{
		static const std::string r("AIRSPACE");
		return r;
	}

	case type_invalid:
	default:
	{
		static const std::string r("INVALID");
		return r;
	}
	}
}

TrafficFlowRestrictions::Result::Result(const std::string& acft, const std::string& equip, char typeofflight)
	: m_aircrafttype(acft), m_equipment(equip), m_typeofflight(typeofflight), m_result(true)
{
	m_aircrafttype.resize(4, ' ');
	if (!m_aircrafttype.empty())
		m_aircrafttypeclass = Aircraft::get_aircrafttypeclass(get_aircrafttype());
	m_aircrafttypeclass.resize(3, '?');
}

TrafficFlowRestrictions::DctParameters::AltRange::AltRange(const set_t& set0, const set_t& set1)
{
	m_set[0] = set0;
	m_set[1] = set1;
}

bool TrafficFlowRestrictions::DctParameters::AltRange::is_inside(unsigned int index, type_t alt) const
{
	return m_set[!!index].is_inside(alt);
}

bool TrafficFlowRestrictions::DctParameters::AltRange::is_empty(void) const
{
	return m_set[0].is_empty() && m_set[1].is_empty();
}

void TrafficFlowRestrictions::DctParameters::AltRange::set_empty(void)
{
	m_set[0].set_empty();
	m_set[1].set_empty();
}

void TrafficFlowRestrictions::DctParameters::AltRange::set_full(void)
{
	m_set[0].set_full();
	m_set[1].set_full();
}

void TrafficFlowRestrictions::DctParameters::AltRange::set_interval(type_t alt0, type_t alt1)
{
	m_set[0] = m_set[1] = set_t::Interval(alt0, alt1);
}

void TrafficFlowRestrictions::DctParameters::AltRange::invert(void)
{
	m_set[0] = ~m_set[0];
	m_set[1] = ~m_set[1];
}

TrafficFlowRestrictions::DctParameters::AltRange& TrafficFlowRestrictions::DctParameters::AltRange::operator&=(const AltRange& x)
{
	m_set[0] &= x.m_set[0];
	m_set[1] &= x.m_set[1];
	return *this;
}

TrafficFlowRestrictions::DctParameters::AltRange& TrafficFlowRestrictions::DctParameters::AltRange::operator|=(const AltRange& x)
{
	m_set[0] |= x.m_set[0];
	m_set[1] |= x.m_set[1];
	return *this;
}

TrafficFlowRestrictions::DctParameters::AltRange& TrafficFlowRestrictions::DctParameters::AltRange::operator^=(const AltRange& x)
{
	m_set[0] ^= x.m_set[0];
	m_set[1] ^= x.m_set[1];
	return *this;
}

TrafficFlowRestrictions::DctParameters::AltRange& TrafficFlowRestrictions::DctParameters::AltRange::operator-=(const AltRange& x)
{
	m_set[0] -= x.m_set[0];
	m_set[1] -= x.m_set[1];
	return *this;
}

TrafficFlowRestrictions::DctParameters::AltRange TrafficFlowRestrictions::DctParameters::AltRange::operator~(void) const
{
	return AltRange(~m_set[0], ~m_set[1]);
}

TrafficFlowRestrictions::DctParameters::DctParameters(const Glib::ustring& ident0, const Point& coord0,
						      const Glib::ustring& ident1, const Point& coord1,
						      AltRange::type_t alt0, AltRange::type_t alt1)
	: m_alt(AltRange::set_t::Interval(alt0, alt1), AltRange::set_t::Interval(alt0, alt1)),
	  m_dist(std::numeric_limits<double>::quiet_NaN())
{
	m_ident[0] = ident0;
	m_ident[1] = ident1;
	m_coord[0] = coord0;
	m_coord[1] = coord1;
}

double TrafficFlowRestrictions::DctParameters::get_dist(void) const
{
	if (std::isnan(m_dist))
		m_dist = m_coord[0].spheric_distance_nmi_dbl(m_coord[1]);
	return m_dist;
}

TrafficFlowRestrictions::Rule::Rule(xmlpp::Document *doc)
	: m_el(0), m_condel(0), m_codetype(codetype_invalid)
{
	if (!doc)
		return;
	xmlpp::Element *el(doc->get_root_node());
	if (!el)
		return;
	if (el->get_name() != "AIXM-Snapshot")
		return;
	xmlpp::Node::NodeList nl(el->get_children("Tfr"));
	for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
		xmlpp::Element *el(dynamic_cast<xmlpp::Element *>(*ni));
		if (!el)
			continue;
		m_el = el;
		parse();
		return;
	}
}

char TrafficFlowRestrictions::Rule::get_codetype_char(codetype_t t)
{
	switch (t) {
	default:
	case codetype_invalid:
		return '-';

	case codetype_mandatory:
		return 'M';

	case codetype_forbidden:
		return 'F';

	case codetype_closed:
		return 'C';
	}
}

bool TrafficFlowRestrictions::Rule::is_valid(void) const
{
	if (!m_el)
		return false;
	if (get_codetype() == codetype_invalid)
		return false;
	return true;
}

void TrafficFlowRestrictions::Rule::next_rule(void)
{
	if (!m_el)
		return;
	for (xmlpp::Node *n = m_el->get_next_sibling(); n; n = n->get_next_sibling()) {
		if (n->get_name() != "Tfr")
			continue;
		xmlpp::Element *el(dynamic_cast<xmlpp::Element *>(n));
		if (!el)
			continue;
		m_el = el;
		parse();
		return;
	}
	m_el = 0;
	parse();
}

void TrafficFlowRestrictions::Rule::parse(void)
{
	m_codetype = codetype_invalid;
	m_codeid.clear();
	m_oprgoal.clear();
	m_descr.clear();
	m_condel = 0;
	if (!m_el)
		return;
	{
		xmlpp::Node::NodeList nl(m_el->get_children("TfrUid"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
			xmlpp::Element *el(dynamic_cast<xmlpp::Element *>(*ni));
			if (!el)
				continue;
			xmlpp::Node::NodeList nl2(el->get_children("codeId"));
			for (xmlpp::Node::NodeList::const_iterator ni(nl2.begin()), ne(nl2.end()); ni != ne; ++ni) {
				xmlpp::Element *el(dynamic_cast<xmlpp::Element *>(*ni));
				if (!el)
					continue;
				xmlpp::TextNode *txt(el->get_child_text());
				if (!txt)
					continue;
				if (txt->is_white_space())
					continue;
				m_codeid = txt->get_content();
				break;
			}
			if (!m_codeid.empty())
				break;
		}
	}
	if (m_codeid.empty())
		return;
	{
		xmlpp::Node::NodeList nl(m_el->get_children(""));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
			xmlpp::Element *el(dynamic_cast<xmlpp::Element *>(*ni));
			if (!el)
				continue;
			if (el->get_name() != "Fcc" && el->get_name() != "Fce")
				continue;
			if (!m_condel) {
				m_condel = el;
				continue;
			}
			std::cerr << "Warning: Rule " << m_codeid << " has multiple conditions" << std::endl;
		}
	}
	{
		xmlpp::Node::NodeList nl(m_el->get_children("txtOprGoal"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
			xmlpp::Element *el(dynamic_cast<xmlpp::Element *>(*ni));
			if (!el)
				continue;
			xmlpp::TextNode *txt(el->get_child_text());
			if (!txt)
				continue;
			if (txt->is_white_space())
				continue;
			m_oprgoal = txt->get_content();
			break;
		}
	}
	{
		xmlpp::Node::NodeList nl(m_el->get_children("txtDescr"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
			xmlpp::Element *el(dynamic_cast<xmlpp::Element *>(*ni));
			if (!el)
				continue;
			xmlpp::TextNode *txt(el->get_child_text());
			if (!txt)
				continue;
			if (txt->is_white_space())
				continue;
			m_descr = txt->get_content();
			break;
		}
	}
	{
		xmlpp::Node::NodeList nl(m_el->get_children("codeType"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
			xmlpp::Element *el(dynamic_cast<xmlpp::Element *>(*ni));
			if (!el)
				continue;
			xmlpp::TextNode *txt(el->get_child_text());
			if (!txt)
				continue;
			if (txt->is_white_space())
				continue;
		        Glib::ustring t(txt->get_content());
			for (codetype_t ct = codetype_mandatory; ct <= codetype_closed; ct = (codetype_t)(ct + 1)) {
				Glib::ustring ctt;
				ctt.push_back(get_codetype_char(ct));
				if (t != ctt)
					continue;
				m_codetype = ct;
				break;
			}
			if (m_codetype != codetype_invalid)
				break;
		}
	}
	if (m_codetype == codetype_invalid)
		return;
}

bool TrafficFlowRestrictions::CompareAirspaces::operator()(const AirspacesDb::Airspace& a, const AirspacesDb::Airspace& b)
{
	if (a.get_typecode() < b.get_typecode())
		return true;
	if (a.get_typecode() > b.get_typecode())
		return false;
	if (a.get_bdryclass() < b.get_bdryclass())
		return true;
	if (a.get_bdryclass() > b.get_bdryclass())
		return false;
	return a.get_icao() < b.get_icao();
}

TrafficFlowRestrictions::CondResult& TrafficFlowRestrictions::CondResult::operator&=(const CondResult& x)
{
	m_result = m_result && x.m_result;
	if (!m_result) {
		m_vertexset.clear();
		m_edgeset.clear();
		m_reflocset.clear();
	} else {
		m_vertexset.insert(x.m_vertexset.begin(), x.m_vertexset.end());
		m_edgeset.insert(x.m_edgeset.begin(), x.m_edgeset.end());
		m_reflocset.insert(x.m_reflocset.begin(), x.m_reflocset.end());
	}
	return *this;
}

TrafficFlowRestrictions::CondResult& TrafficFlowRestrictions::CondResult::operator|=(const CondResult& x)
{
	// determine smaller set of conditions
	if (!m_result) {
		if (!x.m_result) {
			m_vertexset.clear();
			m_edgeset.clear();
			m_reflocset.clear();
		} else {
			m_vertexset = x.m_vertexset;
			m_edgeset = x.m_edgeset;
			m_reflocset = x.m_reflocset;
		}
	} else if (!x.m_result) {
		// keep
	} else {
		if (x.m_vertexset.size() < m_vertexset.size())
			m_vertexset = x.m_vertexset;
		if (x.m_edgeset.size() < m_edgeset.size())
			m_edgeset = x.m_edgeset;
		m_reflocset.insert(x.m_reflocset.begin(), x.m_reflocset.end());
	}
	m_result = m_result || x.m_result;
	return *this;
}

TrafficFlowRestrictions::CondResult TrafficFlowRestrictions::CondResult::operator~(void)
{
	return CondResult(!m_result);
}

unsigned int TrafficFlowRestrictions::CondResult::get_first(void) const
{
	if (m_vertexset.empty()) {
		if (m_edgeset.empty())
			return 0U;
		return *m_edgeset.begin();
	}
	if (m_edgeset.empty())
		return *m_vertexset.begin();
	return std::min(*m_vertexset.begin(), *m_edgeset.begin());
}

unsigned int TrafficFlowRestrictions::CondResult::get_last(void) const
{
	if (m_vertexset.empty()) {
		if (m_edgeset.empty())
			return ~0U;
		return 1U + *--m_edgeset.end();
	}
	if (m_edgeset.empty())
		return *--m_vertexset.end();
	return std::min(*--m_vertexset.end(), 1U + *--m_edgeset.end());
}

const TrafficFlowRestrictions::Graph::vertex_descriptor TrafficFlowRestrictions::invalid_vertex_descriptor;

TrafficFlowRestrictions::Graph::vertex_descriptor TrafficFlowRestrictions::get_vertexdesc(const RuleWpt& wpt)
{
	Graph::vertex_descriptor ux;
	if (m_graph.find_intersection(ux, wpt.get_ident(), wpt.get_coord(), (Vertex::typemask_t)(1U << wpt.get_type()), false, false)) {
		const Vertex& uu(m_graph[ux]);
		if (wpt.get_coord().spheric_distance_nmi_dbl(uu.get_coord()) <= 0.1)
			return ux;
	}
	if (!m_graphbbox.is_inside(wpt.get_coord()))
		return invalid_vertex_descriptor;
	{
		std::ostringstream oss;
		oss << (wpt.is_vor() ? "VOR" : (wpt.is_ndb() ? "NDB" : "Intersection")) << ' ' << wpt.get_ident() << ' '
		    << wpt.get_coord().get_lat_str2() << ' ' << wpt.get_coord().get_lon_str2() << " not found in airway graph";
		m_messages.push_back(Message(oss.str(), Message::type_warning));
	}
	return invalid_vertex_descriptor;
}

TrafficFlowRestrictions::TrafficFlowRestrictions(void)
	: m_airportdb(0), m_navaiddb(0), m_waypointdb(0), m_airwaydb(0), m_airspacedb(0), m_typeofflight('G'), m_shortcircuit(false)
{
}

bool TrafficFlowRestrictions::tracerules_exists(const Glib::ustring& ruleid) const
{
	return m_tracerules.find(ruleid) != m_tracerules.end();
}

bool TrafficFlowRestrictions::tracerules_add(const Glib::ustring& ruleid)
{
	if (ruleid.empty())
		return false;
	std::pair<tracerules_t::iterator,bool> i(m_tracerules.insert(ruleid));
	return i.second;
}

bool TrafficFlowRestrictions::tracerules_erase(const Glib::ustring& ruleid)
{
	tracerules_t::iterator i(m_tracerules.find(ruleid));
	if (i == m_tracerules.end())
		return false;
	m_tracerules.erase(i);
	return true;
}

void TrafficFlowRestrictions::reset_rules(void)
{
	m_tfr = m_loadedtfr;
}

void TrafficFlowRestrictions::prepare_dct_rules(void)
{
	m_dcttfr.clear();
	for (TrafficFlowRules::const_iterator i(m_tfr.begin()), e(m_tfr.end()); i != e; ++i)
		if (i->is_dct())
			m_dcttfr.push_back(*i);
}

unsigned int TrafficFlowRestrictions::count_rules(void) const
{
	return m_tfr.size();
}

unsigned int TrafficFlowRestrictions::count_dct_rules(void) const
{
	return m_dcttfr.size();
}

void TrafficFlowRestrictions::simplify_rules(void)
{
	m_tfr.simplify();
}

void TrafficFlowRestrictions::simplify_rules_bbox(const Rect& bbox)
{
	m_tfr.simplify_bbox(bbox);
}

void TrafficFlowRestrictions::simplify_rules_altrange(int32_t minalt, int32_t maxalt)
{
	m_tfr.simplify_altrange(minalt, maxalt);
}

void TrafficFlowRestrictions::simplify_rules_aircrafttype(const Glib::ustring& acfttype)
{
	if (acfttype.empty())
		return;
	Glib::ustring actype(acfttype);
	actype.resize(4, ' ');
	m_tfr.simplify_aircrafttype(actype);
	simplify_rules_aircraftclass(Aircraft::get_aircrafttypeclass(actype));
}

void TrafficFlowRestrictions::simplify_rules_aircraftclass(const Glib::ustring& acftclass)
{
	if (acftclass.empty())
		return;
	Glib::ustring acclass(acftclass);
	acclass.resize(3, '?');
	m_tfr.simplify_aircraftclass(acclass);
}

void TrafficFlowRestrictions::simplify_rules_equipment(const Glib::ustring& equipment)
{
	m_tfr.simplify_equipment(equipment);
}

void TrafficFlowRestrictions::simplify_rules_typeofflight(char type_of_flight)
{
	m_tfr.simplify_typeofflight(type_of_flight);
}

void TrafficFlowRestrictions::simplify_rules_mil(bool mil)
{
	m_tfr.simplify_mil(mil);
}

void TrafficFlowRestrictions::simplify_rules_gat(bool gat)
{
	m_tfr.simplify_gat(gat);
}

void TrafficFlowRestrictions::simplify_rules_dep(const Glib::ustring& ident, const Point& coord)
{
	m_tfr.simplify_dep(ident, coord);
}

void TrafficFlowRestrictions::simplify_rules_dest(const Glib::ustring& ident, const Point& coord)
{
	m_tfr.simplify_dest(ident, coord);
}

bool TrafficFlowRestrictions::check_dct(std::vector<Message>& msg, DctParameters& dct, bool trace)
{
	m_messages.clear();
	msg.clear();
	DctParameters::AltRange a(m_dcttfr.evaluate_dct(*this, dct, trace));
	dct.get_alt() = a;
	msg.swap(m_messages);
	return !a.is_empty();
}

bool TrafficFlowRestrictions::load(const Glib::ustring& fname)
{
	m_dom.parse_file(fname);
	return m_dom;
}

bool TrafficFlowRestrictions::get_rulewpt(Graph::vertex_descriptor& u, xmlpp::Element *el, const Glib::ustring& suffix)
{
	RuleWpt wpt;
	return get_rulewpt(u, wpt, el, suffix);
}

bool TrafficFlowRestrictions::get_rulewpt(Graph::vertex_descriptor& u, RuleWpt& wpt, xmlpp::Element *el, const Glib::ustring& suffix)
{
	u = invalid_vertex_descriptor;
	wpt = RuleWpt();
	if (!el)
		return false;
	xmlpp::Node::NodeList nl(el->get_children(""));
	for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
		xmlpp::Element *el2(dynamic_cast<xmlpp::Element *>(*ni));
		if (!el2)
			continue;
		Vertex::typemask_t tmask(Vertex::typemask_none);
		Vertex::type_t t(Vertex::type_undefined);
		bool vor(false), ndb(false);
		if (el2->get_name() == "DpnUid" + suffix) {
			tmask = Vertex::typemask_intersection;
			t = Vertex::type_intersection;
		} else if (el2->get_name() == "VorUid" + suffix) {
			tmask = Vertex::typemask_navaid;
			t = Vertex::type_navaid;
			vor = true;
		} else if (el2->get_name() == "NdbUid" + suffix) {
			tmask = Vertex::typemask_navaid;
			t = Vertex::type_navaid;
			ndb = true;
		} else {
			continue;
		}
		Glib::ustring codeid;
		Point pt;
		bool geolat(false), geolon(false);
		{
			xmlpp::Node::NodeList nl(el2->get_children("codeId"));
			for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
				xmlpp::Element *el3(dynamic_cast<xmlpp::Element *>(*ni));
				if (!el3)
					continue;
				xmlpp::TextNode *txt(el3->get_child_text());
				if (!txt)
					continue;
				if (txt->is_white_space())
					continue;
				codeid = txt->get_content();
				if (!codeid.empty())
					break;
			}
		}
		{
			xmlpp::Node::NodeList nl(el2->get_children("geoLat"));
			for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
				xmlpp::Element *el3(dynamic_cast<xmlpp::Element *>(*ni));
				if (!el3)
					continue;
				xmlpp::TextNode *txt(el3->get_child_text());
				if (!txt)
					continue;
				if (txt->is_white_space())
					continue;
				Point pt1;
				if (!(pt1.set_str(txt->get_content()) & Point::setstr_lat))
					continue;
				pt.set_lat(pt1.get_lat());
				geolat = true;
				break;
			}
		}
		{
			xmlpp::Node::NodeList nl(el2->get_children("geoLong"));
			for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
				xmlpp::Element *el3(dynamic_cast<xmlpp::Element *>(*ni));
				if (!el3)
					continue;
				xmlpp::TextNode *txt(el3->get_child_text());
				if (!txt)
					continue;
				if (txt->is_white_space())
					continue;
				Point pt1;
				if (!(pt1.set_str(txt->get_content()) & Point::setstr_lon))
					continue;
				pt.set_lon(pt1.get_lon());
				geolon = true;
				break;
			}
		}
		if (!geolon || !geolat || codeid.empty())
			continue;
		wpt = RuleWpt(t, codeid, pt, vor, ndb);
		Graph::vertex_descriptor ux;
		if (m_graph.find_intersection(ux, codeid, pt, tmask, false, false)) {
			const Vertex& uu(m_graph[ux]);
			if (pt.spheric_distance_nmi_dbl(uu.get_coord()) <= 0.1) {
				u = ux;
				return true;
			}
		}
		if (!m_graphbbox.is_inside(pt))
			return true;
		{
			std::ostringstream oss;
			oss << (vor ? "VOR" : (ndb ? "NDB" : "Intersection")) << ' ' << codeid
			    << ' ' << pt.get_lat_str2() << ' ' << pt.get_lon_str2() << " not found in airway graph";
			m_messages.push_back(Message(oss.str(), Message::type_warning));
		}
		return true;
	}
	return false;
}

TrafficFlowRestrictions::RuleWpt TrafficFlowRestrictions::get_rulewpt(xmlpp::Element *el, const Glib::ustring& suffix)
{
	if (!el)
		return RuleWpt();
	xmlpp::Node::NodeList nl(el->get_children(""));
	for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
		xmlpp::Element *el2(dynamic_cast<xmlpp::Element *>(*ni));
		if (!el2)
			continue;
		FplWpt::type_t t(Vertex::type_undefined);
		bool vor(false), ndb(false);
		if (el2->get_name() == "DpnUid" + suffix) {
			t = Vertex::type_intersection;
		} else if (el2->get_name() == "VorUid" + suffix) {
			t = Vertex::type_navaid;
			vor = true;
		} else if (el2->get_name() == "NdbUid" + suffix) {
			t = Vertex::type_navaid;
			ndb = true;
		} else {
			continue;
		}
		Glib::ustring codeid;
		Point pt;
		bool geolat(false), geolon(false);
		{
			xmlpp::Node::NodeList nl(el2->get_children("codeId"));
			for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
				xmlpp::Element *el3(dynamic_cast<xmlpp::Element *>(*ni));
				if (!el3)
					continue;
				xmlpp::TextNode *txt(el3->get_child_text());
				if (!txt)
					continue;
				if (txt->is_white_space())
					continue;
				codeid = txt->get_content();
				if (!codeid.empty())
					break;
			}
		}
		{
			xmlpp::Node::NodeList nl(el2->get_children("geoLat"));
			for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
				xmlpp::Element *el3(dynamic_cast<xmlpp::Element *>(*ni));
				if (!el3)
					continue;
				xmlpp::TextNode *txt(el3->get_child_text());
				if (!txt)
					continue;
				if (txt->is_white_space())
					continue;
				Point pt1;
				if (!(pt1.set_str(txt->get_content()) & Point::setstr_lat))
					continue;
				pt.set_lat(pt1.get_lat());
				geolat = true;
				break;
			}
		}
		{
			xmlpp::Node::NodeList nl(el2->get_children("geoLong"));
			for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
				xmlpp::Element *el3(dynamic_cast<xmlpp::Element *>(*ni));
				if (!el3)
					continue;
				xmlpp::TextNode *txt(el3->get_child_text());
				if (!txt)
					continue;
				if (txt->is_white_space())
					continue;
				Point pt1;
				if (!(pt1.set_str(txt->get_content()) & Point::setstr_lon))
					continue;
				pt.set_lon(pt1.get_lon());
				geolon = true;
				break;
			}
		}
		if (geolon && geolat && !codeid.empty())
			return RuleWpt(t, codeid, pt, vor, ndb);
	}
	return RuleWpt();
}

TrafficFlowRestrictions::AltRange TrafficFlowRestrictions::get_rulealtrange(xmlpp::Element *el, const Glib::ustring& nodename)
{
	if (!el)
		return AltRange();
	typedef enum {
		altunit_ft,
		altunit_fl,
		altunit_m
	} altunit_t;
	xmlpp::Node::NodeList nl(el->get_children(nodename));
	for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
		xmlpp::Element *el2(dynamic_cast<xmlpp::Element *>(*ni));
		if (!el2)
			continue;
		unsigned int setflags(0);
		int32_t alt[2];
		altunit_t altunit[2];
		AltRange::alt_t altmode[2];
		{
			xmlpp::Node::NodeList nl2(el2->get_children(""));
			for (xmlpp::Node::NodeList::const_iterator ni(nl2.begin()), ne(nl2.end()); ni != ne; ++ni) {
				xmlpp::Element *el3(dynamic_cast<xmlpp::Element *>(*ni));
				if (!el3)
					continue;
				xmlpp::TextNode *txt(el3->get_child_text());
				if (!txt)
					continue;
				if (txt->is_white_space())
					continue;
				unsigned int idx(el3->get_name() == "valDistVerUpper");
				if (el3->get_name() == "valDistVerLower" || idx) {
					const char *cp(txt->get_content().c_str());
					char *ep;
					alt[idx] = strtol(cp, &ep, 10);
					if (cp == ep || *ep)
						continue;
					setflags |= 0x01 << idx;
					continue;
				}
				idx = el3->get_name() == "uomDistVerUpper";
				if (el3->get_name() == "uomDistVerLower" || idx) {
					if (txt->get_content() == "FT")
						altunit[idx] = altunit_ft;
					else if (txt->get_content() == "FL")
						altunit[idx] = altunit_fl;
					else if (txt->get_content() == "M")
						altunit[idx] = altunit_m;
					else
						continue;
					setflags |= 0x04 << idx;
					continue;
				}
				idx = el3->get_name() == "codeDistVerUpper";
				if (el3->get_name() == "codeDistVerLower" || idx) {
					if (txt->get_content() == "STD")
						altmode[idx] = AltRange::alt_std;
					else if (txt->get_content() == "ALT")
						altmode[idx] = AltRange::alt_qnh;
					else if (txt->get_content() == "HEI")
						altmode[idx] = AltRange::alt_height;
					else
						continue;
					setflags |= 0x10 << idx;
					continue;
				}
			}
		}
		if (0x3f & ~setflags)
			continue;
		for (unsigned int i  = 0; i < 2; ++i) {
			switch (altunit[i]) {
			case altunit_fl:
				alt[i] *= 100;
				break;

			case altunit_m:
				alt[i] = Point::round<int32_t,double>(alt[i] * Point::m_to_ft_dbl);
				break;

			default:
				break;
			}
		}
		return AltRange(alt[0], altmode[0], alt[1], altmode[1]);
	}
	return AltRange();
}

const AirspacesDb::Airspace& TrafficFlowRestrictions::get_ruleairspace(xmlpp::Element *el, const Glib::ustring& nodename)
{
	static const AirspacesDb::Airspace invalid_airspace;
	if (!el)
		return invalid_airspace;
	xmlpp::Node::NodeList nl(el->get_children(nodename));
	for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
		xmlpp::Element *el2(dynamic_cast<xmlpp::Element *>(*ni));
		if (!el2)
			continue;
		Glib::ustring icao;
		char bdryclass(0);
		xmlpp::Node::NodeList nl2(el2->get_children(""));
		for (xmlpp::Node::NodeList::const_iterator ni(nl2.begin()), ne(nl2.end()); ni != ne; ++ni) {
			xmlpp::Element *el3(dynamic_cast<xmlpp::Element *>(*ni));
			if (!el3)
				continue;
			xmlpp::TextNode *txt(el3->get_child_text());
			if (!txt)
				continue;
			if (txt->is_white_space())
				continue;
			if (el3->get_name() == "codeId") {
				icao = txt->get_content();
				continue;
			}
			if (el3->get_name() == "codeType") {
				Glib::ustring t(txt->get_content());
				if (t == "D")
					t = "D-EAD";
				uint8_t tc(0);
				char bc(0);
				if (!AirspacesDb::Airspace::set_class_string(tc, bc, t))
					continue;
				if (tc != AirspacesDb::Airspace::typecode_ead)
					continue;
				bdryclass = bc;
				continue;
			}
		}
		if (!bdryclass || icao.empty())
			continue;
		{
			AirspacesDb::Airspace aspc;
			aspc.set_icao(icao);
			aspc.set_typecode(AirspacesDb::Airspace::typecode_ead);
			aspc.set_bdryclass(bdryclass);
			airspacecache_t::const_iterator i(m_airspacecache.find(aspc));
			if (i != m_airspacecache.end())
				return *i;
		}
		{
			AirspacesDb::elementvector_t ev(m_airspacedb->find_by_icao(icao, 0, AirspacesDb::comp_exact, 0, AirspacesDb::Airspace::subtables_none));
			for (AirspacesDb::elementvector_t::iterator i(ev.begin()), e(ev.end()); i != e; ++i) {
				if (i->get_typecode() != AirspacesDb::Airspace::typecode_ead ||
				    i->get_bdryclass() != bdryclass || !i->is_valid())
					continue;
				std::pair<airspacecache_t::iterator,bool> ins(m_airspacecache.insert(*i));
				return *ins.first;
			}
		}
	}
	return invalid_airspace;
}

bool TrafficFlowRestrictions::get_ruledct(RuleDct& dct, xmlpp::Element *el)
{
	dct = RuleDct();
	if (!el)
		return false;
	xmlpp::Node::NodeList nl(el->get_children("Dct"));
	for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
		xmlpp::Element *el2(dynamic_cast<xmlpp::Element *>(*ni));
		if (!el2)
			continue;
		Graph::vertex_descriptor u, v;
		RuleWpt w0, w1;
		if (!get_rulewpt(u, w0, el2, "Sta"))
			continue;
		if (!get_rulewpt(v, w1, el2, "End"))
			continue;
		dct = RuleDct(u, v, w0, w1);
		return true;
	}
	return false;
}

bool TrafficFlowRestrictions::get_ruleawy(RuleRoute& rte, xmlpp::Element *el)
{
	rte = RuleRoute();
	if (!el)
		return false;
	xmlpp::Node::NodeList nl(el->get_children("Rpn"));
	for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
		xmlpp::Element *el2(dynamic_cast<xmlpp::Element *>(*ni));
		if (!el2)
			continue;
		Graph::vertex_descriptor u, v;
		RuleWpt w0, w1;
		if (!get_rulewpt(u, w0, el2, "Sta"))
			continue;
		if (!get_rulewpt(v, w1, el2, "End"))
			continue;
		Glib::ustring awy;
		xmlpp::Node::NodeList nl2(el2->get_children("RteUid"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl2.begin()), ne(nl2.end()); ni != ne; ++ni) {
			xmlpp::Element *el3(dynamic_cast<xmlpp::Element *>(*ni));
			if (!el3)
				continue;
			xmlpp::Node::NodeList nl3(el3->get_children("txtDesig"));
			for (xmlpp::Node::NodeList::const_iterator ni(nl3.begin()), ne(nl3.end()); ni != ne; ++ni) {
				xmlpp::Element *el4(dynamic_cast<xmlpp::Element *>(*ni));
				if (!el4)
					continue;
				xmlpp::TextNode *txt(el4->get_child_text());
				if (!txt)
					continue;
				if (txt->is_white_space())
					continue;
				awy = txt->get_content();
			}
		}
		if (awy.empty())
			continue;
		rte = RuleRoute(u, v, w0, w1, awy);
		return true;
	}
	return false;
}

TrafficFlowRestrictions::Result TrafficFlowRestrictions::check_fplan(const FPlanRoute& route, char type_of_flight,
								     const std::string& acft_type, const std::string& equipment,
								     AirportsDb& airportdb, NavaidsDb& navaiddb, WaypointsDb& waypointdb,
								     AirwaysDb& airwaydb, AirspacesDb& airspacedb)
{
	m_rules.clear();
	m_messages.clear();
	m_airportdb = &airportdb;
	m_navaiddb = &navaiddb;
	m_waypointdb = &waypointdb;
	m_airwaydb = &airwaydb;
	m_airspacedb = &airspacedb;
	Result res(acft_type, equipment, type_of_flight);
	m_aircrafttype = res.get_aircrafttype();
	m_aircrafttypeclass = res.get_aircrafttypeclass();
	m_equipment = res.get_equipment();
	m_typeofflight = res.get_typeofflight();
	res.m_result = (&route) && m_airportdb && m_navaiddb &&
		m_waypointdb && m_airwaydb && m_airspacedb;
	if (res.m_result) {
		m_route = route;
		load_airway_graph();
		res.m_result = check_fplan_depdest() &&
			check_fplan_rules();
	}
	res.m_fplan.swap(m_route);
	res.m_rules.swap(m_rules);
	res.m_messages.swap(m_messages);
	m_route.clear();
	m_rules.clear();
	m_messages.clear();
	m_aircrafttype.clear();
	m_aircrafttypeclass.clear();
	m_equipment.clear();
	m_typeofflight = 'G';
	m_airportdb = 0;
	m_navaiddb = 0;
	m_waypointdb = 0;
	m_airwaydb = 0;
	m_airspacedb = 0;
	return res;
}

bool TrafficFlowRestrictions::check_fplan_depdest(void)
{
	if (m_route.size() < 2) {
		m_messages.push_back(Message("Route has less than 2 waypoints", Message::type_error));
		return false;
	}
	bool ret(true);
	// find waypoint types
	for (unsigned int wptnr = 0; wptnr < m_route.size(); ++wptnr) {
		FplWpt& wpt(m_route[wptnr]);
		wpt.set_wptnr(wptnr);
		{
			Glib::ustring tt(wpt.get_icao());
			Vertex::typemask_t tmask(Vertex::typemask_navaid | Vertex::typemask_intersection);
			if (!wptnr || wptnr + 1U >= m_route.size()) {
				tmask |= Vertex::typemask_airport;
				if (tt == "ZZZZ")
					tt.clear();
			}
			if (tt.empty())
				tt = wpt.get_name();
			Graph::vertex_descriptor u;
			bool ok(false);
			if (!tt.empty())
				ok = m_graph.find_intersection(u, tt, wpt.get_coord(), tmask, false, false);
			if (ok) {
				const Vertex& uu(m_graph[u]);
				double dist(wpt.get_coord().spheric_distance_nmi_dbl(uu.get_coord()));
				ok = dist <= 0.1;
				if (ok) {
					//wpt.set_coord(uu.get_coord());
					wpt.set_vertex_descriptor(u);
					wpt.set_type(uu.get_type());
					if (uu.get_type() == Vertex::type_navaid) {
						NavaidsDb::Navaid el;
						if (uu.get(el) && el.is_valid())
							wpt.set_navaid_type(el.get_navaid_type());
					}
					continue;
				}
				if (wptnr && (wptnr + 1U < m_route.size()) && !wpt.is_ifr() && !m_route[wptnr - 1U].is_ifr())
					continue;
				std::ostringstream oss;
				oss << "Waypoint " << wpt.get_icao() << '/' << wpt.get_name() << " ("
				    << wpt.get_coord().get_lat_str2() << ' ' << wpt.get_coord().get_lon_str2() << ") not found (distance"
				    << dist << "nmi)";
				Message msg(oss.str(), Message::type_error);
				msg.add_vertex(wptnr);
				m_messages.push_back(msg);
				ret = false;
				continue;
			}
		}
		if (wptnr && (wptnr + 1U < m_route.size()) && !wpt.is_ifr() && !m_route[wptnr - 1U].is_ifr())
			continue;
		std::ostringstream oss;
		oss << "Waypoint " << wpt.get_icao() << '/' << wpt.get_name() << " ("
		    << wpt.get_coord().get_lat_str2() << ' ' << wpt.get_coord().get_lon_str2() << ") not found";
		Message msg(oss.str(), Message::type_error);
		msg.add_vertex(wptnr);
		m_messages.push_back(msg);
		ret = false;
	}
	if (m_route.front().get_type() != Vertex::type_airport) {
		Message msg("First Waypoint not an Airport", Message::type_error);
		msg.add_vertex(0U);
		m_messages.push_back(msg);
		ret = false;
	}
	if (m_route.back().get_type() != Vertex::type_airport) {
		Message msg("Last Waypoint not an Airport", Message::type_error);
		msg.add_vertex(m_route.size() - 1U);
		m_messages.push_back(msg);
		ret = false;
	}
	if (!ret)
		return false;
	// Check and Expand Airways
	for (unsigned int wptnr = 0; wptnr < m_route.size(); ++wptnr) {
		FplWpt& wpt(m_route[wptnr]);
		wpt.set_pathname(wpt.get_pathname().uppercase());
		if (wptnr + 1U >= m_route.size()) {
			wpt.set_pathcode(FPlanWaypoint::pathcode_none);
			wpt.set_pathname("");
			continue;
		}
		FplWpt& wptn(m_route[wptnr + 1U]);
		if (!wpt.is_ifr()) {
			switch (wpt.get_pathcode()) {
			case FPlanWaypoint::pathcode_vfrdeparture:
                        case FPlanWaypoint::pathcode_vfrarrival:
                        case FPlanWaypoint::pathcode_vfrtransition:
				break;

			default:
				wpt.set_pathcode(FPlanWaypoint::pathcode_none);
				break;
			}
			continue;
		}
		switch (wpt.get_pathcode()) {
		case FPlanWaypoint::pathcode_sid:
			if (wptnr) {
				wpt.set_pathcode(FPlanWaypoint::pathcode_none);
				wpt.set_pathname("");
			}
			break;

		case FPlanWaypoint::pathcode_star:
			if (wptnr + 2U != m_route.size()) {
				wpt.set_pathcode(FPlanWaypoint::pathcode_none);
				wpt.set_pathname("");
			}
			break;

		case FPlanWaypoint::pathcode_directto:
			wpt.set_pathname("");
			break;

		default:
			wpt.set_pathcode(FPlanWaypoint::pathcode_none);
			wpt.set_pathname("");
			break;

		case FPlanWaypoint::pathcode_airway:
			if (wpt.get_pathname().empty()) {
				wpt.set_pathcode(FPlanWaypoint::pathcode_none);
				break;
			}
			// expand airway
			if (wpt.get_vertex_descriptor() == invalid_vertex_descriptor ||
			    wptn.get_vertex_descriptor() == invalid_vertex_descriptor ||
			    wpt.get_vertex_descriptor() == wptn.get_vertex_descriptor()) {
				std::ostringstream oss;
				oss << "Invalid Airway " << wpt.get_pathname() << " Segment "
				    << wpt.get_ident() << "->" << wptn.get_ident();
				Message msg(oss.str(), Message::type_error);
				msg.add_edge(wptnr);
				m_messages.push_back(msg);
				ret = false;
				wpt.set_pathcode(FPlanWaypoint::pathcode_directto);
				wpt.set_pathname("");
				break;
			}
			{
				std::vector<Graph::vertex_descriptor> predecessors;
				m_graph.shortest_paths(wpt.get_vertex_descriptor(), wpt.get_pathname(), predecessors);
				Graph::vertex_descriptor v(wptn.get_vertex_descriptor());
				if (v == predecessors[v]) {
					std::ostringstream oss;
					oss << "Airway " << wpt.get_pathname() << ": No Segment "
					    << wpt.get_ident() << "->" << wptn.get_ident();
					Message msg(oss.str(), Message::type_error);
					msg.add_edge(wptnr);
					m_messages.push_back(msg);
					ret = false;
					wpt.set_pathcode(FPlanWaypoint::pathcode_directto);
					wpt.set_pathname("");
					break;
				}
				unsigned int wptnrx(wptnr);
				for (;;) {
					FplWpt& wpt(m_route[wptnr]);
					v = predecessors[v];
					if (v == wpt.get_vertex_descriptor())
						break;
					const Vertex& vv(m_graph[v]);
					if (false && vv.is_ident_numeric())
						continue;
					if (true && !vv.is_ifr_fpl())
						continue;
					FplWpt wptx(wpt);
					if (!vv.set(wptx)) {
						std::ostringstream oss;
						oss << "Cannot insert Waypoint " << vv.get_ident();
						Message msg(oss.str(), Message::type_error);
						msg.add_vertex(wptnr);
						m_messages.push_back(msg);
						continue;
					}
					wptx.set_type(vv.get_type());
					if (vv.get_type() == Vertex::type_navaid) {
						NavaidsDb::Navaid el;
						if (vv.get(el) && el.is_valid())
							wptx.set_navaid_type(el.get_navaid_type());
					}
					wptx.set_pathcode(FPlanWaypoint::pathcode_airway);
					wptx.set_pathname(wpt.get_pathname());
					m_route.insert(m_route.begin() + (wptnr + 1U), wptx);
					++wptnrx;
					for (messages_t::iterator mi(m_messages.begin()), me(m_messages.end()); mi != me; ++mi)
						mi->insert_vertex(wptnr + 1U);
				}
				wptnr = wptnrx;
			}
			break;
		}
	}
	return ret;
}

bool TrafficFlowRestrictions::check_fplan_rules(void)
{
	// new style evaluation if loaded
	if (!m_tfr.empty())
		return m_tfr.evaluate(*this, m_rules);
	bool ret(true);
	unsigned int nrrules(0);
	for (Rule rule(get_first_rule()); rule.is_valid(); rule.next_rule()) {
		++nrrules;
		CondResult cond(compute_condition(rule, rule.get_condition()));
		if (cond.get_result() == boost::logic::indeterminate) {
			m_messages.push_back(Message("Cannot evaluate condition", Message::type_warning, rule.get_codeid()));
			continue;
		}
		if (cond.get_result() != true)
			continue;
		RuleResult ruleresult(rule.get_codeid(), rule.get_descr(), rule.get_oprgoal(),
				      (RuleResult::codetype_t)rule.get_codetype(), cond.get_vertexset(), cond.get_edgeset());
		xmlpp::Node::NodeList nl(rule.get_element()->get_children("Tre"));
		if (rule.get_codetype() == Rule::codetype_closed) {
			if (!nl.empty()) 
				m_messages.push_back(Message("Rule closed for cruising has Tre elements", Message::type_warning, rule.get_codeid()));
			m_rules.push_back(ruleresult);
			ret = false;
			continue;
		}
		bool ruleok(rule.get_codetype() == Rule::codetype_forbidden);
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
			xmlpp::Element *tre(dynamic_cast<xmlpp::Element *>(*ni));
			if (!tre)
				continue;
			std::vector<RuleResult::Alternative::Sequence> seq;
			CondResult r(true);
			bool first(true);
			xmlpp::Node::NodeList nl(tre->get_children("Tfe"));
			for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
				xmlpp::Element *tfe(dynamic_cast<xmlpp::Element *>(*ni));
				if (!tfe)
					continue;
				RuleResult::Alternative::Sequence seq1;
				if (first) {
					r = compute_tfe(rule, tfe, seq1);
					first = false;
				} else {
					CondResult r1(compute_tfe(rule, tfe, seq1));
					if (r.get_last() > r1.get_first())
						r = CondResult(false);
					r &= r1;
				}
				if (seq1.get_type() != RuleResult::Alternative::Sequence::type_invalid)
					seq.push_back(seq1);
			}
			{
				RuleResult::Alternative alt(r.get_vertexset(), r.get_edgeset());
				alt.m_seq.swap(seq);
				ruleresult.add_element(alt);
			}
			if (rule.get_codetype() == Rule::codetype_forbidden)
				ruleok = ruleok && !r.get_result();
			else
				ruleok = ruleok || r.get_result();
		}
		if (!ruleok) {
			m_rules.push_back(ruleresult);
			ret = false;
		}
	}
	if (true) {
		std::ostringstream oss;
		oss << "Rules processed: " << nrrules << " Result: " << (ret ? "OK" : "FAIL");
		m_messages.push_back(Message(oss.str(), Message::type_info));
	}
	return ret;
}

void TrafficFlowRestrictions::load_airway_graph(void)
{
        Rect bbox(m_route.get_bbox());
	Rect bboxov(bbox.oversize_nmi(100.f));
	if (boost::num_vertices(m_graph) && m_graphbbox.is_inside(bboxov))
		return;
	// trash airspace cache as well
	m_airspacecache.clear();
	m_graph.clear();
	m_graphbbox = bboxov.oversize_nmi(50.f);
	{
                NavaidsDb::elementvector_t ev(m_navaiddb->find_by_rect(m_graphbbox, ~0, NavaidsDb::Navaid::subtables_none));
                for (NavaidsDb::elementvector_t::iterator i(ev.begin()); i != ev.end();) {
                        switch (i->get_navaid_type()) {
                        case NavaidsDb::Navaid::navaid_vor:
                        case NavaidsDb::Navaid::navaid_vordme:
                        case NavaidsDb::Navaid::navaid_ndb:
                        case NavaidsDb::Navaid::navaid_ndbdme:
                        case NavaidsDb::Navaid::navaid_vortac:
                                ++i;
                                break;

                        default:
                                i = ev.erase(i);
                                break;
                        }
                }
                m_graph.add(ev);
        }
        {
                WaypointsDb::elementvector_t ev(m_waypointdb->find_by_rect(m_graphbbox, ~0, WaypointsDb::Waypoint::subtables_none));
                for (WaypointsDb::elementvector_t::iterator i(ev.begin()); i != ev.end();) {
                        switch (i->get_type()) {
                        case WaypointsDb::Waypoint::type_icao:
                                ++i;
                                break;

                        default:
                                i = ev.erase(i);
                                break;
                        }
                }
                m_graph.add(ev);
        }
        m_graph.add(m_airwaydb->find_by_area(m_graphbbox, ~0, AirwaysDb::Airway::subtables_all));
        m_graph.add(m_airportdb->find_by_rect(m_graphbbox, ~0, NavaidsDb::Navaid::subtables_none));
        if (true) {
                unsigned int counts[4];
                for (unsigned int i = 0; i < sizeof(counts)/sizeof(counts[0]); ++i)
                        counts[i] = 0;
                Graph::vertex_iterator u0, u1;
                for (boost::tie(u0, u1) = boost::vertices(m_graph); u0 != u1; ++u0) {
                        const Vertex& uu(m_graph[*u0]);
                        unsigned int i(uu.get_type());
                        if (i < sizeof(counts)/sizeof(counts[0]))
                                ++counts[i];
                }
                {
                        std::ostringstream oss;
			oss << "Airway area graph: [" << m_graphbbox.get_southwest().get_lat_str2() << ' ' << m_graphbbox.get_southwest().get_lon_str2()
                            << ' ' << m_graphbbox.get_northeast().get_lat_str2() << ' ' << m_graphbbox.get_northeast().get_lon_str2() << "], "
                            << boost::num_vertices(m_graph) << " vertices, " << boost::num_edges(m_graph) << " edges, "
			    << counts[Vertex::type_airport] << " airports, " << counts[Vertex::type_navaid] << " navaids, "
                            << counts[Vertex::type_intersection] << " intersections, " << counts[Vertex::type_mapelement] << " mapelements, route bbox ["
			    << bbox.get_southwest().get_lat_str2() << ' ' << bbox.get_southwest().get_lon_str2()
			    << ' ' << bbox.get_northeast().get_lat_str2() << ' ' << bbox.get_northeast().get_lon_str2() << ']';
			m_messages.push_back(Message(oss.str(), Message::type_info));
                }
	}
}

TrafficFlowRestrictions::CondResult TrafficFlowRestrictions::compute_condition(Rule& rule, xmlpp::Element *cond)
{
	CondResult r(compute_condition_1(rule, cond));
	if (tracerules_exists(rule.get_codeid()) && cond) {
		std::ostringstream oss;
		oss << "Condition " << cond->get_path() << " result: " << (r.get_result() ? 'T' : (!r.get_result() ? 'F' : 'I'));
		Message msg(oss.str(), Message::type_tracecond, rule.get_codeid());
		msg.m_vertexset = r.get_vertexset();
		msg.m_edgeset = r.get_edgeset();
		m_messages.push_back(msg);
	}
	return r;
}

TrafficFlowRestrictions::CondResult TrafficFlowRestrictions::compute_condition_1(Rule& rule, xmlpp::Element *cond)
{
	if (!cond)
		return CondResult();
	if (cond->get_name() == "Fcc") {
		Glib::ustring codeopr(cond->get_attribute_value("codeOpr"));
		if (codeopr == "OR") {
			CondResult r(false);
			xmlpp::Node::NodeList nl(cond->get_children(""));
			for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
				xmlpp::Element *el(dynamic_cast<xmlpp::Element *>(*ni));
				if (!el)
					continue;
				r |= compute_condition(rule, el);
				// short circuit evaluation
				if (m_shortcircuit && r.get_result() == true)
					return r;
			}
			return r;
		}
		if (codeopr == "AND") {
			CondResult r(true);
			xmlpp::Node::NodeList nl(cond->get_children(""));
			for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
				xmlpp::Element *el(dynamic_cast<xmlpp::Element *>(*ni));
				if (!el)
					continue;
			        r &= compute_condition(rule, el);
				// short circuit evaluation
				if (r.get_result() == false)
					return r;
			}
			return r;
		}
		if (codeopr == "SEQ") {
			CondResult r(true);
			bool first(true);
			xmlpp::Node::NodeList nl(cond->get_children(""));
			for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
				xmlpp::Element *el(dynamic_cast<xmlpp::Element *>(*ni));
				if (!el)
					continue;
				if (first) {
					r = compute_condition(rule, el);
					first = false;
				} else {
					CondResult r1(compute_condition(rule, el));
					if (r.get_last() > r1.get_first())
						return CondResult(false);
					r &= r1;
				}
				// short circuit evaluation
				if (r.get_result() == false)
					return r;
			}
			return r;
		}
		if (codeopr == "ANDNOT") {
			CondResult r(true);
			bool first(true);
			xmlpp::Node::NodeList nl(cond->get_children(""));
			for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
				xmlpp::Element *el(dynamic_cast<xmlpp::Element *>(*ni));
				if (!el)
					continue;
				if (first) {
					r = compute_condition(rule, el);
					first = false;
				} else {
					r &= ~compute_condition(rule, el);
				}
				// short circuit evaluation
				if (r.get_result() == false)
					return r;
			}
			return r;
		}
		if (codeopr == "NONE") {
			CondResult r(true);
			bool first(true);
			xmlpp::Node::NodeList nl(cond->get_children(""));
			for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
				xmlpp::Element *el(dynamic_cast<xmlpp::Element *>(*ni));
				if (!el)
					continue;
				if (first) {
					r = compute_condition(rule, el);
					first = false;
				} else {
					m_messages.push_back(Message("Rule has multiple conditions with codeOpr NONE", Message::type_warning, rule.get_codeid()));
				}
			}
			return r;
		}
		m_messages.push_back(Message("Rule has unknown codeOpr " + codeopr, Message::type_warning, rule.get_codeid()));
		return CondResult();
	}
	if (cond->get_name() == "Fce") {
		Glib::ustring coderel(cond->get_attribute_value("codeRelWithLoc"));
		{
			bool is_arr(coderel == "ARR");
			if (is_arr || coderel == "DEP") {
				const FplWpt& wpt(is_arr ? m_route.back() : m_route.front());
				xmlpp::Node::NodeList nl(cond->get_children("AhpUid"));
				for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
					xmlpp::Element *el(dynamic_cast<xmlpp::Element *>(*ni));
					if (!el)
						continue;
					xmlpp::Node::NodeList nl2(el->get_children("codeId"));
					for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
						xmlpp::Element *el2(dynamic_cast<xmlpp::Element *>(*ni2));
						if (!el2)
							continue;
						const xmlpp::TextNode *tn(el2->get_child_text());
						if (!tn)
							continue;
						if (tn->get_content() == wpt.get_icao()) {
							CondResult r(true);
							r.get_vertexset().insert(is_arr ? m_route.size() - 1U : 0U);
							return r;
						}
					}
				}
				return CondResult(false);
			}
		}
		if (coderel == "ACT") {
			// Airspace Activation
			const AirspacesDb::Airspace& aspc(get_ruleairspace(cond));
			if (!aspc.is_valid())
				m_messages.push_back(Message("Rule has unknown airspace " + cond->get_path(), Message::type_warning, rule.get_codeid()));
			// We assume that airspaces are always active at the moment (FIXME)
			return CondResult(true);
		}
		if (coderel == "XNG") {
			AltRange altrange(get_rulealtrange(cond, "Fcl"));
			// Airspace Crossing
			{
				xmlpp::Node::NodeList nl(cond->get_children("AseUid"));
				if (!nl.empty()) {
					const AirspacesDb::Airspace& aspc(get_ruleairspace(cond));
					if (!aspc.is_valid()) {
						m_messages.push_back(Message("Rule Fce element has invalid airspace " + cond->get_path(), Message::type_warning, rule.get_codeid()));
						return CondResult();
					}
					const MultiPolygonHole& poly(aspc.get_polygon());
					Rect bbox(poly.get_bbox());
					// Rect bbox(aspc.get_bbox());
					for (unsigned int wptnr = 0U; wptnr < m_route.size(); ++wptnr) {
						const FplWpt& wpt(m_route[wptnr]);
						if (!wpt.is_ifr() && (!wptnr || !m_route[wptnr - 1U].is_ifr()))
							continue;
						const Point& pt(wpt.get_coord());
						if (!wpt.within(altrange) || !bbox.is_inside(pt) || !poly.windingnumber(pt))
							continue;
						CondResult r(true);
						r.get_vertexset().insert(wptnr);
						return r;
					}
					return CondResult(false);
				}
			}
			// From-To Airspace Crossing
			{
				xmlpp::Node::NodeList nl(cond->get_children("Abc"));
				if (!nl.empty()) {
					for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
						xmlpp::Element *el(dynamic_cast<xmlpp::Element *>(*ni));
						if (!el)
							continue;
						const AirspacesDb::Airspace& aspcfrom(get_ruleairspace(el, "AseUidFrom"));
						const AirspacesDb::Airspace& aspcinto(get_ruleairspace(el, "AseUidInto"));
						if (!aspcfrom.is_valid() || !aspcinto.is_valid()) {
							m_messages.push_back(Message("Rule Fce element has invalid airspace(s) " + cond->get_path(), Message::type_warning, rule.get_codeid()));
							return CondResult();
						}
						const MultiPolygonHole& polyfrom(aspcfrom.get_polygon());
						const MultiPolygonHole& polyinto(aspcinto.get_polygon());
						Rect bboxfrom(polyfrom.get_bbox());
						Rect bboxinto(polyinto.get_bbox());
						// Rect bboxfrom(aspcfrom.get_bbox());
						// Rect bboxinto(aspcinto.get_bbox());
						for (unsigned int wptnr = 1U; wptnr < m_route.size(); ++wptnr) {
							const FplWpt& wpt0(m_route[wptnr - 1U]);
							const FplWpt& wpt1(m_route[wptnr]);
							if (!wpt0.is_ifr())
								continue;
							const Point& pt0(wpt0.get_coord());
							const Point& pt1(wpt1.get_coord());
							if (!wpt0.within(altrange) || !bboxfrom.is_inside(pt0) || !bboxinto.is_inside(pt1) ||
							    !polyfrom.windingnumber(pt0) || !polyinto.windingnumber(pt1))
								continue;
							CondResult r(true);
							r.get_edgeset().insert(wptnr - 1U);
							return r;
						}
					}
					return CondResult(false);
				}
			}
			// Direct
			{
				xmlpp::Node::NodeList nl(cond->get_children("Dct"));
				if (!nl.empty()) {
					RuleDct dct;
					if (!get_ruledct(dct, cond)) {
						m_messages.push_back(Message("Rule Fce element has invalid direct-to " + cond->get_path(), Message::type_warning, rule.get_codeid()));
						return CondResult();
					}
					if (true && tracerules_exists(rule.get_codeid())) {
						std::ostringstream oss;
						oss << "DCT Cond: " << dct.get_fromwpt().get_ident() << " (" << dct.get_from()
						    << ") " << dct.get_towpt().get_ident() << " (" << dct.get_to() << ')';
						m_messages.push_back(Message(oss.str(), Message::type_info, rule.get_codeid()));
					}
					if (dct.get_from() == invalid_vertex_descriptor ||
					    dct.get_to() == invalid_vertex_descriptor)
						return CondResult(false);
					for (unsigned int wptnr = 1U; wptnr < m_route.size(); ++wptnr) {
						const FplWpt& wpt0(m_route[wptnr - 1U]);
						const FplWpt& wpt1(m_route[wptnr]);
						if (!wpt0.is_ifr() || !wpt0.within(altrange) || wpt0.get_pathcode() != FPlanWaypoint::pathcode_directto)
							continue;
						if (wpt0.get_vertex_descriptor() == dct.get_from() &&
						    wpt1.get_vertex_descriptor() == dct.get_to()) {
							CondResult r(true);
							r.get_edgeset().insert(wptnr - 1U);
							return r;
						}
					}
					return CondResult(false);
				}
			}
			// Airway
			{
				xmlpp::Node::NodeList nl(cond->get_children("Rpn"));
				if (!nl.empty()) {
					RuleRoute rte;
					if (!get_ruleawy(rte, cond)) {
						m_messages.push_back(Message("Rule Fce element has invalid airway " + cond->get_path(), Message::type_warning, rule.get_codeid()));
						return CondResult();
					}
					if (true && tracerules_exists(rule.get_codeid())) {
						std::ostringstream oss;
						oss << "Airway Cond: " << rte.get_airway()
						    << ' ' << rte.get_fromwpt().get_ident() << " (" << rte.get_from()
						    << ") " << rte.get_towpt().get_ident() << " (" << rte.get_to() << ')';
						m_messages.push_back(Message(oss.str(), Message::type_info, rule.get_codeid()));
					}
					if (rte.get_from() == invalid_vertex_descriptor ||
					    rte.get_to() == invalid_vertex_descriptor)
						return CondResult(false);
					for (unsigned int wptnr = 1U; wptnr < m_route.size(); ++wptnr) {
						const FplWpt& wpt(m_route[wptnr - 1U]);
						if (wpt.get_vertex_descriptor() != rte.get_from() || !wpt.within(altrange))
							continue;
						CondResult r(true);
						for (unsigned int wptnre(wptnr); wptnre < m_route.size(); ++wptnre) {
							const FplWpt& wpt(m_route[wptnre - 1U]);
							const FplWpt& wpte(m_route[wptnre]);
							if (!wpt.is_ifr() ||
							    wpt.get_pathcode() != FPlanWaypoint::pathcode_airway ||
							    wpt.get_pathname() != rte.get_airway()) {
								r.set_result(false);
								break;
							}
							r.get_edgeset().insert(wptnre - 1U);
							if (wpte.get_vertex_descriptor() == rte.get_to())
								break;
						}
						if (r.get_result())
							return r;
					}
					return CondResult(false);
				}
			}
			// Single Point
			{
				Graph::vertex_descriptor u;
				if (get_rulewpt(u, cond, "Spn")) {
					for (unsigned int wptnr = 0U; wptnr < m_route.size(); ++wptnr) {
						const FplWpt& wpt(m_route[wptnr]);
						if (!wpt.is_ifr() && (!wptnr || !m_route[wptnr - 1U].is_ifr()))
							continue;
						if (wpt.get_vertex_descriptor() != u || !wpt.within(altrange))
							continue;
						CondResult r(true);
						r.get_vertexset().insert(wptnr);
						return r;
					}
					return CondResult(false);
				}
			}
			m_messages.push_back(Message("Rule Fce XNG element has unknown structure " + cond->get_path(), Message::type_warning, rule.get_codeid()));
			return CondResult();
		}
		if (coderel.empty()) {
			xmlpp::Node::NodeList nl(cond->get_children("Acs"));
			for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
				xmlpp::Element *el(dynamic_cast<xmlpp::Element *>(*ni));
				if (!el)
					continue;
				bool match(false);
				xmlpp::Node::NodeList nl2(el->get_children(""));
				for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
					xmlpp::Element *el2(dynamic_cast<xmlpp::Element *>(*ni2));
					if (!el2)
						continue;
					const xmlpp::TextNode *tn(el2->get_child_text());
					if (!tn)
						continue;
					if (el2->get_name() == "codeIcaoAcftType") {
						match = true;
						if (m_aircrafttype != tn->get_content())
							return CondResult(false);
					}
					if (el2->get_name() == "codeTypeEngine") {
						match = true;
						if (m_aircrafttypeclass.substr(2, 1) != tn->get_content())
							return CondResult(false);
					}
					if (el2->get_name() == "codeCapability") {
						if (tn->get_content() == "RNAV") {
							match = true;
							if (m_equipment.find("R") == std::string::npos)
								return CondResult(false);
						}
						if (tn->get_content() == "RVSM") {
							match = true;
							if (m_equipment.find("W") == std::string::npos)
								return CondResult(false);
						}
						if (tn->get_content() == "8.33kHz") {
							match = true;
							if (m_equipment.find("Y") == std::string::npos)
								return CondResult(false);
						}
						if (tn->get_content() == "MLS") {
							match = true;
							if (m_equipment.find("K") == std::string::npos)
								return CondResult(false);
						}
						if (tn->get_content() == "ILS") {
							match = true;
							if (m_equipment.find("L") == std::string::npos)
								return CondResult(false);
						}
					}
					// handle <codeEngineNo>OTHER</codeEngineNo>
				}
				if (true || match)
					return CondResult(true);
			}
			nl = cond->get_children("Fcs");
			for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
				xmlpp::Element *el(dynamic_cast<xmlpp::Element *>(*ni));
				if (!el)
					continue;
				bool match(false);
				xmlpp::Node::NodeList nl2(el->get_children(""));
				for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
					xmlpp::Element *el2(dynamic_cast<xmlpp::Element *>(*ni2));
					if (!el2)
						continue;
					const xmlpp::TextNode *tn(el2->get_child_text());
					if (!tn)
						continue;
					if (el2->get_name() == "codeMil") {
						match = true;
						if ("CIVIL" != tn->get_content())
							return CondResult(false);
					}

				}
				if (true || match)
					return CondResult(true);
			}
		}
		m_messages.push_back(Message("Rule Fce element has unknown code " + coderel + " " + cond->get_path(), Message::type_warning, rule.get_codeid()));
		return CondResult();
	}
	m_messages.push_back(Message("Rule has unknown condition element " + cond->get_name() + " " + cond->get_path(), Message::type_warning, rule.get_codeid()));
	return CondResult();
}

TrafficFlowRestrictions::CondResult TrafficFlowRestrictions::compute_tfe(Rule& rule, xmlpp::Element *tfe, RuleResult::Alternative::Sequence& seq)
{
	CondResult r(compute_tfe_1(rule, tfe, seq));
	if (tracerules_exists(rule.get_codeid()) && tfe) {
		std::ostringstream oss;
		oss << "Tfe " << tfe->get_path() << " result: " << (r.get_result() ? 'T' : (!r.get_result() ? 'F' : 'I'));
		Message msg(oss.str(), Message::type_tracetfe, rule.get_codeid());
		msg.m_vertexset = r.get_vertexset();
		msg.m_edgeset = r.get_edgeset();
		m_messages.push_back(msg);
	}
	return r;
}

TrafficFlowRestrictions::CondResult TrafficFlowRestrictions::compute_tfe_1(Rule& rule, xmlpp::Element *tfe, RuleResult::Alternative::Sequence& seq)
{
	seq = RuleResult::Alternative::Sequence(RuleResult::Alternative::Sequence::type_invalid);
	if (!tfe)
		return CondResult(false);
	AltRange altrange(get_rulealtrange(tfe, "Tfl"));
	// Airway
	{
		xmlpp::Node::NodeList nl(tfe->get_children("Rpn"));
		if (!nl.empty()) {
			RuleRoute rte;
			if (!get_ruleawy(rte, tfe)) {
				m_messages.push_back(Message("Rule Tfe element has invalid airway " + tfe->get_path(), Message::type_warning, rule.get_codeid()));
				return CondResult();
			}
			seq = RuleResult::Alternative::Sequence(RuleResult::Alternative::Sequence::type_airway, rte.get_fromwpt(), rte.get_towpt(),
								altrange.get_lower_alt(), altrange.get_upper_alt(), rte.get_airway());
			if (rte.get_from() == invalid_vertex_descriptor ||
			    rte.get_to() == invalid_vertex_descriptor)
				return CondResult(false);
			for (unsigned int wptnr = 1U; wptnr < m_route.size(); ++wptnr) {
				const FplWpt& wpt(m_route[wptnr - 1U]);
				if (wpt.get_vertex_descriptor() != rte.get_from() || !wpt.within(altrange))
					continue;
				CondResult r(true);
				for (unsigned int wptnre(wptnr); wptnre < m_route.size(); ++wptnre) {
					const FplWpt& wpt(m_route[wptnre - 1U]);
					const FplWpt& wpte(m_route[wptnre]);
					if (!wpt.is_ifr() ||
					    wpt.get_pathcode() != FPlanWaypoint::pathcode_airway ||
					    wpt.get_pathname() != rte.get_airway()) {
						r.set_result(false);
						break;
					}
					r.get_edgeset().insert(wptnre - 1U);
					if (wpte.get_vertex_descriptor() == rte.get_to())
						break;
				}
				if (r.get_result())
					return r;
			}
			return CondResult(false);
		}
	}
	// Direct
	{
		xmlpp::Node::NodeList nl(tfe->get_children("Dct"));
		if (!nl.empty()) {
			RuleDct dct;
			if (!get_ruledct(dct, tfe)) {
				m_messages.push_back(Message("Rule Tfe element has invalid direct-to " + tfe->get_path(), Message::type_warning, rule.get_codeid()));
				return CondResult();
			}
			seq = RuleResult::Alternative::Sequence(RuleResult::Alternative::Sequence::type_dct, dct.get_fromwpt(), dct.get_towpt(),
								altrange.get_lower_alt(), altrange.get_upper_alt());
			if (dct.get_from() == invalid_vertex_descriptor ||
			    dct.get_to() == invalid_vertex_descriptor)
				return CondResult(false);
			for (unsigned int wptnr = 1U; wptnr < m_route.size(); ++wptnr) {
				const FplWpt& wpt0(m_route[wptnr - 1U]);
				const FplWpt& wpt1(m_route[wptnr]);
				if (!wpt0.is_ifr() || !wpt0.within(altrange) || wpt0.get_pathcode() != FPlanWaypoint::pathcode_directto)
					continue;
				if (wpt0.get_vertex_descriptor() == dct.get_from() &&
				    wpt1.get_vertex_descriptor() == dct.get_to()) {
					CondResult r(true);
					r.get_edgeset().insert(wptnr - 1U);
					return r;
				}
			}
			return CondResult(false);
		}
	}
	// Single Point
	{
		Graph::vertex_descriptor u;
		RuleWpt wpt;
		if (get_rulewpt(u, wpt, tfe, "Spn")) {
			seq = RuleResult::Alternative::Sequence(RuleResult::Alternative::Sequence::type_point, wpt,
								RuleResult::Alternative::Sequence::RulePoint(),
								altrange.get_lower_alt(), altrange.get_upper_alt());
			for (unsigned int wptnr = 0U; wptnr < m_route.size(); ++wptnr) {
				const FplWpt& wpt(m_route[wptnr]);
				if (!wpt.is_ifr() && (!wptnr || !m_route[wptnr - 1U].is_ifr()))
					continue;
				if (wpt.get_vertex_descriptor() != u || !wpt.within(altrange))
					continue;
				CondResult r(true);
				r.get_vertexset().insert(wptnr);
				return r;
			}
			return CondResult(false);
		}
	}
	// SID/STAR
	{
		RuleResult::Alternative::Sequence::type_t seqtyp(RuleResult::Alternative::Sequence::type_sid);
		FPlanWaypoint::pathcode_t pcode(FPlanWaypoint::pathcode_sid);
		unsigned int wptnr(0U);
		unsigned int wptnrapt(0U);
		xmlpp::Node::NodeList nl(tfe->get_children("SidUid"));
		if (nl.empty()) {
			nl = tfe->get_children("SiaUid");
			seqtyp = RuleResult::Alternative::Sequence::type_star;
			pcode = FPlanWaypoint::pathcode_star;
			wptnr = std::max((unsigned int)m_route.size(), 2U) - 2U;
			wptnrapt = std::max((unsigned int)m_route.size(), 1U) - 1U;
		}
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
			xmlpp::Element *el(dynamic_cast<xmlpp::Element *>(*ni));
			if (!el)
				continue;
			Glib::ustring awy;
			Glib::ustring aprt;
			{
				xmlpp::Node::NodeList nl2(el->get_children("txtDesig"));
				for (xmlpp::Node::NodeList::const_iterator ni(nl2.begin()), ne(nl2.end()); ni != ne; ++ni) {
					xmlpp::Element *el2(dynamic_cast<xmlpp::Element *>(*ni));
					if (!el2)
						continue;
					xmlpp::TextNode *txt(el2->get_child_text());
					if (!txt)
						continue;
					if (txt->is_white_space())
						continue;
					awy = txt->get_content();
				}
			}
			{
				xmlpp::Node::NodeList nl2(el->get_children("AhpUid"));
				for (xmlpp::Node::NodeList::const_iterator ni(nl2.begin()), ne(nl2.end()); ni != ne; ++ni) {
					xmlpp::Element *el2(dynamic_cast<xmlpp::Element *>(*ni));
					if (!el2)
						continue;
					xmlpp::Node::NodeList nl3(el2->get_children("codeId"));
					for (xmlpp::Node::NodeList::const_iterator ni(nl3.begin()), ne(nl3.end()); ni != ne; ++ni) {
						xmlpp::Element *el3(dynamic_cast<xmlpp::Element *>(*ni));
						if (!el3)
							continue;
						xmlpp::TextNode *txt(el3->get_child_text());
						if (!txt)
							continue;
						if (txt->is_white_space())
							continue;
						aprt = txt->get_content();
					}
				}
			}
			if (awy.empty() || aprt.empty())
				continue;
			seq = RuleResult::Alternative::Sequence(seqtyp, RuleResult::Alternative::Sequence::RulePoint(),
								RuleResult::Alternative::Sequence::RulePoint(),
								altrange.get_lower_alt(), altrange.get_upper_alt(),
								awy, aprt);
			const FplWpt& wpt(m_route[wptnr]);
			const FplWpt& wptapt(m_route[wptnrapt]);
			CondResult r(wptapt.is_ifr() && wpt.is_ifr() && wpt.within(altrange) &&
				     wpt.get_pathcode() == pcode && wpt.get_pathname() == awy && wptapt.get_ident() == aprt);
			if (!r.get_result())
				return r;
			r.get_edgeset().insert(wptnr);
			r.get_vertexset().insert(wptnrapt);
			return r;
		}
	}
	// Airspace
	{
		xmlpp::Node::NodeList nl(tfe->get_children("AseUid"));
		if (!nl.empty()) {
			const AirspacesDb::Airspace& aspc(get_ruleairspace(tfe));
			if (!aspc.is_valid()) {
				m_messages.push_back(Message("Rule Tfe element has invalid airspace " + tfe->get_path(), Message::type_warning, rule.get_codeid()));
				return CondResult();
			}
			seq = RuleResult::Alternative::Sequence(RuleResult::Alternative::Sequence::type_airspace,
								RuleResult::Alternative::Sequence::RulePoint(),
								RuleResult::Alternative::Sequence::RulePoint(),
								altrange.get_lower_alt(), altrange.get_upper_alt(),
								aspc.get_icao(), "", aspc.get_bdryclass());
			const MultiPolygonHole& poly(aspc.get_polygon());
			Rect bbox(poly.get_bbox());
			// Rect bbox(aspc.get_bbox());
			for (unsigned int wptnr = 0U; wptnr < m_route.size(); ++wptnr) {
				const FplWpt& wpt(m_route[wptnr]);
				if (!wpt.is_ifr() && (!wptnr || !m_route[wptnr - 1U].is_ifr()))
					continue;
				const Point& pt(wpt.get_coord());
				if (!wpt.within(altrange) || !bbox.is_inside(pt) || !poly.windingnumber(pt))
					continue;
				CondResult r(true);
				r.get_vertexset().insert(wptnr);
				return r;
			}
			return CondResult(false);
		}
	}
	return CondResult();
}
