#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include <iomanip>

#include <boost/graph/dijkstra_shortest_paths.hpp>
#include <boost/graph/johnson_all_pairs_shortest.hpp>
#include <boost/graph/floyd_warshall_shortest.hpp>
#include <boost/graph/filtered_graph.hpp>

#include "cfmuautoroute.hh"
#include "icaofpl.h"
#include "airdata.h"

std::string CFMUAutoroute::Performance::get_fpltas(void) const
{
	std::ostringstream oss;
	oss << 'N' << std::setw(4) << std::setfill('0') << Point::round<int,double>(get_tas());
	return oss.str();
}

const bool CFMUAutoroute::newtfr;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaymatchall;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaymatchnone;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaymatchawy;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaymatchdctawy;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaymatchdctawysidstar;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaymatchdctawysidstarlevel;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaymatchsidstar;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaymatchlevel;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaylevelup;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwayleveldown;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaysid;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaystar;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaydct;
const bool CFMUAutoroute::lgraphrestartedgescan;
const bool CFMUAutoroute::lgraphawyvertdct;

CFMUAutoroute::CFMUAutoroute()
	: m_route(*(FPlan *)0), m_routetime(0), m_routefuel(0), m_deptimepenalty(0), m_depfuelpenalty(0),
	  m_dctlimit(50.0), m_dctpenalty(1.0), m_vfraspclimit(100.0), m_qnh(IcaoAtmosphere::std_sealevel_pressure), m_isaoffs(0),
	  m_enginerpm(2400), m_enginemp(22), m_starttime(0), m_childtimeoutcount(0), m_pathprobe(-1),
	  m_opttarget(opttarget_time), m_childrun(false), m_forceenrifr(false), m_tfrenable(true), m_tfrloaded(false),
	  m_honourawylevels(true), m_running(false), m_done(false)
{
	m_iterationcount[0] = m_iterationcount[1] = 0;
	for (unsigned int i = 0; i < 2; ++i) {
		m_airport[i] = AirportsDb::Airport();
		m_airportconn[i].set_invalid();
		m_airportconnlimit[i] = 40.0;
		m_airportifr[i] = true;
	}
	m_levels[0] = 50;
	m_levels[1] = 120;
	time(&m_starttime);
	clear();
	// 10 o'clock
	set_deptime(3600*10);
	m_tfrenable = get_tfr_available();
}

CFMUAutoroute::~CFMUAutoroute()
{
	child_close();
	clear();
}

void CFMUAutoroute::set_opttarget(opttarget_t t)
{
	if (m_opttarget != t)
		clear();
	m_opttarget = t;
}

const Glib::ustring& to_str(CFMUAutoroute::opttarget_t t)
{
	switch (t) {
	case CFMUAutoroute::opttarget_time:
	{
		static const Glib::ustring r("time");
		return r;
	}

	case CFMUAutoroute::opttarget_fuel:
	{
		static const Glib::ustring r("fuel");
		return r;
	}

	default:
	{
		static const Glib::ustring r("??");
		return r;
	}
	}
}

const Glib::ustring& to_str(CFMUAutoroute::log_t l)
{
	switch (l) {
	case CFMUAutoroute::log_fplproposal:
	{
		static const Glib::ustring r("fplproposal");
		return r;
	}

	case CFMUAutoroute::log_fplvalidation:
	{
		static const Glib::ustring r("fplvalidation");
		return r;
	}

	case CFMUAutoroute::log_graphrule:
	{
		static const Glib::ustring r("graphrule");
		return r;
	}

	case CFMUAutoroute::log_graphchange:
	{
		static const Glib::ustring r("graphchange");
		return r;
	}

	case CFMUAutoroute::log_normal:
	{
		static const Glib::ustring r("normal");
		return r;
	}

	case CFMUAutoroute::log_debug0:
	{
		static const Glib::ustring r("debug0");
		return r;
	}

	case CFMUAutoroute::log_debug1:
	{
		static const Glib::ustring r("debug1");
		return r;
	}

	default:
	{
		static const Glib::ustring r("??");
		return r;
	}
	}
}

void CFMUAutoroute::set_departure(const AirportsDb::Airport& el)
{
	if (m_airport[0].is_valid() != el.is_valid() ||
	    m_airport[0].get_icao() != el.get_icao() ||
	    m_airport[0].get_name() != el.get_name() ||
	    m_airport[0].get_coord() != el.get_coord())
		clear();
	m_airport[0] = el;
}

bool CFMUAutoroute::set_departure(const Glib::ustring& icao, const Glib::ustring& name)
{
	AirportsDb::Airport el;
	if (!find_airport(el, icao, name))
		return false;
	set_departure(el);
	return true;
}

void CFMUAutoroute::set_departure_ifr(bool ifr)
{
	ifr = !!ifr;
	if (m_airportifr[0] != ifr)
		clear();
	m_airportifr[0] = ifr;
}

void CFMUAutoroute::set_sid(const Point& pt)
{
	if (m_airportconn[0] != pt)
		clear();
	m_airportconn[0] = pt;
}

bool CFMUAutoroute::set_sid(const Glib::ustring& name)
{
	Point ptnear(m_airport[0].get_coord());
	if (!m_airport[0].is_valid())
		ptnear.set_invalid();
	Point pt;
	if (!find_point(pt, name, m_airportifr[0], ptnear))
		return false;
	set_sid(pt);
	return true;
}

void CFMUAutoroute::set_sidlimit(double l)
{
	if (std::isnan(l))
		l = 1.;
	l = std::max(l, 0.);
	if (m_airportconnlimit[0] != l)
		clear();
	m_airportconnlimit[0] = l;
}

void CFMUAutoroute::set_destination(const AirportsDb::Airport& el)
{
	if (m_airport[1].is_valid() != el.is_valid() ||
	    m_airport[1].get_icao() != el.get_icao() ||
	    m_airport[1].get_name() != el.get_name() ||
	    m_airport[1].get_coord() != el.get_coord())
		clear();
	m_airport[1] = el;
}

bool CFMUAutoroute::set_destination(const Glib::ustring& icao, const Glib::ustring& name)
{
	AirportsDb::Airport el;
	if (!find_airport(el, icao, name))
		return false;
	set_destination(el);
	return true;
}

void CFMUAutoroute::set_destination_ifr(bool ifr)
{
	ifr = !!ifr;
	if (m_airportifr[1] != ifr)
		clear();
	m_airportifr[1] = ifr;
}

void CFMUAutoroute::set_star(const Point& pt)
{
	if (m_airportconn[1] != pt)
		clear();
	m_airportconn[1] = pt;
}

bool CFMUAutoroute::set_star(const Glib::ustring& name)
{
	Point ptnear(m_airport[1].get_coord());
	if (!m_airport[1].is_valid())
		ptnear.set_invalid();
	Point pt;
	if (!find_point(pt, name, m_airportifr[1], ptnear))
		return false;
	set_star(pt);
	return true;
}

void CFMUAutoroute::set_starlimit(double l)
{
	if (std::isnan(l))
		l = 0.;
	l = std::max(l, 0.);
	if (m_airportconnlimit[1] != l)
		clear();
	m_airportconnlimit[1] = l;
}

void CFMUAutoroute::set_dctlimit(double l)
{
	if (std::isnan(l))
		l = 0.;
	l = std::max(l, 0.);
	if (m_dctlimit != l)
		clear();
	m_dctlimit = l;
}

void CFMUAutoroute::set_dctpenalty(double l)
{
	if (std::isnan(l))
		l = 1.;
	l = std::max(std::min(l, 1.e6), 1.e-6);
	if (m_dctpenalty != l)
		clear();
	m_dctpenalty = l;
}

void CFMUAutoroute::set_vfraspclimit(double l)
{
	if (std::isnan(l))
		l = 0.;
	l = std::max(l, 0.);
	if (m_vfraspclimit != l)
		clear();
	m_vfraspclimit = l;
}

void CFMUAutoroute::set_force_enroute_ifr(bool ifr)
{
	ifr = !!ifr;
	if (m_forceenrifr != ifr)
		clear();
	m_forceenrifr = ifr;
}

bool CFMUAutoroute::get_tfr_available(void) const
{
	std::string fname(get_tfr_filename());
	if (false)
		std::cerr << "TFR Filename " << fname << " exists "
			  << (Glib::file_test(fname, Glib::FILE_TEST_EXISTS) ? "yes" : "no") << " dir "
			  << (Glib::file_test(fname, Glib::FILE_TEST_IS_DIR) ? "yes" : "no") << std::endl;
	if (!Glib::file_test(fname, Glib::FILE_TEST_EXISTS))
		return false;
	if (Glib::file_test(fname, Glib::FILE_TEST_IS_DIR))
		return false;
	return true;
}

void CFMUAutoroute::set_tfr_enabled(bool ena)
{
	m_tfrenable = ena && get_tfr_available();	
}

void CFMUAutoroute::set_qnh(double q)
{
	if (std::isnan(q))
		q = IcaoAtmosphere::std_sealevel_pressure;
	q = std::max(std::min(q, 1300.), 700.);
	if (m_qnh != q)
		clear();
	m_qnh = q;
}

void CFMUAutoroute::set_isaoffs(double t)
{
	if (std::isnan(t))
		t = 0.;
	t = std::max(std::min(t, 100.), -100.);
	if (m_isaoffs != t)
		clear();
	m_isaoffs = t;
}

void CFMUAutoroute::set_engine_rpm(double l)
{
	if (m_enginerpm != l)
		clear();
	m_enginerpm = l;
}

void CFMUAutoroute::set_engine_mp(double l)
{
	if (m_enginemp != l)
		clear();
	m_enginemp = l;
}

void CFMUAutoroute::set_levels(int base, int top)
{
	if (m_levels[0] != base ||
	    m_levels[1] != top)
		clear();
	m_levels[0] = base;
	m_levels[1] = top;
}

void CFMUAutoroute::set_honour_awy_levels(bool awylvl)
{
	awylvl = !!awylvl;
	if (m_honourawylevels != awylvl)
		clear();
	m_honourawylevels = awylvl;
}

void CFMUAutoroute::set_deptime(time_t t)
{
	if (m_route.get_time_offblock_unix() != t)
		clear();
	m_route.set_time_offblock_unix(t);
}

void CFMUAutoroute::set_db_directories(const std::string& dir_main, Engine::auxdb_mode_t auxdbmode, const std::string& dir_aux)
{
	closedb();
	m_dbdir_main = dir_main;
        switch (auxdbmode) {
	case Engine::auxdb_prefs:
		m_dbdir_aux = (Glib::ustring)m_prefs.globaldbdir;
		break;

	case Engine::auxdb_override:
		m_dbdir_aux = dir_aux;
		break;

	default:
		m_dbdir_aux.clear();
		break;
        }
	set_tfr_enabled(get_tfr_enabled());
}

void CFMUAutoroute::opendb(void)
{
	if (!m_airportdb.is_open()) {
		try {
			m_airportdb.open(get_db_maindir());
			if (!get_db_auxdir().empty() && m_airportdb.is_dbfile_exists(get_db_auxdir()))
				m_airportdb.attach(get_db_auxdir());
			if (m_airportdb.is_aux_attached())
				std::cerr << "Auxillary airports database attached" << std::endl;
		} catch (const std::exception& e) {
			m_signal_log(log_normal, std::string("Error opening airports database: ") + e.what());
		}
	}
	if (!m_navaiddb.is_open()) {
		try {
			m_navaiddb.open(get_db_maindir());
			if (!get_db_auxdir().empty() && m_navaiddb.is_dbfile_exists(get_db_auxdir()))
				m_navaiddb.attach(get_db_auxdir());
			if (m_navaiddb.is_aux_attached())
				std::cerr << "Auxillary navaids database attached" << std::endl;
		} catch (const std::exception& e) {
			m_signal_log(log_normal, std::string("Error opening navaids database: ") + e.what());
		}
	}
	if (!m_waypointdb.is_open()) {
		try {
			m_waypointdb.open(get_db_maindir());
			if (!get_db_auxdir().empty() && m_waypointdb.is_dbfile_exists(get_db_auxdir()))
				m_waypointdb.attach(get_db_auxdir());
			if (m_waypointdb.is_aux_attached())
				std::cerr << "Auxillary waypoints database attached" << std::endl;
		} catch (const std::exception& e) {
			m_signal_log(log_normal, std::string("Error opening waypoints database: ") + e.what());
		}
	}
	if (!m_airspacedb.is_open()) {
		try {
			m_airspacedb.open(get_db_maindir());
			if (!get_db_auxdir().empty() && m_airspacedb.is_dbfile_exists(get_db_auxdir()))
				m_airspacedb.attach(get_db_auxdir());
			if (m_airspacedb.is_aux_attached())
				std::cerr << "Auxillary airspaces database attached" << std::endl;
		} catch (const std::exception& e) {
			m_signal_log(log_normal, std::string("Error opening airspaces database: ") + e.what());
		}
	}
	if (!m_airwaydb.is_open()) {
		try {
			m_airwaydb.open(get_db_maindir());
			if (!get_db_auxdir().empty() && m_airwaydb.is_dbfile_exists(get_db_auxdir()))
				m_airwaydb.attach(get_db_auxdir());
			if (m_airwaydb.is_aux_attached())
				std::cerr << "Auxillary airways database attached" << std::endl;
		} catch (const std::exception& e) {
			m_signal_log(log_normal, std::string("Error opening airways database: ") + e.what());
		}
	}
	if (!m_mapelementdb.is_open()) {
		try {
			m_mapelementdb.open(get_db_auxdir().empty() ? Engine::get_default_aux_dir() : get_db_auxdir());
		} catch (const std::exception& e) {
			m_signal_log(log_normal, std::string("Error opening mapelements database: ") + e.what());
		}
	}
	try {
		m_topodb.open(get_db_auxdir().empty() ? Engine::get_default_aux_dir() : get_db_auxdir());
	} catch (const std::exception& e) {
		m_signal_log(log_normal, std::string("Error opening topo database: ") + e.what());
	}
}

void CFMUAutoroute::closedb(void)
{
	if (m_airportdb.is_open())
		m_airportdb.close();
	if (m_navaiddb.is_open())
		m_navaiddb.close();
	if (m_waypointdb.is_open())
		m_waypointdb.close();
	if (m_airwaydb.is_open())
		m_airwaydb.close();
	if (m_airspacedb.is_open())
		m_airspacedb.close();
	if (m_mapelementdb.is_open())
		m_mapelementdb.close();
	m_topodb.close();
}

bool CFMUAutoroute::find_airport(AirportsDb::Airport& el, const Glib::ustring& icao, const Glib::ustring& name)
{
	opendb();
	AirportsDb::elementvector_t ev;
	if (!icao.empty())
		ev = m_airportdb.find_by_icao(icao, 0, AirportsDb::comp_contains, 0, AirportsDb::Airport::subtables_all);
	if (!name.empty()) {
		AirportsDb::elementvector_t ev1(m_airportdb.find_by_name(name, 0, AirportsDb::comp_contains, 0, AirportsDb::Airport::subtables_all));
		ev.insert(ev.end(), ev1.begin(), ev1.end());
	}
	if (ev.empty())
		return false;
	// remove duplicate airports
	for (unsigned int i = 0; i < ev.size(); ++i)
		for (unsigned int j = i + 1; j < ev.size();) {
			if (ev[i].get_sourceid() != ev[j].get_sourceid()) {
				++j;
				continue;
			}
			ev.erase(ev.begin() + j);
		}
	if (ev.size() != 1) {
		typedef std::multimap<int,unsigned int> priomap_t;
		priomap_t priomap;
		Glib::ustring icaou(icao.uppercase()), nameu(name.uppercase());
		for (unsigned int i = 0; i < ev.size(); ++i) {
			const AirportsDb::Airport& arpt(ev[i]);
			Glib::ustring iu(arpt.get_icao().uppercase());
			Glib::ustring nu(arpt.get_name().uppercase());
			int prio(0);
			if (iu == icaou)
				prio -= 2;
			else if (iu.find(icaou) != Glib::ustring::npos)
				--prio;
			if (nu == nameu)
				prio -= 2;
			else if (nu.find(nameu) != Glib::ustring::npos)
				--prio;
			priomap.insert(std::make_pair(prio, i));
		}
		priomap_t::const_iterator it(priomap.begin());
		if (it == priomap.end() || it->first == 0)
			return false;
		priomap_t::const_iterator it2(it);
		++it2;
		if (it2 != priomap.end() && it2->first == it->first)
			return false;
		el = ev[it->second];
		return true;
	}
	el = ev[0];
	return true;
}

bool CFMUAutoroute::find_point(Point& pt, const Glib::ustring& name, bool ifr, const Point& ptnear)
{
	if (name.empty())
		return false;
	typedef std::vector<Point> result_t;
	result_t result;
	Glib::ustring nameu(name.uppercase());
	{
		NavaidsDb::elementvector_t ev(m_navaiddb.find_by_text(nameu, 0, NavaidsDb::comp_contains, 0, NavaidsDb::Navaid::subtables_all));
		for (NavaidsDb::elementvector_t::const_iterator ei(ev.begin()), ee(ev.end()); ei != ee; ++ei) {
			const NavaidsDb::Navaid& el(*ei);
			if (!el.is_valid() || (el.get_icao().uppercase().find(nameu) == Glib::ustring::npos &&
					       el.get_name().uppercase().find(nameu) == Glib::ustring::npos))
				continue;
			result.push_back(el.get_coord());
		}
	}
	{
		WaypointsDb::elementvector_t ev(m_waypointdb.find_by_text(nameu, 0, WaypointsDb::comp_contains, 0, WaypointsDb::Waypoint::subtables_all));
		for (WaypointsDb::elementvector_t::const_iterator ei(ev.begin()), ee(ev.end()); ei != ee; ++ei) {
			const WaypointsDb::Waypoint& el(*ei);
			if (!el.is_valid() || el.get_name().uppercase().find(nameu) == Glib::ustring::npos)
				continue;
			result.push_back(el.get_coord());
		}
	}
	if (!ifr) {
		{
			Point pt;
			if (!((Point::setstr_lat | Point::setstr_lon) & ~pt.set_str(name)) && !pt.is_invalid())
				result.push_back(pt);
		}
		{
			AirportsDb::elementvector_t ev(m_airportdb.find_by_text(nameu, 0, AirportsDb::comp_contains, 0, AirportsDb::Airport::subtables_all));
			for (AirportsDb::elementvector_t::const_iterator ei(ev.begin()), ee(ev.end()); ei != ee; ++ei) {
				const AirportsDb::Airport& el(*ei);
				if (!el.is_valid() || (el.get_icao().uppercase().find(nameu) == Glib::ustring::npos &&
						       el.get_name().uppercase().find(nameu) == Glib::ustring::npos))
					continue;
				result.push_back(el.get_coord());
			}
		}
		{
			MapelementsDb::elementvector_t ev(m_mapelementdb.find_by_text(nameu, 0, MapelementsDb::comp_contains, 0, MapelementsDb::Mapelement::subtables_all));
			for (MapelementsDb::elementvector_t::const_iterator ei(ev.begin()), ee(ev.end()); ei != ee; ++ei) {
				const MapelementsDb::Mapelement& el(*ei);
				if (!el.is_valid() || el.get_name().uppercase().find(nameu) == Glib::ustring::npos)
					continue;
				switch (el.get_typecode()) {
				case MapelementsDb::Mapelement::maptyp_city:
				case MapelementsDb::Mapelement::maptyp_village:
				case MapelementsDb::Mapelement::maptyp_spot:
				case MapelementsDb::Mapelement::maptyp_landmark:
				case MapelementsDb::Mapelement::maptyp_lake:
				case MapelementsDb::Mapelement::maptyp_lake_t:
					break;

				default:
					continue;
				}
				result.push_back(el.get_coord());
			}
		}
	}
	if (ptnear.is_invalid()) {
		if (result.size() == 1) {
			pt = result[0];
			return true;
		}
		return false;
	}
	if (result.empty())
		return false;
	uint64_t dist(ptnear.simple_distance_rel(result[0]));
	unsigned int idx(0);
	for (unsigned i = 1, j = result.size(); i < j; ++i) {
		uint64_t d(ptnear.simple_distance_rel(result[i]));
		if (d >= dist)
			continue;
		dist = d;
		idx = i;
	}
	pt = result[idx];
	return true;
}

Glib::ustring CFMUAutoroute::get_simplefpl(void) const
{
	if (m_performance.empty() || m_route.get_nrwpt() < 3)
		return "";
	std::ostringstream fpl;
	std::ostringstream otherinfo;
	char flightrules(m_route.get_flightrules());
	const FPlanWaypoint& wptdep(m_route[0]);
	const FPlanWaypoint& wptdest(m_route[m_route.get_nrwpt() - 1]);
	int alt(std::numeric_limits<int>::min());
	uint16_t flags(wptdep.get_flags());
	if (flightrules == 'V') {
		alt = m_route.max_altitude(flags);
	} else if (flags & FPlanWaypoint::altflag_ifr) {
		alt = m_route[1].get_altitude();
		flags &= ~FPlanWaypoint::altflag_standard;
		flags |= m_route[1].get_flags() & FPlanWaypoint::altflag_standard;
		if (m_pathprobe >= 0) {
			if (m_pathprobe != 0)
				flags &= ~FPlanWaypoint::altflag_ifr;
			if (m_pathprobe >= m_route.get_nrwpt() && 
			    !(m_route[m_pathprobe].get_flags() & FPlanWaypoint::altflag_ifr))
				flightrules = 'V';
			else
				flightrules = (flags & FPlanWaypoint::altflag_ifr) ? 'Y' : 'Z';
		}
	}
	fpl << "-(FPL-";
	if (!m_done || m_aircraft.get_callsign().empty())
		fpl << m_callsign;
	else
		fpl << m_aircraft.get_callsign();
	fpl << '-' << flightrules << "G -1" << m_aircraft.get_icaotype()
	    << '/' << m_aircraft.get_wakecategory() << " -" << m_aircraft.get_equipment() << '/' << m_aircraft.get_transponder()
	    << " -" << wptdep.get_icao();
	if (wptdep.get_icao() == "ZZZZ")
		otherinfo << "DEP/" << wptdep.get_name() << ' ';
	if (wptdest.get_icao() == "ZZZZ")
		otherinfo << "DEST/" << wptdest.get_name() << ' ';
	if (!m_aircraft.get_pbn().empty())
		otherinfo << "PBN/" << m_aircraft.get_pbn() << ' ';
	{
		struct tm utm;
		time_t deptime(wptdep.get_time_unix());
		gmtime_r(&deptime, &utm);
		fpl << std::setw(2) << std::setfill('0') << utm.tm_hour
		    << std::setw(2) << std::setfill('0') << utm.tm_min
		    << " -";
		if (deptime >= m_starttime) {
			// also add date if in the future
			otherinfo << "DOF/" << std::setw(2) << std::setfill('0') << (utm.tm_year % 100)
				  << std::setw(2) << std::setfill('0') << (utm.tm_mon + 1)
				  << std::setw(2) << std::setfill('0') << utm.tm_mday << ' ';
		}
	}
	if (alt == std::numeric_limits<int>::min())
		fpl << m_performance[findperfindex(m_route.max_altitude())].get_fpltas() << "VFR";
	else
		fpl << m_performance[findperfindex(alt)].get_fpltas()
		    << ((flags & FPlanWaypoint::altflag_standard) ? 'F' : 'A')
		    << std::setw(3) << std::setfill('0') << ((alt + 50) / 100);
	std::ostringstream eet;
	for (unsigned int idx = 1;;) {
		const FPlanWaypoint& wpt(m_route[idx]);
		++idx;
		if (wpt.get_pathcode() == FPlanWaypoint::pathcode_sid) {
			if (idx + 1 >= m_route.get_nrwpt())
				break;
			continue;
		}
		Glib::ustring ident(wpt.get_icao());
		if (ident.empty() || ident == wptdep.get_icao() || ident == wptdest.get_icao())
			ident = wpt.get_name();
		{
			const FPlanWaypoint& wptprev(m_route[idx-2]);
			if (ident.empty() || 
			    ((Engine::AirwayGraphResult::Vertex::is_ident_numeric(ident) || ident.size() < 2) &&
			     wpt.get_pathcode() == FPlanWaypoint::pathcode_airway &&
			     wptprev.get_pathcode() == FPlanWaypoint::pathcode_airway &&
			     wpt.get_pathname() == wptprev.get_pathname() &&
			     (wpt.get_flags() & wptprev.get_flags() & FPlanWaypoint::altflag_ifr))) {
				if (idx + 1 >= m_route.get_nrwpt())
					break;
				continue;
			}
		}
		fpl << ' ' << ident;
		uint16_t flgchg(flags);
		flags = wpt.get_flags();
		if (m_pathprobe >= 0 && idx != m_pathprobe + 1)
			flags &= ~FPlanWaypoint::altflag_ifr;
		flgchg ^= flags;
		{
			int16_t a(wpt.get_altitude());
			if (alt != a)
				flgchg |= FPlanWaypoint::altflag_standard;
			alt = a;
		}
		if (flgchg & FPlanWaypoint::altflag_standard)
			fpl << '/' << m_performance[findperfindex(alt)].get_fpltas()
			    << FPlanWaypoint::get_altchar(flags)
			    << std::setw(3) << std::setfill('0') << ((alt + 50) / 100);
		if (idx + 1 >= m_route.get_nrwpt() || wpt.get_pathcode() == FPlanWaypoint::pathcode_star)
			break;
		if (flgchg & FPlanWaypoint::altflag_ifr) {
			fpl << ' ' << FPlanWaypoint::get_ruleschar(flags) << "FR";
			if (flags & FPlanWaypoint::altflag_ifr) {
				unsigned int tm(wpt.get_time_unix() - wptdep.get_time_unix());
				tm += 30;
				tm /= 60;
				unsigned int hr(tm / 60);
				tm -= hr * 60;
				eet << ident << std::setw(2) << std::setfill('0') << hr
				    << std::setw(2) << std::setfill('0') << tm << ' ';
			}
		}
		if (true || (flags & FPlanWaypoint::altflag_ifr)) {
			switch (wpt.get_pathcode()) {
			case FPlanWaypoint::pathcode_directto:
				fpl << " DCT";
				break;

			case FPlanWaypoint::pathcode_airway:
				fpl << ' ' << wpt.get_pathname();
				break;

			default:
				break;
			}
		}
	}
	{
		uint16_t flgchg(flags);
		flags = wptdest.get_flags();
		flgchg ^= flags;
		if (flgchg & FPlanWaypoint::altflag_ifr && m_pathprobe < 0)
			fpl << ' ' << FPlanWaypoint::get_ruleschar(flags) << "FR";
	}
	fpl << " -" << wptdest.get_icao();
	{
		unsigned int tm(wptdest.get_time_unix() - wptdep.get_time_unix());
		tm += 30;
		tm /= 60;
		unsigned int hr(tm / 60);
		tm -= hr * 60;
		fpl << std::setw(2) << std::setfill('0') << hr
		    << std::setw(2) << std::setfill('0') << tm
		    << " -" << otherinfo.str();
	}
	if (!eet.str().empty())
		fpl << "EET/" << eet.str();
	fpl << ')';
	return fpl.str();
}

void CFMUAutoroute::updatefpl(void)
{
	m_routefuel = m_depfuelpenalty;
	double rttime(0);
	time_t tm(m_route.get_time_offblock_unix());
	tm += 5 * 60;
	unsigned int prevpi(0);
	for (unsigned int i = 0; i < m_route.get_nrwpt(); ++i) {
		FPlanWaypoint& wpt(m_route[i]);
		wpt.set_qff_hpa(m_qnh);
		wpt.set_isaoffset_kelvin(m_isaoffs);
		if (i > 0) {
			unsigned int pi(findperfindex(wpt.get_altitude()));
			double tmadd(0);
			if (i == 1)
				tmadd = m_deptimepenalty;
			for (; prevpi < pi; ++prevpi) {
				const Performance& perf(m_performance[prevpi]);
				tmadd += perf.get_timepenalty();
				m_routefuel += perf.get_fuelpenalty();
			}
			prevpi = pi;
			const Performance& perf(m_performance[prevpi]);
			double x(m_route[i - 1].get_coord().spheric_distance_nmi_dbl(wpt.get_coord()));
			x *= perf.get_secpernmi();
			tmadd += x;
			m_routefuel += x * perf.get_fuelpersec();
			tm += Point::round<time_t,double>(tmadd);
			rttime += tmadd;
		}
		wpt.set_time_unix(tm);
	}
	tm += 5 * 60;
	m_route.set_time_onblock_unix(tm);
	m_routetime = Point::round<unsigned int,double>(rttime);
}

bool CFMUAutoroute::is_running(void) const
{
	return m_running;
}

bool CFMUAutoroute::is_errorfree(void) const
{
	return m_done;
}

double CFMUAutoroute::get_gcdistance(void) const
{
	if (!m_airport[0].is_valid() || !m_airport[1].is_valid())
		return std::numeric_limits<double>::quiet_NaN();
	return m_airport[0].get_coord().spheric_distance_nmi_dbl(m_airport[1].get_coord());
}

double CFMUAutoroute::get_routedistance(void) const
{
	if (m_route.get_nrwpt())
		return m_route.total_distance_nmi();
	return std::numeric_limits<double>::quiet_NaN();
}

unsigned int CFMUAutoroute::get_routetime(void) const
{
	if (m_route.get_nrwpt())
		return m_routetime;
	return 0;
}

double CFMUAutoroute::get_routefuel(void) const
{
	if (m_route.get_nrwpt())
		return m_routefuel;
	return std::numeric_limits<double>::quiet_NaN();
}

unsigned int CFMUAutoroute::get_mintime(void) const
{
	if (!m_airport[0].is_valid() || !m_airport[1].is_valid() || m_performance.empty())
		return 0;
	double dist(m_airport[0].get_coord().spheric_distance_nmi_dbl(m_airport[1].get_coord()));
	double x(std::numeric_limits<double>::max());
	double climb(0);
	for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
		const Performance& perf(m_performance[pi]);
		double y(climb + dist * perf.get_secpernmi());
		x = std::min(x, y);
		climb += perf.get_timepenalty();
	}
	if (x == std::numeric_limits<double>::max())
		return 0;
	return Point::round<unsigned int, double>(x + m_deptimepenalty);
}

double CFMUAutoroute::get_minfuel(void) const
{
	if (!m_airport[0].is_valid() || !m_airport[1].is_valid() || m_performance.empty())
		return std::numeric_limits<double>::quiet_NaN();
	double dist(m_airport[0].get_coord().spheric_distance_nmi_dbl(m_airport[1].get_coord()));
	double x(std::numeric_limits<double>::max());
	double climb(0);
	for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
		const Performance& perf(m_performance[pi]);
		double y(climb + dist * perf.get_secpernmi() * perf.get_fuelpersec());
		x = std::min(x, y);
		climb += perf.get_fuelpenalty();
	}
	if (x == std::numeric_limits<double>::max())
		return std::numeric_limits<double>::quiet_NaN();
	return x + m_depfuelpenalty;
}

void CFMUAutoroute::start(void)
{
	start(false);
}

void CFMUAutoroute::cont(void)
{
	start(true);
}

void CFMUAutoroute::start(bool cont)
{
	stop();
	child_run();
	m_done = false;
	time(&m_starttime);
	bool ok(get_departure().is_valid() && get_destination().is_valid());
	bool perferr(false);
	if (ok) {
		computeperformance();
		perferr = m_performance.empty();
		ok = ok && !perferr;
	}
	if (!ok) {
		m_running = true;
		m_signal_statuschange(statusmask_starting);
		if (perferr)
			m_signal_log(log_normal, "Performance calculation error - aircraft unable to climb to minimum level");
		clear();
		return;
	}
	if (!(m_airportifr[0] || m_airportifr[1] || m_forceenrifr)) {
		clear();
		m_running = true;
		m_signal_statuschange(statusmask_starting);
		m_signal_log(log_normal, "Starting VFR router");
		m_iterationcount[0] = m_iterationcount[1] = 0;
		m_done = start_vfr();
		if (m_done) {
			m_iterationcount[0] = 1;
			m_signal_log(log_fplproposal, get_simplefpl());
		}
		stop();
		return;
	}
	if (!start_ifr(cont))
		stop();
}

void CFMUAutoroute::computeperformance(void)
{
	m_performance.clear();
	m_deptimepenalty = 0;
	m_depfuelpenalty = 0;
	int level(m_levels[0]);
	if (m_airportifr[0] || m_airportifr[1] || m_forceenrifr)
		level = ((level + 9) / 10) * 10;
	else
		level = ((level + 4) / 10) * 10 + 5;
	Aircraft::Climb& climb(m_aircraft.get_climb());
	Aircraft::Cruise& cruise(m_aircraft.get_cruise());
	AirData ad;
	ad.set_qnh_tempoffs(m_qnh, m_isaoffs);
	while (level <= m_levels[1]) {
		ad.set_pressure_altitude(level * 100);
		if (ad.get_density_altitude() > climb.get_ceiling())
			break;
		
		double tas(0), fuelflow(0);
		// compute cruise performance
                {
                        double bhp(0), rpm(get_engine_rpm()), mp(get_engine_mp()), da(ad.get_density_altitude());
                        cruise.calculate(tas, fuelflow, bhp, rpm, mp, da);
                        m_signal_log(log_normal, Glib::ustring::format("Cruise: DA ", std::fixed, std::setprecision(0), da) + 
				     Glib::ustring::format(" TAS ", std::fixed, std::setprecision(0), tas) + 
				     Glib::ustring::format(" Fuelflow ", std::fixed, std::setprecision(1), fuelflow) + 
				     Glib::ustring::format(" RPM ", std::fixed, std::setprecision(0), rpm) + 
				     Glib::ustring::format(" MP ", std::fixed, std::setprecision(1), mp) + 
				     Glib::ustring::format(" Fuel/NMI ", std::fixed, std::setprecision(4), fuelflow / tas));
                }
		if (tas <= 0)
			break;
		double secpernmi(3600.0 / tas);
		double ff(fuelflow * (1.0 / 3600.0));
		double timepenalty(0), fuelpenalty(0);
		bool climberr(false);
		// compute climb performance
		{
			double t0(climb.altitude_to_time(ad.get_density_altitude()));
			ad.set_pressure_altitude(level * 100 + 1000);
			double t1(climb.altitude_to_time(ad.get_density_altitude()));
			climberr = climb.time_to_climbrate(t1) < 100;
			if (!climberr) {
				timepenalty = t1 - t0;
				double cldist(climb.time_to_distance(t1) - climb.time_to_distance(t0));
				if (cldist > 0) {
					timepenalty = std::max(timepenalty - cldist * secpernmi, 0.);
					fuelpenalty = climb.time_to_fuel(t1) - climb.time_to_fuel(t0);
					fuelpenalty = std::max(fuelpenalty - cldist * secpernmi * ff, 0.);
				}				
			}
		}
		m_performance.push_back(Performance(level * 100, secpernmi, ff, timepenalty, fuelpenalty));
		m_signal_log(log_normal, Glib::ustring::format("Climb: DA ", std::fixed, std::setprecision(0), ad.get_density_altitude()) + 
			     Glib::ustring::format(" dTime ", std::fixed, std::setprecision(1), timepenalty) + 
			     Glib::ustring::format(" dFuel ", std::fixed, std::setprecision(3), fuelpenalty));
		if (climberr)
			break;
		level += 10;
	}
	if (m_performance.empty())
		return;
	// compute departure climb performance
	{
		double t0(climb.altitude_to_time(m_airport[0].get_elevation()));
		double t1(climb.altitude_to_time(m_performance[0].get_altitude()));
		m_deptimepenalty = t1 - t0;
		double cldist(climb.time_to_distance(t1) - climb.time_to_distance(t0));
		if (cldist > 0) {
			m_deptimepenalty = std::max(m_deptimepenalty - cldist * m_performance[0].get_secpernmi(), 0.);
			m_depfuelpenalty = climb.time_to_fuel(t1) - climb.time_to_fuel(t0);
			m_depfuelpenalty = std::max(m_depfuelpenalty - cldist * m_performance[0].get_secpernmi() * m_performance[0].get_fuelpersec(), 0.);
		}				
	}

}

unsigned int CFMUAutoroute::findperfindex(int alt) const
{
	int adiff(std::numeric_limits<int>::max());
	unsigned int idx(0);
	for (unsigned int i = 0; i < m_performance.size(); ++i) {
		int ad(abs(alt - m_performance[i].get_altitude()));
		if (ad > adiff)
			continue;
		adiff = ad;
		idx = i;
	}
	return idx;
}

bool CFMUAutoroute::start_vfr(void)
{
	VFRRouter r(*this);
	if (!r.load_graph(get_bbox())) {
		std::ostringstream oss;
		oss << "VFR router: cannot load routing graph for bbox " << get_bbox();
		m_signal_log(log_debug0, oss.str());
		return false;
	}
	{
		std::ostringstream oss;
		oss << "VFR area graph after loading: " << boost::num_vertices(r.get_graph()) << " vertices, "
		    << boost::num_edges(r.get_graph()) << " edges";
		m_signal_log(log_debug0, oss.str());
	}
	if (!r.set_departure(get_departure())) {
		std::ostringstream oss;
		oss << "VFR router: cannot find departure airport " << get_departure().get_icao_name();
		m_signal_log(log_debug0, oss.str());
		return false;
	}
	if (!r.set_destination(get_destination())) {
		std::ostringstream oss;
		oss << "VFR router: cannot find destination airport " << get_destination().get_icao_name();
		m_signal_log(log_debug0, oss.str());
		return false;
	}
	bool havesid(r.add_sid(get_sid()));
	if (!havesid && !get_sid().is_invalid()) {
		std::ostringstream oss;
		oss << "VFR router: cannot find SID point " << get_sid().get_lat_str2() << ' ' << get_sid().get_lon_str2();
		m_signal_log(log_debug0, oss.str());
		return false;
	}
	havesid = havesid || r.add_sid(get_departure());
	bool havestar(r.add_star(get_star()));
	if (!havestar && !get_star().is_invalid()) {
		std::ostringstream oss;
		oss << "VFR router: cannot find STAR point " << get_star().get_lat_str2() << ' ' << get_star().get_lon_str2();
		m_signal_log(log_debug0, oss.str());
		return false;
	}
	havestar = havestar || r.add_star(get_destination());
	{
		std::ostringstream oss;
		oss << "VFR area graph after SID/STAR: " << boost::num_vertices(r.get_graph()) << " vertices, "
		    << boost::num_edges(r.get_graph()) << " edges";
		m_signal_log(log_debug0, oss.str());
	}
	r.add_dct(get_dctlimit(),
		  havesid ? std::numeric_limits<double>::quiet_NaN() : get_sidlimit(),
		  havestar ? std::numeric_limits<double>::quiet_NaN() : get_starlimit());
	{
		std::ostringstream oss;
		oss << "VFR area graph after DCT: " << boost::num_vertices(r.get_graph()) << " vertices, "
		    << boost::num_edges(r.get_graph()) << " edges";
		m_signal_log(log_debug0, oss.str());
	}
	r.set_metric();
	r.exclude_airspace(m_levels[0], m_levels[1], get_vfraspclimit());
	if (!r.autoroute(m_route)) {
		std::ostringstream oss;
		oss << "VFR router: cannot find route " << get_departure().get_icao_name()
		    << "->" << get_destination().get_icao_name();
		m_signal_log(log_debug0, oss.str());
		return false;
	}
	{
		const Performance& perf(m_performance[m_performance.size() / 2U]);
		for (unsigned int i = 0; i < m_route.get_nrwpt(); ++i) {
			FPlanWaypoint& wpt(m_route[i]);
			uint16_t flags(wpt.get_flags());
			flags &= ~FPlanWaypoint::altflag_ifr;
			if (i > 0 && i + 1 < m_route.get_nrwpt()) {
				flags |= FPlanWaypoint::altflag_standard;
				wpt.set_altitude(perf.get_altitude());
			}
			wpt.set_flags(flags);
		}
	}
	updatefpl();
	return true;
}

bool CFMUAutoroute::start_ifr(bool cont, bool optfuel)
{
	cont = cont && (m_superdep < m_lsupervertices.size())
		&& (m_superdest < m_lsupervertices.size())
		&& boost::num_vertices(m_lgraph);
	if (!cont) {
		clear();
		m_callsign.clear();
		for (unsigned int i = 0; i < 5; ++i) {
			char ch('A' + (rand() % ('Z' - 'A' + 1)));
			m_callsign.push_back(ch);
		}
		if (get_tfr_enabled()) {
			if (m_tfrloaded) {
				m_tfr.reset_rules();
			} else if (newtfr) {
				opendb();
				m_signal_log(log_normal, "Loading traffic flow restrictions " + get_tfr_filename());
				std::vector<TrafficFlowRestrictions::Message> msg;
				Glib::TimeVal tv;
				tv.assign_current_time();
				bool ok(m_tfr.load_rules(msg, get_tfr_filename(), m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb));
				{
					Glib::TimeVal tv1;
					tv1.assign_current_time();
					tv = tv1 - tv;
				}
				for (unsigned int i = 0; i < msg.size(); ++i)
					logmessage(msg[i]);
				if (!ok) {
					set_tfr_enabled(false);
					m_signal_log(log_normal, "Cannot load traffic flow restrictions " + get_tfr_filename());
				} else {
					m_tfr.reset_rules();
					std::ostringstream oss;
					oss << m_tfr.count_rules() << " rules loaded traffic flow restrictions " << get_tfr_filename()
					    << ", " << std::fixed << std::setprecision(3) << tv.as_double() << 's';
					m_signal_log(log_normal, oss.str());
				}
			} else {
				if (!m_tfr.load(get_tfr_filename())) {
					set_tfr_enabled(false);
					m_signal_log(log_normal, "Cannot load traffic flow restrictions " + get_tfr_filename());
				} else {
					m_signal_log(log_normal, "Loaded traffic flow restrictions " + get_tfr_filename());
					TrafficFlowRestrictions::Rule rule(m_tfr.get_first_rule());
					for (;;) {
						if (!rule.is_valid())
							break;
						std::ostringstream oss;
						oss << "Rule " << rule.get_codeid() << ' ' << rule.get_codetype_char();
						m_signal_log(log_debug0, oss.str());
						rule.next_rule();
					}
				}
			}
		}
		if (get_tfr_enabled()) {
			//m_tfr.tracerules_add("LS1A");
			//m_tfr.tracerules_add("ED1A");
			if (newtfr) {
				m_tfr.reset_rules();
				if (true) {
					Glib::TimeVal tv;
					tv.assign_current_time();
					m_tfr.simplify_rules_typeofflight('G');
					m_tfr.simplify_rules_mil(false);
					m_tfr.simplify_rules_gat(true);
					{
						Glib::TimeVal tv1;
						tv1.assign_current_time();
						tv = tv1 - tv;
					}
					std::ostringstream oss;
					oss << m_tfr.count_rules() << " rules after fixing type of flight G CIV GAT, "
					    << std::fixed << std::setprecision(3) << tv.as_double() << 's';
					m_signal_log(log_normal, oss.str());
				}
				if (true) {
					Glib::TimeVal tv;
					tv.assign_current_time();
					if (m_airport[0].is_valid() && m_airport[0].get_icao().size() == 4 && m_airport[0].get_icao() != "ZZZZ")
						m_tfr.simplify_rules_dep(m_airport[0].get_icao(), m_airport[0].get_coord());
					if (m_airport[1].is_valid() && m_airport[1].get_icao().size() == 4 && m_airport[1].get_icao() != "ZZZZ")
						m_tfr.simplify_rules_dest(m_airport[1].get_icao(), m_airport[1].get_coord());
					{
						Glib::TimeVal tv1;
						tv1.assign_current_time();
						tv = tv1 - tv;
					}
					std::ostringstream oss;
					oss << m_tfr.count_rules() << " rules after fixing departure " << m_airport[0].get_icao()
					    << " destination " << m_airport[1].get_icao()
					    << ", " << std::fixed << std::setprecision(3) << tv.as_double() << 's';
					m_signal_log(log_normal, oss.str());
				}
				if (true) {
					Glib::TimeVal tv;
					tv.assign_current_time();
					m_tfr.simplify_rules_aircrafttype(m_aircraft.get_aircrafttype());
					{
						Glib::TimeVal tv1;
						tv1.assign_current_time();
						tv = tv1 - tv;
					}
					std::ostringstream oss;
					oss << m_tfr.count_rules() << " rules after fixing aircraft type " << m_aircraft.get_aircrafttype()
					    << ", " << std::fixed << std::setprecision(3) << tv.as_double() << 's';
					m_signal_log(log_normal, oss.str());
				}
				if (true) {
					Glib::TimeVal tv;
					tv.assign_current_time();
					m_tfr.simplify_rules_equipment(m_aircraft.get_equipment());
					{
						Glib::TimeVal tv1;
						tv1.assign_current_time();
						tv = tv1 - tv;
					}
					std::ostringstream oss;
					oss << m_tfr.count_rules() << " rules after fixing aircraft equipment " << m_aircraft.get_equipment()
					    << ", " << std::fixed << std::setprecision(3) << tv.as_double() << 's';
					m_signal_log(log_normal, oss.str());
				}
				if (true) {
					Glib::TimeVal tv;
					tv.assign_current_time();
					m_tfr.simplify_rules_altrange(m_levels[0] * 100, m_levels[1] * 100);
					{
						Glib::TimeVal tv1;
						tv1.assign_current_time();
						tv = tv1 - tv;
					}
					std::ostringstream oss;
					oss << m_tfr.count_rules() << " rules after fixing cruise levels F"
					    << std::setw(3) << std::setfill('0') << m_levels[0] << "..F"
					    << std::setw(3) << std::setfill('0') << m_levels[1]
					    << ", " << std::fixed << std::setprecision(3) << tv.as_double() << 's';
					m_signal_log(log_normal, oss.str());
				}
				if (true) {
					Glib::TimeVal tv;
					tv.assign_current_time();
					Rect bbox(get_bbox().oversize_nmi(200.f));
					m_tfr.simplify_rules_bbox(bbox);
					{
						Glib::TimeVal tv1;
						tv1.assign_current_time();
						tv = tv1 - tv;
					}
					std::ostringstream oss;
					oss << m_tfr.count_rules() << " rules after fixing area ["
					    << bbox.get_southwest().get_lat_str2() << ' ' << bbox.get_southwest().get_lon_str2() << ' '
					    << bbox.get_northeast().get_lat_str2() << ' ' << bbox.get_northeast().get_lon_str2()
					    << "], " << std::fixed << std::setprecision(3) << tv.as_double() << 's';
					m_signal_log(log_normal, oss.str());
				}
				{
					Glib::TimeVal tv;
					tv.assign_current_time();
					m_tfr.prepare_dct_rules();
					{
						Glib::TimeVal tv1;
						tv1.assign_current_time();
						tv = tv1 - tv;
					}
					std::ostringstream oss;
					oss << m_tfr.count_dct_rules() << " DCT rules, "
					    << std::fixed << std::setprecision(3) << tv.as_double() << 's';
					m_signal_log(log_normal, oss.str());
				}
			}
		} else {
			m_signal_log(log_normal, "Traffic flow restrictions disabled");
		}
	}
	m_running = true;
	m_signal_statuschange(statusmask_starting);
	m_signal_log(log_normal, cont ? "Restarting IFR router" : "Starting IFR router");
	if (!cont) {
		m_iterationcount[0] = m_iterationcount[1] = 0;
		clearlgraph();
		if (!lgraphload(get_bbox()))
			return false;
		lgraphkillinvalidsuper();
		lgraphadddct();
		if (true) {
			std::ostringstream oss;
			oss << "Airway graph after DCT: " << boost::num_vertices(m_lgraph) << " vertices, "
			    << boost::num_edges(m_lgraph) << " edges";
			m_signal_log(log_debug0, oss.str());
		}
		if (!lgraphaddsidstar()) {
			if (true)
				m_signal_log(log_debug0, "Cannot add SID/STAR to Airway graph");
			return false;
		}
		if (true) {
			std::ostringstream oss;
			oss << "Airway graph after SID/STAR: " << boost::num_vertices(m_lgraph) << " vertices, "
			    << boost::num_edges(m_lgraph) << " edges";
	       		m_signal_log(log_debug0, oss.str());
		}
	}
	lgraphmodified();
	lgraphroute();
	return true;
}

void CFMUAutoroute::stop(void)
{
	m_childtimeoutcount = 0;
	m_connchildtimeout.disconnect();
	m_pathprobe = -1;
	if (m_running) {
		m_running = false;
		if (m_done)
			m_signal_statuschange(statusmask_stoppingdone);
		else
			m_signal_statuschange(statusmask_stoppingerror);
	}
}

void CFMUAutoroute::preload(void)
{
	child_run();
}

void CFMUAutoroute::clear(void)
{
	stop();
	while (m_route.get_nrwpt())
		m_route.delete_wpt(0);
	m_routetime = 0;
	m_routefuel = 0;
	m_validationresponse.clear();
	clearlgraph();
	m_done = false;
}

void CFMUAutoroute::child_run(void)
{
	if (m_childrun)
		return;
	m_signal_log(log_normal, "Connecting to validation server...");
	int childstdin(-1), childstdout(-1);
	try {
		std::vector<std::string> argv, env;
#ifdef __MINGW32__
		std::string workdir;
		{
			gchar *instdir(g_win32_get_package_installation_directory_of_module(0));
			if (instdir) {
				workdir = instdir;
				argv.push_back(Glib::build_filename(instdir, "bin", "cfmuvalidateserver.exe"));
			} else {
				argv.push_back(Glib::build_filename(PACKAGE_PREFIX_DIR "/bin", "cfmuvalidateserver.exe"));
				workdir = PACKAGE_PREFIX_DIR;
			}
		}
#else
		std::string workdir(PACKAGE_DATA_DIR);
		argv.push_back(Glib::build_filename(PACKAGE_LIBEXEC_DIR, "cfmuvalidateserver"));
		env.push_back("PATH=/bin:/usr/bin");
		{
			bool found(false);
			std::string disp(Glib::getenv("DISPLAY", found));
			if (found)
				env.push_back("DISPLAY=" + disp);
		}
#endif
		if (false) {
			std::cerr << "Spawn: workdir " << workdir << " argv:" << std::endl;
			for (std::vector<std::string>::const_iterator i(argv.begin()), e(argv.end()); i != e; ++i)
				std::cerr << "  " << *i << std::endl;
			std::cerr << "env:" << std::endl;
			for (std::vector<std::string>::const_iterator i(env.begin()), e(env.end()); i != e; ++i)
				std::cerr << "  " << *i << std::endl;
		}
#ifdef __MINGW32__
		// inherit environment; otherwise, this may fail on WindowsXP with STATUS_SXS_ASSEMBLY_NOT_FOUND (C0150004)
		Glib::spawn_async_with_pipes(workdir, argv,
					     Glib::SPAWN_DO_NOT_REAP_CHILD | Glib::SPAWN_STDERR_TO_DEV_NULL,
					     sigc::slot<void>(), &m_childpid, &childstdin, &childstdout, 0);
#else
		Glib::spawn_async_with_pipes(workdir, argv, env,
					     Glib::SPAWN_DO_NOT_REAP_CHILD | Glib::SPAWN_STDERR_TO_DEV_NULL,
					     sigc::slot<void>(), &m_childpid, &childstdin, &childstdout, 0);
#endif
		m_childrun = true;
	} catch (Glib::SpawnError& e) {
		m_childrun = false;
		m_signal_log(log_normal, "Cannot run validator: " + e.what());
		return;
	}
	m_connchildwatch = Glib::signal_child_watch().connect(sigc::mem_fun(this, &CFMUAutoroute::child_watch), m_childpid);
	m_childchanstdin = Glib::IOChannel::create_from_fd(childstdin);
	m_childchanstdout = Glib::IOChannel::create_from_fd(childstdout);
	m_childchanstdin->set_encoding();
	m_childchanstdout->set_encoding();
	m_childchanstdin->set_close_on_unref(true);
	m_childchanstdout->set_close_on_unref(true);
	m_connchildstdout = Glib::signal_io().connect(sigc::mem_fun(this, &CFMUAutoroute::child_stdout_handler), m_childchanstdout,
						      Glib::IO_IN | Glib::IO_ERR | Glib::IO_HUP);
}

bool CFMUAutoroute::child_stdout_handler(Glib::IOCondition iocond)
{
        Glib::ustring line;
        Glib::IOStatus iostat(m_childchanstdout->read_line(line));
        if (iostat == Glib::IO_STATUS_ERROR ||
            iostat == Glib::IO_STATUS_EOF) {
		if (iostat == Glib::IO_STATUS_ERROR)
			m_signal_log(log_debug0, "Child stdout error, stopping...");
		else
			m_signal_log(log_debug0, "Child stdout eof, stopping...");
                child_close();
		stop();
		return true;
        }
	{
		Glib::ustring::size_type n(line.size());
		while (n > 0) {
			--n;
			if (!isspace(line[n]) && line[n] != '\n' && line[n] != '\r') {
				++n;
				break;
			}
		}
		line = Glib::ustring(line, 0, n);
	}
	if (!line.empty())
		m_validationresponse.push_back(line);
	m_signal_log(log_fplvalidation, line);
	if (line.empty()) {
		parse_response();
	}
        return true;
}

void CFMUAutoroute::child_watch(GPid pid, int child_status)
{
	if (!m_childrun)
		return;
	if (m_childpid != pid) {
		Glib::spawn_close_pid(pid);
		return;
	}
	Glib::spawn_close_pid(pid);
	m_childrun = false;
	m_connchildwatch.disconnect();
	m_connchildstdout.disconnect();
	child_close();
	child_timeout();
}

void CFMUAutoroute::child_close(void)
{
	m_connchildstdout.disconnect();
	if (m_childchanstdin) {
		m_childchanstdin->close();
		m_childchanstdin.reset();
	}
	if (m_childchanstdout) {
		m_childchanstdout->close();
		m_childchanstdout.reset();
	}
	if (!m_childrun)
		return;
	m_childrun = false;
	m_connchildwatch.disconnect();
	Glib::signal_child_watch().connect(sigc::hide(sigc::ptr_fun(&Glib::spawn_close_pid)), m_childpid);
}

Rect CFMUAutoroute::get_bbox(void) const
{
	if (!get_departure().is_valid() && !get_destination().is_valid())
		return Rect();
	if (!get_departure().is_valid())
		return Rect(get_destination().get_coord(), get_destination().get_coord());
	if (!get_destination().is_valid())
		return Rect(get_departure().get_coord(), get_departure().get_coord());
	Point pt1(get_departure().get_coord());
	Point pt2(get_destination().get_coord() - pt1);
	return Rect(Point(std::min(pt2.get_lon(), (Point::coord_t)0), std::min(pt2.get_lat(), (Point::coord_t)0)) + pt1,
		    Point(std::max(pt2.get_lon(), (Point::coord_t)0), std::max(pt2.get_lat(), (Point::coord_t)0)) + pt1);
}


void CFMUAutoroute::logmessage(const TrafficFlowRestrictions::Message& msg)
{
	std::ostringstream oss;
	oss << msg.get_type_char() << ':';
	if (!msg.get_rule().empty())
		oss << " R:" << msg.get_rule();
	if (!msg.get_vertexset().empty()) {
		oss << " V:";
		bool subseq(false);
		for (TrafficFlowRestrictions::Message::set_t::const_iterator i(msg.get_vertexset().begin()), e(msg.get_vertexset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	if (!msg.get_edgeset().empty()) {
		oss << " E:";
		bool subseq(false);
		for (TrafficFlowRestrictions::Message::set_t::const_iterator i(msg.get_edgeset().begin()), e(msg.get_edgeset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	oss << ' ' << msg.get_text();
	log_t lt(log_debug1);
	switch (msg.get_type()) {
	case TrafficFlowRestrictions::Message::type_error:
		lt = log_normal;
		break;

	case TrafficFlowRestrictions::Message::type_warning:
		lt = log_debug0;
		break;

	default:
		lt = log_debug1;
		break;
	}
	m_signal_log(lt, oss.str());
}

CFMUAutoroute::VFRRouter::VFRRouter(CFMUAutoroute& ar)
	: m_ar(ar)
{
}

bool CFMUAutoroute::VFRRouter::load_graph(const Rect& bbox)
{
	m_bbox = bbox.oversize_nmi(100.f);
	if (false) {
		std::ostringstream oss;
		oss << "Loading VFR area graph: " << m_bbox;
		m_ar.m_signal_log(log_debug0, oss.str());
	}
	m_ar.opendb();
	m_graph.clear();
	m_graph.add(m_ar.m_airportdb.find_by_rect(m_bbox, ~0, AirportsDb::Airport::subtables_none));
	m_graph.add(m_ar.m_navaiddb.find_by_rect(m_bbox, ~0, NavaidsDb::Navaid::subtables_none));
	m_graph.add(m_ar.m_waypointdb.find_by_rect(m_bbox, ~0, WaypointsDb::Waypoint::subtables_none));
	m_graph.add(m_ar.m_mapelementdb.find_by_rect(m_bbox, ~0, MapelementsDb::Mapelement::subtables_none));
	// We do not need airways for VFR routing
	//m_graph.add(m_ar.m_airwaydb.find_by_area(m_bbox, ~0, AirwaysDb::Airway::subtables_all));
	//m_graph.add(m_ar.m_airwaydb.find_by_rect(m_bbox, ~0, AirwaysDb::Airway::subtables_all));
	if (true) {
		if (false) {
			std::ostringstream oss;
			oss << "VFR area graph: " << m_bbox << ", "
			    << boost::num_vertices(m_graph) << " vertices, " << boost::num_edges(m_graph) << " edges";
			m_ar.m_signal_log(log_debug0, oss.str());
		}
		unsigned int counts[4];
		for (unsigned int i = 0; i < sizeof(counts)/sizeof(counts[0]); ++i)
			counts[i] = 0;
		Graph::vertex_iterator u0, u1;
		for (boost::tie(u0, u1) = boost::vertices(m_graph); u0 != u1; ++u0) {
			const Vertex& uu(m_graph[*u0]);
			unsigned int i(uu.get_type());
			if (i < sizeof(counts)/sizeof(counts[0]))
				++counts[i];
		}
		{
			std::ostringstream oss;
			oss << "VFR area graph: " << counts[Vertex::type_airport] << " airports, " << counts[Vertex::type_navaid] << " navaids, "
			    << counts[Vertex::type_intersection] << " intersections, " << counts[Vertex::type_mapelement] << " mapelements";
			m_ar.m_signal_log(log_debug0, oss.str());
		}
	}
	return true;
}

CFMUAutoroute::VFRRouter::Graph::vertex_descriptor CFMUAutoroute::VFRRouter::find_airport(const AirportsDb::Airport& arpt)
{
	if (!arpt.is_valid())
		return boost::num_vertices(m_graph);
	Graph::vertex_iterator u0, u1;
	for (boost::tie(u0, u1) = boost::vertices(m_graph); u0 != u1; ++u0) {
		const Vertex& uu(m_graph[*u0]);
		AirportsDb::Airport a;
		if (!uu.get(a))
			continue;
		if (!a.is_valid())
			continue;
		if (arpt.get_icao() != a.get_icao() ||
		    arpt.get_name() != a.get_name() ||
		    arpt.get_coord() != a.get_coord())
			continue;
		return *u0;
	}
	return boost::num_vertices(m_graph);
}

bool CFMUAutoroute::VFRRouter::set_departure(const AirportsDb::Airport& arpt)
{
	m_dep = find_airport(arpt);
	return m_dep < boost::num_vertices(m_graph);
}

bool CFMUAutoroute::VFRRouter::set_destination(const AirportsDb::Airport& arpt)
{
	m_dest = find_airport(arpt);
	return m_dest < boost::num_vertices(m_graph);
}

void CFMUAutoroute::VFRRouter::add_dct(double dctlim, double sidlim, double starlim)
{
	Graph::vertex_iterator u0, u1;
	for (boost::tie(u0, u1) = boost::vertices(m_graph); u0 != u1; ++u0) {
		Graph::vertex_descriptor u(*u0);
		const Vertex& uu(m_graph[u]);
		Graph::vertex_iterator v0, v1;
		for (boost::tie(v0, v1) = boost::vertices(m_graph); v0 != v1; ++v0) {
			Graph::vertex_descriptor v(*v0);
			if (u == v)
				continue;
			const Vertex& vv(m_graph[*v0]);
			bool kill(u == m_dep && v == m_dest);
			kill = kill || (uu.get_type() == Vertex::type_intersection && uu.is_ident_numeric());
			kill = kill || (vv.get_type() == Vertex::type_intersection && vv.is_ident_numeric());
			double dist, lim(dctlim);
			if (!kill) {
				dist = uu.get_coord().spheric_distance_nmi_dbl(vv.get_coord());
				if (u == m_dep)
					lim = sidlim;
				if (v == m_dest)
					lim = starlim;
				if (!std::isnan(lim))
					kill = dist > lim;
			}
			if (kill) {
				for (;;) {
					bool done(true);
					Graph::out_edge_iterator e0, e1;
					for (boost::tie(e0, e1) = boost::out_edges(u, m_graph); e0 != e1;) {
						Graph::out_edge_iterator ex(e0);
						++e0;
						if (boost::target(*ex, m_graph) != v)
							continue;
						boost::remove_edge(ex, m_graph);
						if (lgraphrestartedgescan) {
							done = false;
							break;
						}
					}
					if (done)
						break;
				}
				continue;
			}
			if (std::isnan(lim))
				continue;
			Graph::edge_descriptor e;
			bool hase;
			tie(e, hase) = boost::edge(u, v, m_graph);
			if (hase)
				continue;
			boost::add_edge(u, v, Edge("", 0, 600, AirwaysDb::Airway::airway_invalid, dist), m_graph);
		}
	}
}

void CFMUAutoroute::VFRRouter::remove_sid(void)
{
	for (;;) {
		bool done(true);
		Graph::out_edge_iterator e0, e1;
		for (boost::tie(e0, e1) = boost::out_edges(m_dep, m_graph); e0 != e1;) {
			Graph::out_edge_iterator ex(e0);
			++e0;
			boost::remove_edge(ex, m_graph);
			if (lgraphrestartedgescan) {
				done = false;
				break;
			}
		}
		if (done)
			break;
	}
}

bool CFMUAutoroute::VFRRouter::add_sid(const Point& pt)
{
	if (pt.is_invalid())
		return false;
	Graph::vertex_descriptor v;
	if (!m_graph.find_nearest(v, pt, Vertex::typemask_all))
		return false;
	if (v == m_dest)
		return false;
	const Vertex& vv(m_graph[v]);
	double dist(pt.spheric_distance_nmi_dbl(vv.get_coord()));
	if (dist > 0.1)
		return false;
	// remove departure out edges
	remove_sid();
	// add edge
	boost::add_edge(m_dep, v, Edge("", 0, 600, AirwaysDb::Airway::airway_invalid, dist), m_graph);
	m_sid.clear();
	return true;
}

bool CFMUAutoroute::VFRRouter::add_sid(const AirportsDb::Airport& arpt)
{
	bool added(false);
	m_sid.clear();
	for (unsigned int i = 0, j = arpt.get_nrvfrrte(); i < j; ++i) {
		const AirportsDb::Airport::VFRRoute& rte(arpt.get_vfrrte(i));
		if (!rte.size() ||
		    rte.get_name().empty())
			continue;
		const AirportsDb::Airport::VFRRoute::VFRRoutePoint& rtept(rte[rte.size()-1]);
		if (rtept.get_pathcode() != AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_departure)
			continue;
		Graph::vertex_descriptor v;
		{
			bool found(m_graph.find_nearest(v, rtept.get_coord(), Vertex::typemask_all));
			if (found) {
				const Vertex& vv(m_graph[v]);
				double dist(rtept.get_coord().spheric_distance_nmi_dbl(vv.get_coord()));
				if (dist > 0.1)
					found = false;
			}
			if (!found)
				v = boost::add_vertex(Vertex(rte.get_name(), rtept.get_coord(), Vertex::type_vfrreportingpt), m_graph);
		}
		if (!added)
			remove_sid();
		added = true;
		boost::add_edge(m_dep, v, Edge("", 0, 600, AirwaysDb::Airway::airway_invalid, rte.compute_distance()), m_graph);
		m_sid.push_back(rte);
	}
	return added;
}

void CFMUAutoroute::VFRRouter::remove_star(void)
{
	for (;;) {
		bool done(true);
		Graph::in_edge_iterator e0, e1;
		for (boost::tie(e0, e1) = boost::in_edges(m_dest, m_graph); e0 != e1;) {
			Graph::in_edge_iterator ex(e0);
			++e0;
			boost::remove_edge(*ex, m_graph);
			if (lgraphrestartedgescan) {
				done = false;
				break;
			}
		}
		if (done)
			break;
	}
}

bool CFMUAutoroute::VFRRouter::add_star(const Point& pt)
{
	if (pt.is_invalid())
		return false;
	Graph::vertex_descriptor u;
	if (!m_graph.find_nearest(u, pt, Vertex::typemask_all))
		return false;
	if (u == m_dep)
		return false;
	const Vertex& uu(m_graph[u]);
	double dist(pt.spheric_distance_nmi_dbl(uu.get_coord()));
	if (dist > 0.1)
		return false;
	// remove destination in edges
	remove_star();
	// add edge
	boost::add_edge(u, m_dest, Edge("", 0, 600, AirwaysDb::Airway::airway_invalid, dist), m_graph);
	m_star.clear();
	return true;
}

bool CFMUAutoroute::VFRRouter::add_star(const AirportsDb::Airport& arpt)
{
	bool added(false);
	m_star.clear();
	for (unsigned int i = 0, j = arpt.get_nrvfrrte(); i < j; ++i) {
		const AirportsDb::Airport::VFRRoute& rte(arpt.get_vfrrte(i));
		if (!rte.size() || rte.get_name().empty())
			continue;
		const AirportsDb::Airport::VFRRoute::VFRRoutePoint& rtept(rte[0]);
		if (rtept.get_pathcode() != AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_arrival)
			continue;
		Graph::vertex_descriptor v;
		{
			bool found(m_graph.find_nearest(v, rtept.get_coord(), Vertex::typemask_all));
			if (found) {
				const Vertex& vv(m_graph[v]);
				double dist(rtept.get_coord().spheric_distance_nmi_dbl(vv.get_coord()));
				if (dist > 0.1)
					found = false;
			}
			if (!found)
				v = boost::add_vertex(Vertex(rte.get_name(), rtept.get_coord(), Vertex::type_vfrreportingpt), m_graph);
		}
		if (!added)
			remove_star();
		added = true;
		boost::add_edge(v, m_dest, Edge("", 0, 600, AirwaysDb::Airway::airway_invalid, rte.compute_distance()), m_graph);
		m_star.push_back(rte);
	}
	return added;
}

void CFMUAutoroute::VFRRouter::set_metric(void)
{
	Graph::edge_iterator e0, e1;
	for (boost::tie(e0, e1) = boost::edges(m_graph); e0 != e1; ++e0) {
		Edge& ee(m_graph[*e0]);
		ee.set_metric(ee.get_dist());
	}
}

void CFMUAutoroute::VFRRouter::exclude_airspace(int baselevel, int toplevel, double maxarea)
{
	if (maxarea <= 0)
		return;
	AirspacesDb::elementvector_t airspaces(m_ar.m_airspacedb.find_by_rect(m_bbox, 0, AirspacesDb::element_t::subtables_none));
	baselevel *= 100;
	toplevel *= 100;
	for (AirspacesDb::elementvector_t::const_iterator ai(airspaces.begin()), ae(airspaces.end()); ai != ae; ++ai) {
		const AirspacesDb::Airspace& aspc(*ai);
		{
			char bc(aspc.get_bdryclass());
			uint8_t tc(aspc.get_typecode());
			switch (tc) {
			default:
				// do not care about airspace classes E, F, G
				if (bc < 'A' || bc > 'D') {
					if (false) {
						std::ostringstream oss;
						oss << "VFR: exclude_airspace: skipping " << aspc.get_icao_name() << ' ' << aspc.get_class_string()
							  << " altitude: " << baselevel << "..." << toplevel << " due to class" << std::endl;
						m_ar.m_signal_log(log_debug0, oss.str());
					}
					continue;
				}
				break;

			case AirspacesDb::Airspace::typecode_specialuse:
				// Special Use airspaces; only care about prohibited, reserved and danger areas
				if (bc != 'P' && bc != 'R' && bc != 'D') {
					if (false) {
						std::ostringstream oss;
						oss << "VFR: exclude_airspace: skipping " << aspc.get_icao_name() << ' ' << aspc.get_class_string()
						    << " altitude: " << baselevel << "..." << toplevel << " due to class" << std::endl;
						m_ar.m_signal_log(log_debug0, oss.str());
					}
					continue;
				}
				break;

			case AirspacesDb::Airspace::typecode_ead:
				// EAD airspaces; only care about prohibited, reserved and danger areas
                                if (bc != AirspacesDb::Airspace::bdryclass_ead_d_amc && 
				    bc != AirspacesDb::Airspace::bdryclass_ead_d && 
				    bc != AirspacesDb::Airspace::bdryclass_ead_d_other && 
				    bc != AirspacesDb::Airspace::bdryclass_ead_p && 
				    bc != AirspacesDb::Airspace::bdryclass_ead_r_amc && 
				    bc != AirspacesDb::Airspace::bdryclass_ead_r) {
					if (false) {
						std::ostringstream oss;
						oss << "VFR: exclude_airspace: skipping " << aspc.get_icao_name() << ' ' << aspc.get_class_string()
						    << " altitude: " << baselevel << "..." << toplevel << " due to class" << std::endl;
						m_ar.m_signal_log(log_debug0, oss.str());
					}
					continue;
				}
				break;
			}
		}
		if (baselevel >= aspc.get_altupr_corr() || toplevel <= aspc.get_altlwr_corr()) {
			if (false) {
				std::ostringstream oss;
				oss << "VFR: exclude_airspace: skipping " << aspc.get_icao_name() << ' ' << aspc.get_class_string()
				    << " due to altitude: " << baselevel << "..." << toplevel << " fully outside "
				    << aspc.get_altlwr_corr() << "..." << aspc.get_altupr_corr() << std::endl;
				m_ar.m_signal_log(log_debug0, oss.str());
			}
			continue;
		}
		if (false && !(baselevel >= aspc.get_altlwr_corr() && toplevel <= aspc.get_altupr_corr())) {
			if (false) {
				std::ostringstream oss;
				oss << "VFR: exclude_airspace: skipping " << aspc.get_icao_name() << ' ' << aspc.get_class_string()
				    << " due to altitude: " << baselevel << "..." << toplevel << " not inside "
				    << aspc.get_altlwr_corr() << "..." << aspc.get_altupr_corr() << std::endl;
				m_ar.m_signal_log(log_debug0, oss.str());
			}
			continue;
		}
		Rect bbox(aspc.get_bbox());
		const MultiPolygonHole& poly(aspc.get_polygon());
		double area(fabs(poly.get_simple_area_nmi2_dbl()));
		if (area > maxarea) {
			if (false) {
				std::ostringstream oss;
				oss << "VFR: exclude_airspace: skipping " << aspc.get_icao_name() << ' ' << aspc.get_class_string()
				    << " due to area: " << area << "nmi^2" << std::endl;
				m_ar.m_signal_log(log_debug0, oss.str());
			}
			continue;
		}
		unsigned int killcnt(0);
		Graph::edge_iterator e0, e1;
		for (boost::tie(e0, e1) = boost::edges(m_graph); e0 != e1; ++e0) {
			Edge& ee(m_graph[*e0]);
			if (ee.get_metric() == std::numeric_limits<double>::max())
				continue;
			Graph::vertex_descriptor u(boost::source(*e0, m_graph));
			if (u == m_dep)
				continue;
			Graph::vertex_descriptor v(boost::target(*e0, m_graph));
			if (v == m_dest)
				continue;
			const Vertex& uu(m_graph[u]);
			const Vertex& vv(m_graph[v]);
			bool uui(bbox.is_inside(uu.get_coord()));
			bool vvi(bbox.is_inside(vv.get_coord()));
			{
				Rect r(uu.get_coord(), uu.get_coord());
				r.add(vv.get_coord());
				if (!r.is_intersect(bbox))
					continue;
			}
			if ((uui && poly.windingnumber(uu.get_coord())) ||
			    (vvi && poly.windingnumber(vv.get_coord())) ||
			    poly.is_intersection(uu.get_coord(), vv.get_coord())) {
				//ee.set_metric(std::numeric_limits<double>::max());
				ee.set_metric(ee.get_dist() * 1000);
				++killcnt;
				if (false) {
					std::ostringstream oss;
					oss << "VFR: airspace: kill edge " << uu.get_ident() << "->" << vv.get_ident() << std::endl;
					m_ar.m_signal_log(log_debug0, oss.str());
				}
			}
		}
		if (false || killcnt) {
			std::ostringstream oss;
			oss << "VFR: exclude_airspace: " << aspc.get_icao_name() << ' ' << aspc.get_class_string()
			    << " altitude: " << aspc.get_altlwr_corr() << "..." << aspc.get_altupr_corr()
			    << " area: " << area << "nmi^2: removed " << killcnt << " edges";
			m_ar.m_signal_log(log_debug0, oss.str());
		}
	}
}

bool CFMUAutoroute::VFRRouter::autoroute(FPlanRoute& route)
{
	while (route.get_nrwpt())
		route.delete_wpt(0);
	if (m_dep == m_dest) {
		if (true) {
			std::ostringstream oss;
			oss << "VFR: departure and destination are the same";
			m_ar.m_signal_log(log_debug0, oss.str());
		}
		return false;
	}
	std::vector<Graph::vertex_descriptor> predecessors;
	m_graph.shortest_paths_metric(m_dep, predecessors);
	{
		const Vertex& vv(m_graph[m_dest]);
		if (!vv.insert(route, 0)) {
			if (true) {
				std::ostringstream oss;
				oss << "VFR: cannot insert " << vv.get_ident() << " into fpl" ;
				m_ar.m_signal_log(log_debug0, oss.str());
			}
			return false;
		}
	}
	Graph::vertex_descriptor v(m_dest);
	while (v != m_dep) {
		Graph::vertex_descriptor u(predecessors[v]);
		if (u == v) {
			if (true) {
				std::ostringstream oss;
				oss << "VFR: no route found; "
				    << boost::out_degree(m_dep, m_graph) << " SID edges, "
				    << boost::in_degree(m_dest, m_graph) << " STAR edges";
				m_ar.m_signal_log(log_debug0, oss.str());
			}
			return false;
		}
		const Vertex& uu(m_graph[u]);
		const Vertex& vv(m_graph[v]);
		bool rteadded(false);
		if (v == m_dest && uu.is_match(Vertex::typemask_vfrreportingpt)) {
			for (sidstar_t::const_iterator si(m_star.begin()), se(m_star.end()); si != se; ++si) {
				const AirportsDb::Airport::VFRRoute& rte(*si);
				if (!rte.size() || rte.get_name() != uu.get_ident())
					continue;
				const AirportsDb::Airport::VFRRoute::VFRRoutePoint& rtept(rte[0]);
				if (rtept.get_pathcode() != AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_arrival)
					continue;
				AirportsDb::Airport arpt;
				m_graph[m_dest].get(arpt);
				for (unsigned int i = 0, j = rte.size() - 1U; i < j; ++i) {
					const AirportsDb::Airport::VFRRoute::VFRRoutePoint& rtept(rte[i]);
					FPlanWaypoint wpt;
					if (rtept.is_at_airport() && arpt.is_valid()) {
						Engine::DbObject::set(wpt, arpt);
					} else {
						if (arpt.is_valid())
							wpt.set_icao(arpt.get_icao());
						wpt.set_name(rtept.get_name());
						wpt.set_coord(rtept.get_coord());
						wpt.set_altitude(rtept.get_altitude());
					}
					wpt.set_pathcode(FPlanWaypoint::pathcode_vfrarrival);
					wpt.set_pathname(rte.get_name());
					route.insert_wpt(i, wpt);
				}
				rteadded = true;
				break;
			}
		}
		if (u == m_dep && vv.is_match(Vertex::typemask_vfrreportingpt)) {
			for (sidstar_t::const_iterator si(m_sid.begin()), se(m_sid.end()); si != se; ++si) {
				const AirportsDb::Airport::VFRRoute& rte(*si);
				if (!rte.size() || rte.get_name() != vv.get_ident())
					continue;
				const AirportsDb::Airport::VFRRoute::VFRRoutePoint& rtept(rte[rte.size()-1]);
				if (rtept.get_pathcode() != AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_departure)
					continue;
				AirportsDb::Airport arpt;
				m_graph[m_dep].get(arpt);
				for (unsigned int i = 0, j = rte.size() - 1U; i < j; ++i) {
					const AirportsDb::Airport::VFRRoute::VFRRoutePoint& rtept(rte[i]);
					FPlanWaypoint wpt;
					if (rtept.is_at_airport() && arpt.is_valid()) {
						Engine::DbObject::set(wpt, arpt);
					} else {
						if (arpt.is_valid())
							wpt.set_icao(arpt.get_icao());
						wpt.set_name(rtept.get_name());
						wpt.set_coord(rtept.get_coord());
						wpt.set_altitude(rtept.get_altitude());
					}
					wpt.set_pathcode(FPlanWaypoint::pathcode_vfrdeparture);
					wpt.set_pathname(rte.get_name());
					route.insert_wpt(i, wpt);
				}
				rteadded = true;
				break;
			}
		}
		if (!rteadded && uu.is_match(Vertex::typemask_vfrreportingpt)) {
			// FIXME!!
			std::ostringstream oss;
			oss << "VFR: do not know what to do with VFR arr/dep point " << uu.get_ident();
			m_ar.m_signal_log(log_debug0, oss.str());
			rteadded = true;
		}
		if (!rteadded && !uu.insert(route, 0)) {
			if (true) {
				std::ostringstream oss;
				oss << "VFR: cannot insert " << uu.get_ident() << " into fpl";
				m_ar.m_signal_log(log_debug0, oss.str());
			}
			return false;
		}
		v = u;
	}
	return true;
}

bool CFMUAutoroute::LSuperVertex::is_airport(void) const
{
	if (!m_obj)
		return false;
	return m_obj->is_airport();
}

bool CFMUAutoroute::LSuperVertex::is_navaid(void) const
{
	if (!m_obj)
		return false;
	return m_obj->is_navaid();
}

bool CFMUAutoroute::LSuperVertex::is_intersection(void) const
{
	if (!m_obj)
		return false;
	return m_obj->is_intersection();
}

bool CFMUAutoroute::LSuperVertex::is_mapelement(void) const
{
	if (!m_obj)
		return false;
	return m_obj->is_mapelement();
}

bool CFMUAutoroute::LSuperVertex::get(AirportsDb::Airport& el) const
{
	if (!m_obj)
		return false;
	return m_obj->get(el);
}

bool CFMUAutoroute::LSuperVertex::get(NavaidsDb::Navaid& el) const
{
	if (!m_obj)
		return false;
	return m_obj->get(el);
}

bool CFMUAutoroute::LSuperVertex::get(WaypointsDb::Waypoint& el) const
{
	if (!m_obj)
		return false;
	return m_obj->get(el);
}

bool CFMUAutoroute::LSuperVertex::get(MapelementsDb::Mapelement& el) const
{
	if (!m_obj)
		return false;
	return m_obj->get(el);
}

bool CFMUAutoroute::LSuperVertex::set(FPlanWaypoint& el) const
{
	if (!m_obj)
		return false;
	return m_obj->set(el);
}

unsigned int CFMUAutoroute::LSuperVertex::insert(FPlanRoute& route, uint32_t nr) const
{
	if (!m_obj)
		return false;
	return m_obj->insert(route, nr);
}

const Glib::ustring& CFMUAutoroute::LSuperVertex::get_ident(void) const
{
	if (get_object())
		return get_object()->get_ident();
	return Engine::DbObject::empty;
}

bool CFMUAutoroute::LSuperVertex::is_ident_valid(void) const
{
	const Glib::ustring& id(get_ident());
	return id.size() >= 2 && !Engine::AirwayGraphResult::Vertex::is_ident_numeric(id);
}

const double CFMUAutoroute::LEdge::maxmetric = std::numeric_limits<double>::max();

bool CFMUAutoroute::LEdge::is_match(lgraphairwayindex_t ident, lgraphairwayindex_t id)
{
	switch (id) {
	case airwaymatchall:
		return true;

	case airwaymatchnone:
		return false;

	case airwaymatchawy:
		return ident < airwaydct;

	case airwaymatchdctawy:
		return ident <= airwaydct;

	case airwaymatchdctawysidstar:
		return ident <= airwaysid;

	case airwaymatchdctawysidstarlevel:
		return ident <= airwaylevelup;

	case airwaymatchsidstar:
		return ident >= airwaystar && ident <= airwaysid;

	case airwaymatchlevel:
		return ident >= airwayleveldown && ident <= airwaylevelup;

	default:
		return ident == id;
	}
}

CFMUAutoroute::LVertexEdge::LVertexEdge(LGraph::edge_descriptor e, const LGraph& g)
{
	const LEdge& ee(g[e]);
	m_vertex = boost::target(e, g);
	m_edge = ee.get_ident();
}

bool CFMUAutoroute::LVertexEdge::operator<(const LVertexEdge& x) const
{
	if (get_vertex() < x.get_vertex())
		return true;
	if (get_vertex() > x.get_vertex())
		return false;
	return get_edge() < x.get_edge();
}

bool CFMUAutoroute::LRoute::operator<(const LRoute& x) const
{
	if (get_metric() < x.get_metric())
		return true;
	if (get_metric() > x.get_metric())
		return false;
	for (size_type i(0), n(std::min(size(), x.size())); i < n; ++i) {
		if ((*this)[i] < x[i])
			return true;
		if ((*this)[i] > x[i])
			return false;
	}
	return size() < x.size();
}

bool CFMUAutoroute::LRoute::is_repeated_nodes(const LGraph& g, const LSuperVertices& supervertices) const
{
	typedef std::set<unsigned int> nmap_t;
	nmap_t nmap;
	typedef std::set<Glib::ustring> imap_t;
	imap_t imap;
	for (const_iterator i(begin()), e(end()); i != e;) {
		const LVertex& v(g[i->get_vertex()]);
		unsigned int superindex0(v.get_superindex());
		{
			std::pair<nmap_t::iterator,bool> x(nmap.insert(superindex0));
			if (!x.second)
				return true;
		}
		if (!supervertices.empty()) {
			const LSuperVertex& super0(supervertices[superindex0]);
			std::pair<imap_t::iterator,bool> x(imap.insert(super0.get_ident()));
			if (!x.second)
				return true;
		}
		++i;
		while (i != e) {
			const LVertex& v1(g[i->get_vertex()]);
			if (v1.get_superindex() != superindex0)
				break;
			++i;
		}
		if (false) {
			if (i == e)
				break;
			if (!supervertices.empty()) {
				const LVertex& v1(g[i->get_vertex()]);
				const LSuperVertex& super0(supervertices[v.get_superindex()]);
				const LSuperVertex& super1(supervertices[v1.get_superindex()]);
				if (super0.get_ident() == super1.get_ident())
					return true;
			}
		}
	}
	return false;
}

bool CFMUAutoroute::LRoute::is_existing(const LGraph& g) const
{
	for (LRoute::size_type i(1); i < this->size(); ++i) {
		LGraph::edge_descriptor e;
		bool ok;
		boost::tie(e, ok) = this->get_edge(i, g);
		if (!ok)
			return false;
	}
	return true;
}

std::pair<CFMUAutoroute::LGraph::edge_descriptor,bool> CFMUAutoroute::LRoute::get_edge(const LVertexEdge& u, const LVertexEdge& v, const LGraph& g)
{
	return get_edge(u.get_vertex(), v, g);
}

std::pair<CFMUAutoroute::LGraph::edge_descriptor,bool> CFMUAutoroute::LRoute::get_edge(LGraph::vertex_descriptor u, const LVertexEdge& v, const LGraph& g)
{
	LGraph::out_edge_iterator e0, e1;
	for (boost::tie(e0, e1) = boost::out_edges(u, g); e0 != e1; ++e0) {
		if (boost::target(*e0, g) != v.get_vertex())
			continue;
		const LEdge& ee(g[*e0]);
		if (ee.get_ident() != v.get_edge())
			continue;
		return std::pair<LGraph::edge_descriptor,bool>(*e0, true);
	}
	return std::pair<LGraph::edge_descriptor,bool>(LGraph::edge_descriptor(), false);
}

std::pair<CFMUAutoroute::LGraph::edge_descriptor,bool> CFMUAutoroute::LRoute::get_edge(size_type i, const LGraph& g) const
{
	if (i < 1 || i >= size())
		return std::pair<LGraph::edge_descriptor,bool>(LGraph::edge_descriptor(), false);
	return get_edge((*this)[i - 1], (*this)[i], g);
}

CFMUAutoroute::LEdgeAdd::LEdgeAdd(LGraph::vertex_descriptor v0, LGraph::vertex_descriptor v1, double metric)
	: m_v0(v0), m_v1(v1), m_metric(metric)
{
}

bool CFMUAutoroute::LEdgeAdd::operator<(const LEdgeAdd& x) const
{
	if (get_v0() < x.get_v0())
		return true;
	if (get_v0() > x.get_v0())
		return false;
	return get_v1() < x.get_v1();
}

void CFMUAutoroute::clearlgraph(void)
{
	m_lgraph.clear();
	m_lsupervertices.clear();
	m_lgraphairwaytable.clear();
	m_lgraphairwayrevtable.clear();
	m_solutionvertices.clear();
	m_superdep = m_superdest = 0;
	lgraphmodified();
}

CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::lgraphairway2index(const Glib::ustring& awy, bool create)
{
	Glib::ustring awyu(awy.uppercase());
	if (awyu.empty())
		return airwaydct;
	if (create) {
		std::pair<lgraphairwayrevtable_t::iterator,bool> ins(m_lgraphairwayrevtable.insert(std::make_pair(awyu, m_lgraphairwaytable.size())));
		if (ins.second)
			m_lgraphairwaytable.push_back(awyu);
		return ins.first->second;
	}
	lgraphairwayrevtable_t::const_iterator i(m_lgraphairwayrevtable.find(awyu));
	if (i == m_lgraphairwayrevtable.end())
		return airwaymatchnone;
	return i->second;
}

const Glib::ustring& CFMUAutoroute::lgraphindex2airway(lgraphairwayindex_t idx, bool specialnames)
{
	switch (idx) {
	case airwaydct:
		if (specialnames) {
			static const Glib::ustring r("DCT");
			return r;
		}
		return Engine::DbObject::empty;

	case airwaysid:
		if (specialnames) {
			static const Glib::ustring r("SID");
			return r;
		}
		return Engine::DbObject::empty;

	case airwaystar:
		if (specialnames) {
			static const Glib::ustring r("STAR");
			return r;
		}
		return Engine::DbObject::empty;

	case airwaylevelup:
		if (specialnames) {
			static const Glib::ustring r("LEVEL+");
			return r;
		}
		return Engine::DbObject::empty;

	case airwayleveldown:
		if (specialnames) {
			static const Glib::ustring r("LEVEL-");
			return r;
		}
		return Engine::DbObject::empty;

	case airwaymatchall:
	{
		static const Glib::ustring r("[MATCHALL]");
		return r;
	}

	case airwaymatchnone:
	{
		static const Glib::ustring r("[MATCHALL]");
		return r;
	}

	case airwaymatchdctawy:
	{
		static const Glib::ustring r("[MATCHDCTAWY]");
		return r;
	}

	case airwaymatchdctawysidstar:
	{
		static const Glib::ustring r("[MATCHDCTAWYSIDSTAR]");
		return r;
	}

	case airwaymatchdctawysidstarlevel:
	{
		static const Glib::ustring r("[MATCHDCTAWYSIDSTARLEVEL]");
		return r;
	}

	case airwaymatchsidstar:
	{
		static const Glib::ustring r("[MATCHDCTAWYSIDSTAR]");
		return r;
	}

	case airwaymatchlevel:
	{
		static const Glib::ustring r("[MATCHDCTAWYLEVEL]");
		return r;
	}

	default:
		break;
	}
	if (idx < m_lgraphairwaytable.size())
		return m_lgraphairwaytable[idx];
	{
		static const Glib::ustring r("[??]");
		return r;
	}
}

unsigned int CFMUAutoroute::lgraphaddvertex(Engine::DbObject::ptr_t obj, bool supernode, bool levelchg)
{
	if (!obj)
		throw std::runtime_error("LGraph Vertex: object null");
	unsigned int nr(1);
	if (supernode)
		nr = m_performance.size();
	if (!nr)
		return ~0U;
	unsigned int idx(m_lsupervertices.size());
	LGraph::vertex_descriptor v0, v1;
	for (unsigned int i = 0; i < nr; ++i) {
		LGraph::vertex_descriptor vn(boost::add_vertex(LVertex(idx), m_lgraph));
		if (!i) {
			v0 = v1 = vn;
			continue;
		}
		if (vn != v1 + 1)
			throw std::runtime_error("LGraph Vertex: nonconsecutive descriptors");
		v1 = vn;
	}
	m_lsupervertices.push_back(LSuperVertex(obj, v0, v1));
	if (!supernode || !levelchg)
		return idx;
	for (unsigned int i = 1; i < nr; ++i) {
		LGraph::vertex_descriptor u0(v0 + (i - 1)), u1(v0 + i);
		const Performance& perf(m_performance[i - 1]);
		boost::add_edge(u0, u1, LEdge(airwaylevelup, is_optfuel() ? perf.get_fuelpenalty() : perf.get_timepenalty()), m_lgraph);
		boost::add_edge(u1, u0, LEdge(airwayleveldown, 0), m_lgraph);
	}
	return idx;
}

bool CFMUAutoroute::lgraphload(const Rect& bbox)
{
	typedef Engine::AirwayGraphResult::Graph Graph;
	typedef Engine::AirwayGraphResult::Vertex Vertex;
	typedef Engine::AirwayGraphResult::Edge Edge;

	Rect bbox1(bbox.oversize_nmi(100.f));
	if (true) {
		std::ostringstream oss;
		oss << "Loading Airway graph: [" << bbox1.get_southwest().get_lat_str2() << ' ' << bbox1.get_southwest().get_lon_str2()
		    << ' ' << bbox1.get_northeast().get_lat_str2() << ' ' << bbox1.get_northeast().get_lon_str2() << ']';
		m_signal_log(log_debug0, oss.str());
	}
	Graph g;
	opendb();
	g.clear();
	{
		NavaidsDb::elementvector_t ev(m_navaiddb.find_by_rect(bbox1, ~0, NavaidsDb::Navaid::subtables_none));
		for (NavaidsDb::elementvector_t::iterator i(ev.begin()); i != ev.end();) {
			switch (i->get_navaid_type()) {
			case NavaidsDb::Navaid::navaid_vor:
			case NavaidsDb::Navaid::navaid_vordme:
			case NavaidsDb::Navaid::navaid_ndb:
			case NavaidsDb::Navaid::navaid_ndbdme:
			case NavaidsDb::Navaid::navaid_vortac:
				++i;
				break;

			default:
				i = ev.erase(i);
				break;
			}
		}
		g.add(ev);
	}
	{
		WaypointsDb::elementvector_t ev(m_waypointdb.find_by_rect(bbox1, ~0, WaypointsDb::Waypoint::subtables_none));
		for (WaypointsDb::elementvector_t::iterator i(ev.begin()); i != ev.end();) {
			switch (i->get_type()) {
			case WaypointsDb::Waypoint::type_icao:
				++i;
				break;

			default:
				i = ev.erase(i);
				break;
			}
		}
		g.add(ev);
	}
	g.add(m_airwaydb.find_by_area(bbox1, ~0, AirwaysDb::Airway::subtables_all));
	if (true) {
		{
			std::ostringstream oss;
			oss << "Airway area graph: [" << bbox1.get_southwest().get_lat_str2() << ' ' << bbox1.get_southwest().get_lon_str2()
			    << ' ' << bbox1.get_northeast().get_lat_str2() << ' ' << bbox1.get_northeast().get_lon_str2() << "], "
			    << boost::num_vertices(g) << " vertices, " << boost::num_edges(g) << " edges";
			m_signal_log(log_debug0, oss.str());
		}
		unsigned int counts[4];
		for (unsigned int i = 0; i < sizeof(counts)/sizeof(counts[0]); ++i)
			counts[i] = 0;
		Graph::vertex_iterator u0, u1;
		for (boost::tie(u0, u1) = boost::vertices(g); u0 != u1; ++u0) {
			const Vertex& uu(g[*u0]);
			unsigned int i(uu.get_type());
			if (i < sizeof(counts)/sizeof(counts[0]))
				++counts[i];
		}
		{
			std::ostringstream oss;
			oss << "Airway area graph: " << counts[Vertex::type_airport] << " airports, " << counts[Vertex::type_navaid] << " navaids, "
			    << counts[Vertex::type_intersection] << " intersections, " << counts[Vertex::type_mapelement] << " mapelements";
			m_signal_log(log_debug0, oss.str());
		}
	}
	typedef std::map<Graph::vertex_descriptor,unsigned int> vmap_t;
	vmap_t vmap;
	// create vertices
	Graph::vertex_iterator u0, u1;
	for (boost::tie(u0, u1) = boost::vertices(g); u0 != u1; ++u0) {
		const Vertex& uu(g[*u0]);
		if (uu.get_type() == Vertex::type_intersection && uu.is_ident_numeric())
			continue;
		Engine::DbObject::ptr_t obj(uu.get_object());
		if (!obj) {
			{
				WaypointsDb::Waypoint el;
				el.set_icao("");
				el.set_name(uu.get_ident());
				el.set_coord(uu.get_coord());
                                el.set_usage(WaypointsDb::Waypoint::usage_user);
                                el.set_type(WaypointsDb::Waypoint::type_other);
				Engine::DbObject::ptr_t p(Engine::DbObject::create(el));
				obj.swap(p);
			}
			if (true) {
				std::ostringstream oss;
				oss << "lgraphload: creating artificial object for vertex " << uu.get_ident();
				m_signal_log(log_debug0, oss.str());
			}
		}
		if (false && !boost::out_degree(*u0, g) && !boost::in_degree(*u0, g))
			continue;
		unsigned int super(lgraphaddvertex(obj, true, true));
		vmap.insert(std::make_pair(*u0, super));
	}
	// create edges
	Graph::edge_iterator e0, e1;
	for (boost::tie(e0, e1) = boost::edges(g); e0 != e1; ++e0) {
		Edge& ee(g[*e0]);
		// FIXME: once we have better base and top level information, this heuristic should be removed
		{
			const Glib::ustring& id(ee.get_ident());
			if (id.size() >= 3 && (id[0] == 'U' || id[0] == 'u') && std::isalpha(id[1]) && std::isdigit(id[2]))
				continue;
		}
		unsigned int superidx0, superidx1;
		{
			vmap_t::const_iterator i(vmap.find(boost::source(*e0, g)));
			if (i == vmap.end())
				continue;
			superidx0 = i->second;
			i = vmap.find(boost::target(*e0, g));
			if (i == vmap.end())
				continue;
			superidx1 = i->second;
		}
		lgraphairwayindex_t awyname(lgraphairway2index(ee.get_ident(), true));
		const LSuperVertex& super0(m_lsupervertices[superidx0]);
		const LSuperVertex& super1(m_lsupervertices[superidx1]);
		if (superidx0 == superidx1 || super0.get_v0() == super1.get_v0()) {
			std::ostringstream oss;
			oss << "Routing Graph: refusing to add self-edge " << super0.get_ident();
			m_signal_log(log_debug0, oss.str());
			continue;
		}
		int minaltitude(std::numeric_limits<int>::min());
		int maxaltitude(std::numeric_limits<int>::max());
		if (ee.get_corridor5_elev() != AirwaysDb::Airway::nodata) {
			minaltitude = ee.get_corridor5_elev();
			if (minaltitude > 5000)
				minaltitude += 1000;
			minaltitude += 1000;
		}
		int mindctaltitude(minaltitude);
		if (get_honour_awy_levels()) {
			int mi(ee.get_base_level()), ma(ee.get_top_level());
			mi *= 100;
			ma *= 100;
			minaltitude = std::max(minaltitude, mi);
			maxaltitude = std::min(maxaltitude, ma);
		}
		TrafficFlowRestrictions::DctParameters dctrule(super0.get_ident(), 
							       super0.get_coord(), super1.get_ident(), super1.get_coord(),
							       mindctaltitude, m_performance[m_performance.size()-1].get_altitude()+1);
		bool dctcalc(!get_tfr_enabled());
		for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
			const Performance& perf(m_performance[pi]);
			if (perf.get_altitude() < mindctaltitude)
				continue;
			//double dist(ee.get_dist());
			double dist(super0.get_coord().spheric_distance_nmi_dbl(super1.get_coord()));
			double x(dist * perf.get_secpernmi());
			if (is_optfuel())
				x *= perf.get_fuelpersec();
			LGraph::vertex_descriptor v0(super0.get_v0() + pi);
			LGraph::vertex_descriptor v1(super1.get_v0() + pi);
			if (v0 > super0.get_v1() || v1 > super1.get_v1())
				throw std::runtime_error("CFMUAutoroute::lgraphload: vertex error");
			bool outsideawyvert(perf.get_altitude() < minaltitude || perf.get_altitude() > maxaltitude);
			if (outsideawyvert || awyname == airwaydct) {
				if (!lgraphawyvertdct)
					continue;
				if (!dctcalc) {
					std::vector<TrafficFlowRestrictions::Message> msg;
					m_tfr.check_dct(msg, dctrule);
					for (unsigned int i = 0; i < msg.size(); ++i)
						logmessage(msg[i]);
					dctcalc = true;
				}
			}
			if (outsideawyvert) {
				if (!dctrule.get_alt().is_inside(0, perf.get_altitude()))
					continue;
				LGraph::edge_descriptor e;
				bool ok;
				tie(e, ok) = boost::edge(v0, v1, m_lgraph);
				if (ok)
					continue;
				x *= get_dctpenalty();
				boost::add_edge(v0, v1, LEdge(airwaydct, x), m_lgraph);
				continue;
			}
			if (awyname == airwaydct) {
				if (!dctrule.get_alt().is_inside(0, perf.get_altitude()))
					continue;
				LGraph::edge_descriptor e;
				bool ok;
				tie(e, ok) = boost::edge(v0, v1, m_lgraph);
				if (ok)
					continue;
				x *= get_dctpenalty();
			}
			LGraph::out_edge_iterator e0, e1;
			for (boost::tie(e0, e1) = boost::out_edges(v0, m_lgraph); e0 != e1; ++e0) {
				if (boost::target(*e0, m_lgraph) != v1)
					continue;
				LEdge& ee(m_lgraph[*e0]);
				if (!ee.is_match(airwaydct))
					continue;
				ee.set_ident(awyname);
				ee.set_metric(x);
				break;
			}
			if (e0 == e1)
				boost::add_edge(v0, v1, LEdge(awyname, x), m_lgraph);
		}
	}
	{
		std::ostringstream oss;
		oss << "Routing Graph after airway population: " << boost::num_vertices(m_lgraph) << " V, " << boost::num_edges(m_lgraph) << " E";
		m_signal_log(log_debug0, oss.str());
	}
	lgraphmodified();
	return true;
}

void CFMUAutoroute::lgraphkillinvalidsuper(void)
{
	bool work(false);
	for (LSuperVertices::const_iterator si(m_lsupervertices.begin()), se(m_lsupervertices.end()); si != se; ++si) {
		const LSuperVertex& super(*si);
		if (super.is_ident_valid())
			continue;
		for (LGraph::vertex_descriptor v0(super.get_v0()); v0 <= super.get_v1(); ++v0) {
			work = work || boost::out_degree(v0, m_lgraph) || boost::in_degree(v0, m_lgraph);
			LGraph::in_edge_iterator e0, z0;
			for(boost::tie(e0, z0) = boost::in_edges(v0, m_lgraph); e0 != z0; ++e0) {
				if (boost::target(*e0, m_lgraph) != v0 || boost::source(*e0, m_lgraph) == v0)
					continue;
				const LEdge& ee0(m_lgraph[*e0]);
				if (!ee0.is_match(airwaymatchawy))
					continue;
				LGraph::out_edge_iterator e1, z1;
				for(boost::tie(e1, z1) = boost::out_edges(v0, m_lgraph); e1 != z1; ++e1) {
					if (boost::source(*e1, m_lgraph) != v0 || boost::target(*e1, m_lgraph) == v0)
						continue;
					const LEdge& ee1(m_lgraph[*e1]);
					if (!ee1.is_match(ee0.get_ident()))
						continue;
					if (boost::source(*e0, m_lgraph) == boost::target(*e1, m_lgraph))
						continue;
					boost::add_edge(boost::source(*e0, m_lgraph), boost::target(*e1, m_lgraph),
							LEdge(ee0.get_ident(), ee0.get_metric() + ee1.get_metric()), m_lgraph);
					if (true) {
						std::ostringstream oss;
						oss << "lgraphkillinvalidsuper: avoiding " << lgraphvertexname(v0)
						    << ": " << lgraphvertexname(boost::source(*e0, m_lgraph))
						    << ' ' << lgraphindex2airway(ee0.get_ident())
						    << ' ' << lgraphvertexname(boost::target(*e1, m_lgraph))
						    << " metric " << ee0.get_metric() + ee1.get_metric();
						m_signal_log(log_debug0, oss.str());
					}
				}
			}
			boost::clear_vertex(v0, m_lgraph);
		}
	}
	if (!work)
		return;
	{
		std::ostringstream oss;
		oss << "Routing Graph after invalid vertex removal: " << boost::num_vertices(m_lgraph) << " V, " << boost::num_edges(m_lgraph) << " E";
		m_signal_log(log_debug0, oss.str());
	}
	lgraphmodified();
}

#if 0

void CFMUAutoroute::lgraphadddct(void)
{
	if (get_dctlimit() <= 0)
		return;
	std::vector<std::vector<double> > awydist(boost::num_vertices(m_lgraph));
	for (LGraph::vertex_descriptor i = 0; i < boost::num_vertices(m_lgraph); ++i)
		awydist[i].resize(boost::num_vertices(m_lgraph), 0.0);
	if (!boost::johnson_all_pairs_shortest_paths(m_lgraph, awydist, boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)))) {
	//if (!boost::floyd_warshall_all_pairs_shortest_paths(m_lgraph, awydist, boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)))) {
		m_signal_log(log_debug0, "DCT: airway distances contain negative cycles, ignoring");
		for (LGraph::vertex_descriptor i = 0; i < boost::num_vertices(m_lgraph); ++i)
			for (LGraph::vertex_descriptor j = 0; j < boost::num_vertices(m_lgraph); ++j)
				awydist[i][j] = std::numeric_limits<double>::max();
	} else if (false) {
		for (LGraph::vertex_descriptor i = 0; i < boost::num_vertices(m_lgraph); ++i)
			for (LGraph::vertex_descriptor j = 0; j < boost::num_vertices(m_lgraph); ++j) {
				double d(awydist[i][j]);
				if (d == std::numeric_limits<double>::max())
					continue;
				const LVertex& v0(m_lgraph[i]);
				const LVertex& v1(m_lgraph[j]);
				const LSuperVertex& super0(m_lsupervertices[v0.get_superindex()]);
				const LSuperVertex& super1(m_lsupervertices[v1.get_superindex()]);
				std::ostringstream oss;
				oss << "DCT: AWY Distance " << super0.get_ident() << "->" << super1.get_ident()
				    << " metric " << std::setprecision(1);
				if (d < 1000.0)
					oss << std::fixed;
				oss << d << "nmi";
				m_signal_log(log_debug0, oss.str());
			}
	}
	typedef std::set<LEdgeAdd> edgeadd_t;
	edgeadd_t ea;
	unsigned int count_distpair(0);
	unsigned int count_awy(0);
	for (LSuperVertices::const_iterator s0i(m_lsupervertices.begin()), s0e(m_lsupervertices.end()); s0i != s0e; ++s0i) {
		const LSuperVertex& super0(*s0i);
		if (!super0.is_ident_valid())
			continue;
		for (LSuperVertices::const_iterator s1i(s0i + 1); s1i != s0e; ++s1i) {
			const LSuperVertex& super1(*s1i);
			if (!super1.is_ident_valid())
				continue;
			double dist(super0.get_coord().spheric_distance_nmi_dbl(super1.get_coord()));
			if (dist > get_dctlimit())
				continue;
			if (super0.get_ident() == super1.get_ident())
				continue;
			++count_distpair;
			int minalt(0);
			if (true) {
				TopoDb30::Profile prof(m_topodb.get_profile(super0.get_coord(), super1.get_coord(), 5));
                                if (prof.empty()) {
                                        std::ostringstream oss;
                                        oss << "DCT: Error computing ground clearance: " << super0.get_ident() << ' '
					    << super1.get_ident() << " dist " << std::setprecision(1) << dist << "nmi";
					m_signal_log(log_debug0, oss.str());
				} else {
					TopoDb30::elev_t el(prof.get_maxelev());
					if (el != TopoDb30::nodata) {
						minalt = prof.get_maxelev() * Point::m_to_ft;
						if (minalt >= 5000)
							minalt += 1000;
						minalt += 1000;
					}
				}
			}
			for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
				const Performance& perf(m_performance[pi]);
				if (perf.get_altitude() < minalt)
					continue;
				double x(dist * perf.get_secpernmi() * get_dctpenalty());
				if (is_optfuel())
					x *= perf.get_fuelpersec();
				LGraph::vertex_descriptor v0(super0.get_v0() + pi);
				LGraph::vertex_descriptor v1(super1.get_v0() + pi);
				if (v0 > super0.get_v1() || v1 > super1.get_v1())
					throw std::runtime_error("CFMUAutoroute::lgraphadddct: vertex error");
				// omit DCT if the airway distance is not (significantly) longer
				// i.e. prefer airways
				// recompute airway distance
				if (false && awydist[v0][v1] == std::numeric_limits<double>::max()) {
					std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph), 0);
					std::vector<double> distances(boost::num_vertices(m_lgraph), 0);
					dijkstra_shortest_paths(m_lgraph, v0, boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)).
								predecessor_map(&predecessors[0]).distance_map(&distances[0]));
					if (distances[v1] == std::numeric_limits<double>::max()) {
						double m(0);
						LGraph::vertex_descriptor v(v1);
						while (v != v0) {
							LGraph::vertex_descriptor u(predecessors[v]);
							if (u == v) {
								m = std::numeric_limits<double>::max();
								break;
							}
							LGraph::edge_descriptor e;
							bool ok(false);
							double bestm(std::numeric_limits<double>::max());
							LGraph::out_edge_iterator e0, e1;
							for (boost::tie(e0, e1) = boost::out_edges(u, m_lgraph); e0 != e1; ++e0) {
								if (boost::target(*e0, m_lgraph) != v)
									continue;
								const LEdge& ee(m_lgraph[*e0]);
								if (std::isnan(ee.get_metric()) || std::isinf(ee.get_metric()))
									continue;
								if (ee.get_metric() >= bestm)
									continue;
								bestm = ee.get_metric();
								ok = true;
								e = *e0;
							}
							if (!ok) {
								m = std::numeric_limits<double>::max();
								break;
							}
							m += bestm;
							v = u;
						}
						distances[v1] = m;
					}
					if (distances[v1] != std::numeric_limits<double>::max()) {
						awydist[v0][v1] = distances[v1];
						if (true) {
							Glib::ustring path;
							{
								LGraph::vertex_descriptor v(v1);
								while (v != v0) {
									LGraph::vertex_descriptor u(predecessors[v]);
									if (u == v) {
										path.clear();
										break;
									}
									path = "->" + lgraphvertexname(v) + path;
								}
								path = lgraphvertexname(v0) + path;
							}
							std::ostringstream oss;
							oss << "DCT: " << super0.get_ident() << "->" << super1.get_ident()
							    << " johnson metric error " << awydist[v0][v1] << " path " << path;
							m_signal_log(log_debug0, oss.str());
						}
					}
				}
				if (false && awydist[v1][v0] == std::numeric_limits<double>::max()) {
					std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph), 0);
					std::vector<double> distances(boost::num_vertices(m_lgraph), 0);
					dijkstra_shortest_paths(m_lgraph, v1, boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)).
								predecessor_map(&predecessors[0]).distance_map(&distances[0]));
					if (distances[v0] == std::numeric_limits<double>::max()) {
						double m(0);
						LGraph::vertex_descriptor v(v0);
						while (v != v1) {
							LGraph::vertex_descriptor u(predecessors[v]);
							if (u == v) {
								m = std::numeric_limits<double>::max();
								break;
							}
							LGraph::edge_descriptor e;
							bool ok(false);
							double bestm(std::numeric_limits<double>::max());
							LGraph::out_edge_iterator e0, e1;
							for (boost::tie(e0, e1) = boost::out_edges(u, m_lgraph); e0 != e1; ++e0) {
								if (boost::target(*e0, m_lgraph) != v)
									continue;
								const LEdge& ee(m_lgraph[*e0]);
								if (std::isnan(ee.get_metric()) || std::isinf(ee.get_metric()))
									continue;
								if (ee.get_metric() >= bestm)
									continue;
								bestm = ee.get_metric();
								ok = true;
								e = *e0;
							}
							if (!ok) {
								m = std::numeric_limits<double>::max();
								break;
							}
							m += bestm;
							v = u;
						}
						distances[v0] = m;
					}
					if (distances[v0] != std::numeric_limits<double>::max()) {
						awydist[v1][v0] = distances[v0];
						if (true) {
							Glib::ustring path;
							{
								LGraph::vertex_descriptor v(v0);
								while (v != v1) {
									LGraph::vertex_descriptor u(predecessors[v]);
									if (u == v) {
										path.clear();
										break;
									}
									path = "->" + lgraphvertexname(v) + path;
								}
								path = lgraphvertexname(v0) + path;
							}
							std::ostringstream oss;
							oss << "DCT: " << super1.get_ident() << "->" << super0.get_ident()
							    << " johnson metric error " << awydist[v1][v0] << " path " << path;
							m_signal_log(log_debug0, oss.str());
						}
					}
				}
				LGraph::edge_descriptor e;
				bool ok;
				tie(e, ok) = boost::edge(v0, v1, m_lgraph);
				if (!ok) {
					if (awydist[v0][v1] <= 1.01 * x)
						++count_awy;
				        else
						ea.insert(LEdgeAdd(v0, v1, x));
					if (false && count_distpair < 50 && pi + 1 == m_performance.size()) {
						std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph), 0);
						dijkstra_shortest_paths(m_lgraph, v0, boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)).
									predecessor_map(&predecessors[0]));
						double distd(0);
						LGraph::vertex_descriptor v(v1);
						while (v != v0) {
							LGraph::vertex_descriptor u(predecessors[v]);
							if (u == v) {
								distd = std::numeric_limits<double>::max();
								break;
							}
							LGraph::edge_descriptor e;
							bool ok(false);
							double bestm(std::numeric_limits<double>::max());
							LGraph::out_edge_iterator e0, e1;
							for (boost::tie(e0, e1) = boost::out_edges(u, m_lgraph); e0 != e1; ++e0) {
								if (boost::target(*e0, m_lgraph) != v)
									continue;
								const LEdge& ee(m_lgraph[*e0]);
								if (std::isnan(ee.get_metric()) || std::isinf(ee.get_metric()))
									continue;
								if (ee.get_metric() >= bestm)
									continue;
								bestm = ee.get_metric();
								ok = true;
								e = *e0;
							}
							if (!ok) {
								distd = std::numeric_limits<double>::max();
								break;
							}
							distd += bestm;
							v = u;
						}
						std::ostringstream oss;
						oss << "DCT: " << super0.get_ident() << "->" << super1.get_ident()
						    << " F" << ((perf.get_altitude() + 50) / 100) << ": DCT D "
						    << std::setprecision(1) << dist << " M "
						    << std::setprecision(1) << x
						    << " AWY D "
						    << std::setprecision(1) << distd << " M "
						    << std::setprecision(1) << awydist[v0][v1];
						m_signal_log(log_debug0, oss.str());
					}
				}
				tie(e, ok) = boost::edge(v1, v0, m_lgraph);
				if (!ok) {
					if (awydist[v1][v0] <= 1.01 * x)
						++count_awy;
					else
						ea.insert(LEdgeAdd(v1, v0, x));
					if (false && count_distpair < 50 && pi + 1 == m_performance.size()) {
						std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph), 0);
						dijkstra_shortest_paths(m_lgraph, v1, boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)).
									predecessor_map(&predecessors[0]));
						double distd(0);
						LGraph::vertex_descriptor v(v0);
						while (v != v1) {
							LGraph::vertex_descriptor u(predecessors[v]);
							if (u == v) {
								distd = std::numeric_limits<double>::max();
								break;
							}
							LGraph::edge_descriptor e;
							bool ok(false);
							double bestm(std::numeric_limits<double>::max());
							LGraph::out_edge_iterator e0, e1;
							for (boost::tie(e0, e1) = boost::out_edges(u, m_lgraph); e0 != e1; ++e0) {
								if (boost::target(*e0, m_lgraph) != v)
									continue;
								const LEdge& ee(m_lgraph[*e0]);
								if (std::isnan(ee.get_metric()) || std::isinf(ee.get_metric()))
									continue;
								if (ee.get_metric() >= bestm)
									continue;
								bestm = ee.get_metric();
								ok = true;
								e = *e0;
							}
							if (!ok) {
								distd = std::numeric_limits<double>::max();
								break;
							}
							distd += bestm;
							v = u;
						}
						std::ostringstream oss;
						oss << "DCT: " << super1.get_ident() << "->" << super0.get_ident()
						    << " F" << ((perf.get_altitude() + 50) / 100) << ": DCT D "
						    << std::setprecision(1) << dist << " M "
						    << std::setprecision(1) << x << " AWY D "
						    << std::setprecision(1) << distd << " M "
						    << std::setprecision(1) << awydist[v1][v0];
						m_signal_log(log_debug0, oss.str());
					}
				}
			}
		}
	}
	unsigned int count_dct(0);
	for (edgeadd_t::const_iterator ie(ea.begin()), ee(ea.end()); ie != ee; ++ie) {
		LGraph::edge_descriptor e;
		bool ok;
		tie(e, ok) = boost::add_edge(ie->get_v0(), ie->get_v1(), LEdge(airwaydct, ie->get_metric()), m_lgraph);
		if (ok)
			++count_dct;
	}
	{
		std::ostringstream oss;
		oss << "Routing Graph after DCT population: " << boost::num_vertices(m_lgraph) << " V, " << boost::num_edges(m_lgraph) << " E, "
		    << count_distpair << " vertex pairs, " << count_dct << " DCT edges, " << count_awy << " AWY edges preferred";
		m_signal_log(log_debug0, oss.str());
	}
	lgraphmodified();
}

#elif 0

template <class T>
struct DijkstraCombineLimit
{
	const T m_limit;
	const T m_inf;

	DijkstraCombineLimit() : m_limit(std::numeric_limits<T>::max()), m_inf(std::numeric_limits<T>::max()) { }
	DijkstraCombineLimit(T lim) : m_limit(lim), m_inf(std::numeric_limits<T>::max()) { }
	DijkstraCombineLimit(T lim, T inf) : m_limit(lim), m_inf(inf) { }

	T operator()(const T& a, const T& b) const
	{
		if (a == m_inf)
			return m_inf;
		if (b == m_inf)
			return m_inf;
		double z(a + b);
		if (z > m_limit)
			return m_inf;
		return z;
	}
};

void CFMUAutoroute::lgraphadddct(void)
{
	if (get_dctlimit() <= 0)
		return;
	typedef std::set<LEdgeAdd> edgeadd_t;
	edgeadd_t ea;
	unsigned int count_distpair(0);
	unsigned int count_awy(0);
	for (LSuperVertices::const_iterator s0i(m_lsupervertices.begin()), s0e(m_lsupervertices.end()); s0i != s0e; ++s0i) {
		const LSuperVertex& super0(*s0i);
		if (!super0.is_ident_valid())
			continue;
		for (LSuperVertices::const_iterator s1i(s0i + 1); s1i != s0e; ++s1i) {
			const LSuperVertex& super1(*s1i);
			if (!super1.is_ident_valid())
				continue;
			double dist(super0.get_coord().spheric_distance_nmi_dbl(super1.get_coord()));
			if (dist > get_dctlimit())
				continue;
			if (super0.get_ident() == super1.get_ident())
				continue;
			++count_distpair;
			if (true && !(count_distpair & ((1 << 10) - 1))) {
				std::ostringstream oss;
				oss << "DCT: processed " << count_distpair << " pairs";
				m_signal_log(log_debug0, oss.str());
			}
			int minalt(0);
			if (true) {
				TopoDb30::Profile prof(m_topodb.get_profile(super0.get_coord(), super1.get_coord(), 5));
                                if (prof.empty()) {
                                        std::ostringstream oss;
                                        oss << "DCT: Error computing ground clearance: " << super0.get_ident() << ' '
					    << super1.get_ident() << " dist " << std::setprecision(1) << dist << "nmi";
					m_signal_log(log_debug0, oss.str());
				} else {
					TopoDb30::elev_t el(prof.get_maxelev());
					if (el != TopoDb30::nodata) {
						minalt = prof.get_maxelev() * Point::m_to_ft;
						if (minalt >= 5000)
							minalt += 1000;
						minalt += 1000;
					}
				}
			}
			for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
				const Performance& perf(m_performance[pi]);
				if (perf.get_altitude() < minalt)
					continue;
				double x(dist * perf.get_secpernmi() * get_dctpenalty());
				if (is_optfuel())
					x *= perf.get_fuelpersec();
				LGraph::vertex_descriptor v0(super0.get_v0() + pi);
				LGraph::vertex_descriptor v1(super1.get_v0() + pi);
				if (v0 > super0.get_v1() || v1 > super1.get_v1())
					throw std::runtime_error("CFMUAutoroute::lgraphadddct: vertex error");
				// omit DCT if the airway distance is not (significantly) longer
				// i.e. prefer airways
				LGraph::edge_descriptor e;
				bool ok;
				tie(e, ok) = boost::edge(v0, v1, m_lgraph);
				if (!ok) {
					std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph));
					std::vector<double> distances(boost::num_vertices(m_lgraph));
					DijkstraCombineLimit<double> cmb(1.1 * x);
					dijkstra_shortest_paths(m_lgraph, v0, boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)).
								predecessor_map(&predecessors[0]).distance_map(&distances[0]).
								distance_combine(cmb));
					if (distances[v1] <= 1.01 * x)
						++count_awy;
				        else
						ea.insert(LEdgeAdd(v0, v1, x));
				}
				tie(e, ok) = boost::edge(v1, v0, m_lgraph);
				if (!ok) {
					std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph));
					std::vector<double> distances(boost::num_vertices(m_lgraph));
					DijkstraCombineLimit<double> cmb(1.1 * x);
					dijkstra_shortest_paths(m_lgraph, v1, boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)).
								predecessor_map(&predecessors[0]).distance_map(&distances[0]).
								distance_combine(cmb));
					if (distances[v0] <= 1.01 * x)
						++count_awy;
					else
						ea.insert(LEdgeAdd(v1, v0, x));
				}
			}
		}
	}
	unsigned int count_dct(0);
	for (edgeadd_t::const_iterator ie(ea.begin()), ee(ea.end()); ie != ee; ++ie) {
		LGraph::edge_descriptor e;
		bool ok;
		tie(e, ok) = boost::add_edge(ie->get_v0(), ie->get_v1(), LEdge(airwaydct, ie->get_metric()), m_lgraph);
		if (ok)
			++count_dct;
	}
	{
		std::ostringstream oss;
		oss << "Routing Graph after DCT population: " << boost::num_vertices(m_lgraph) << " V, " << boost::num_edges(m_lgraph) << " E, "
		    << count_distpair << " vertex pairs, " << count_dct << " DCT edges, " << count_awy << " AWY edges preferred";
		m_signal_log(log_debug0, oss.str());
	}
	lgraphmodified();
}

#else

void CFMUAutoroute::lgraphadddct(void)
{
	if (get_dctlimit() <= 0)
		return;
	//const unsigned int awydistsz(m_lsupervertices.size());
	//double awydist[awydistsz][awydistsz];
	std::vector<std::vector<double> > awydist(m_lsupervertices.size());
	for (LGraph::vertex_descriptor i = 0; i < m_lsupervertices.size(); ++i)
		awydist[i].resize(m_lsupervertices.size(), 0.0);
	{
		typedef boost::adjacency_list<boost::listS, boost::vecS, boost::directedS, boost::no_property, boost::property<boost::edge_weight_t, double> > DGraph;
		DGraph dg(m_lsupervertices.size());
		boost::property_map<DGraph, boost::edge_weight_t>::type dgdistmap(boost::get(boost::edge_weight, dg));
		{
			LGraph::edge_iterator e0, e1;
			for (boost::tie(e0, e1) = boost::edges(m_lgraph); e0 != e1; ++e0) {
				const LVertex& uu(m_lgraph[boost::source(*e0, m_lgraph)]);
				const LVertex& vv(m_lgraph[boost::target(*e0, m_lgraph)]);
				DGraph::edge_descriptor e;
				bool ok;
				boost::tie(e, ok) = add_edge((DGraph::vertex_descriptor)uu.get_superindex(), (DGraph::vertex_descriptor)vv.get_superindex(), dg);
				if (!ok)
					continue;
				const LSuperVertex& superu(m_lsupervertices[uu.get_superindex()]);
				const LSuperVertex& superv(m_lsupervertices[vv.get_superindex()]);
				dgdistmap[e] = superu.get_coord().spheric_distance_nmi_dbl(superv.get_coord());
			}
		}
		if (!boost::johnson_all_pairs_shortest_paths(dg, awydist)) {
			m_signal_log(log_debug0, "DCT: airway distances contain negative cycles, ignoring");
			for (LGraph::vertex_descriptor i = 0; i < m_lsupervertices.size(); ++i)
				for (LGraph::vertex_descriptor j = 0; j < m_lsupervertices.size(); ++j)
					awydist[i][j] = std::numeric_limits<double>::max();
		}
	}
	unsigned int count_distpair(0);
	unsigned int count_awy(0);
	unsigned int count_dct(0);
	unsigned int count_dctrules(0);
	for (unsigned int superidx0(0), nrsuper(m_lsupervertices.size()); superidx0 < nrsuper; ++superidx0) {
		const LSuperVertex& super0(m_lsupervertices[superidx0]);
		if (!super0.is_ident_valid())
			continue;
		for (unsigned int superidx1(superidx0 + 1); superidx1 < nrsuper; ++superidx1) {
			const LSuperVertex& super1(m_lsupervertices[superidx1]);
			if (!super1.is_ident_valid())
				continue;
			double dist(super0.get_coord().spheric_distance_nmi_dbl(super1.get_coord()));
			if (dist > get_dctlimit())
				continue;
			if (super0.get_ident() == super1.get_ident())
				continue;
			++count_distpair;
			if (true && !(count_distpair & ((1 << 12) - 1))) {
				std::ostringstream oss;
				oss << "DCT: processed " << count_distpair << " pairs";
				m_signal_log(log_debug0, oss.str());
			}
			if (dist * 1.01 > std::max(awydist[superidx0][superidx1], awydist[superidx1][superidx0]))
				continue;
			int minalt(0);
			TrafficFlowRestrictions::DctParameters dctrule(super0.get_ident(), super0.get_coord(), super1.get_ident(), super1.get_coord(),
								       m_performance[0].get_altitude(), m_performance[m_performance.size()-1].get_altitude()+1);
			if (true && get_tfr_enabled()) {
				std::vector<TrafficFlowRestrictions::Message> msg;
				m_tfr.check_dct(msg, dctrule, false && (super0.get_ident() == "BERSU" || super1.get_ident() == "BERSU"));
				for (unsigned int i = 0; i < msg.size(); ++i)
					logmessage(msg[i]);
				if (false && (super0.get_ident() == "ZUE" || super0.get_ident() == "ROMIR" || super0.get_ident() == "KUDIS" ||
					     super1.get_ident() == "ZUE" || super1.get_ident() == "ROMIR" || super1.get_ident() == "KUDIS")) {
					std::ostringstream oss;
                                        oss << "DCT: rules: " << super0.get_ident() << ' ' << super1.get_ident()
					    << " alt " << dctrule.get_alt(0).to_str() << " / " << dctrule.get_alt(1).to_str();
					if (dctrule.get_alt().is_empty())
						oss << " (E)";
					m_signal_log(log_debug0, oss.str());
				}
			}
			if (dctrule.get_alt().is_empty()) {
				count_dctrules += 2 * m_performance.size();
				continue;
			}
			if (true) {
				TopoDb30::Profile prof(m_topodb.get_profile(super0.get_coord(), super1.get_coord(), 5));
                                if (prof.empty()) {
                                        std::ostringstream oss;
                                        oss << "DCT: Error computing ground clearance: " << super0.get_ident() << ' '
					    << super1.get_ident() << " dist " << std::setprecision(1) << dist << "nmi";
					m_signal_log(log_debug0, oss.str());
				} else {
					TopoDb30::elev_t el(prof.get_maxelev());
					if (el != TopoDb30::nodata) {
						minalt = prof.get_maxelev() * Point::m_to_ft;
						if (minalt >= 5000)
							minalt += 1000;
						minalt += 1000;
					}
				}
			}
			for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
				const Performance& perf(m_performance[pi]);
				if (perf.get_altitude() < minalt)
					continue;
				double x(dist * perf.get_secpernmi() * get_dctpenalty());
				if (is_optfuel())
					x *= perf.get_fuelpersec();
				LGraph::vertex_descriptor v0(super0.get_v0() + pi);
				LGraph::vertex_descriptor v1(super1.get_v0() + pi);
				if (v0 > super0.get_v1() || v1 > super1.get_v1())
					throw std::runtime_error("CFMUAutoroute::lgraphadddct: vertex error");
				if (false && ((super0.get_ident() == "BERSU" && super1.get_ident() == "WIL") ||
					      (super0.get_ident() == "WIL" && super1.get_ident() == "BERSU"))) {
					std::ostringstream oss;
					oss << "DCT: " << super0.get_ident() << ' ' << super1.get_ident()
					    << " alt " << perf.get_altitude() << " dist " << dist << " awy "
					    << awydist[superidx0][superidx1] << " / " << awydist[superidx1][superidx0]
					    << " dctrule " << (int)dctrule.get_alt().is_inside(0, perf.get_altitude())
					    << " / " << (int)dctrule.get_alt().is_inside(1, perf.get_altitude())
					    << " metric " << x;
					m_signal_log(log_debug0, oss.str());
				}
				// omit DCT if the airway distance is not (significantly) longer
				// i.e. prefer airways
				LGraph::edge_descriptor e;
				bool ok;
				if (dctrule.get_alt().is_inside(0, perf.get_altitude())) {
					tie(e, ok) = boost::edge(v0, v1, m_lgraph);
					if (!ok) {
						if (awydist[superidx0][superidx1] <= 1.01 * dist) {
							++count_awy;
						} else {
							tie(e, ok) = boost::add_edge(v0, v1, LEdge(airwaydct, x), m_lgraph);
							if (ok)
								++count_dct;
						}
					}
				} else {
					++count_dctrules;
				}
				if (dctrule.get_alt().is_inside(1, perf.get_altitude())) {
					tie(e, ok) = boost::edge(v1, v0, m_lgraph);
					if (!ok) {
						if (awydist[superidx1][superidx0] <= 1.01 * dist) {
							++count_awy;
						} else {
							tie(e, ok) = boost::add_edge(v1, v0, LEdge(airwaydct, x), m_lgraph);
							if (ok)
								++count_dct;
						}
					}
				} else {
					++count_dctrules;
				}
			}
		}
	}
	{
		std::ostringstream oss;
		oss << "Routing Graph after DCT population: " << boost::num_vertices(m_lgraph) << " V, " << boost::num_edges(m_lgraph) << " E, "
		    << count_distpair << " vertex pairs, " << count_dct << " DCT edges, " << count_awy << " AWY edges preferred, "
		    << count_dctrules << " DCT edges prevented by TFR rules";
		m_signal_log(log_debug0, oss.str());
	}
	lgraphmodified();
}

#endif

bool CFMUAutoroute::lgraphaddsidstar(void)
{
	if (!get_departure().is_valid() || !get_destination().is_valid())
		return false;
	static const bool duplicate_sidstar_nodes = true;
	unsigned int lastsuper(m_lsupervertices.size());
	m_superdep = lgraphaddvertex(Engine::DbObject::create(get_departure()), false, false);
	m_superdest = lgraphaddvertex(Engine::DbObject::create(get_destination()), false, false);
	const LSuperVertex& sdep(m_lsupervertices[m_superdep]);
	const LSuperVertex& sdest(m_lsupervertices[m_superdest]);
	typedef std::vector<unsigned int> sv_t;
	sv_t sidv, starv;
	{
		unsigned int sidsv(lastsuper), starsv(lastsuper);
		double sidd(std::numeric_limits<double>::max()), stard(sidd);
		sv_t sida, stara;
		for (unsigned int s0i(0); s0i < lastsuper; ++s0i) {
			const LSuperVertex& super0(m_lsupervertices[s0i]);
			double dist(super0.get_coord().spheric_distance_nmi_dbl(sdep.get_coord()));
			if (dist <= get_sidlimit())
				sidv.push_back(s0i);
			dist = super0.get_coord().spheric_distance_nmi_dbl(sdest.get_coord());
			if (dist <= get_starlimit())
				starv.push_back(s0i);
			if (!get_sid().is_invalid()) {
				dist = super0.get_coord().spheric_distance_nmi_dbl(get_sid());
				if (dist < sidd) {
					sidd = dist;
					sidsv = s0i;
				}
			}
			if (!get_star().is_invalid()) {
				dist = super0.get_coord().spheric_distance_nmi_dbl(get_star());
				if (dist < stard) {
					stard = dist;
					starsv = s0i;
				}
			}
			for (unsigned int i = 0, j = get_departure().get_nrvfrrte(); i < j; ++i) {
				const AirportsDb::Airport::VFRRoute& rte(get_departure().get_vfrrte(i));
				if (!rte.size())
					continue;
				const AirportsDb::Airport::VFRRoute::VFRRoutePoint& rtept(rte[rte.size()-1]);
				if (rtept.get_pathcode() != AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_sid)
					continue;
				if (super0.get_coord().spheric_distance_nmi_dbl(rtept.get_coord()) > 0.1)
					continue;
				sida.push_back(s0i);
				break;
			}
			for (unsigned int i = 0, j = get_destination().get_nrvfrrte(); i < j; ++i) {
				const AirportsDb::Airport::VFRRoute& rte(get_destination().get_vfrrte(i));
				if (!rte.size())
					continue;
				const AirportsDb::Airport::VFRRoute::VFRRoutePoint& rtept(rte[0]);
				if (rtept.get_pathcode() != AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_star)
					continue;
				if (super0.get_coord().spheric_distance_nmi_dbl(rtept.get_coord()) > 0.1)
					continue;
				stara.push_back(s0i);
				break;
			}
		}
		if (!sida.empty())
			sida.swap(sidv);
		if (!stara.empty())
			stara.swap(starv);
		if (sidd > 0.1)
			sidsv = lastsuper;
		if (stard > 0.1)
			starsv = lastsuper;
		if (sidsv != lastsuper) {
			sidv.clear();
			sidv.push_back(sidsv);
		} else if (!get_sid().is_invalid()) {
			if (true) {
				std::ostringstream oss;
				oss << "IFR: SID point " << get_sid().get_lat_str2() << ' ' << get_sid().get_lon_str2() << " not found";
				m_signal_log(log_debug0, oss.str());
			}
			return false;
		}
		if (starsv != lastsuper) {
			starv.clear();
			starv.push_back(starsv);
		} else if (!get_star().is_invalid()) {
			if (true) {
				std::ostringstream oss;
				oss << "IFR: STAR point " << get_star().get_lat_str2() << ' ' << get_star().get_lon_str2() << " not found";
				m_signal_log(log_debug0, oss.str());
			}
			return false;
		}
	}
	typedef std::map<unsigned int, unsigned int> nodemap_t;
	nodemap_t nodemap;
	if (duplicate_sidstar_nodes) {
		for (sv_t::const_iterator svi(sidv.begin()), sve(sidv.end()); svi != sve; ++svi) {
			unsigned int superidx(*svi);
			if (nodemap.find(superidx) != nodemap.end())
				continue;
			const LSuperVertex& super(m_lsupervertices[superidx]);
			unsigned int newsuperidx(lgraphaddvertex(super.get_object(), true, false));
			nodemap.insert(nodemap_t::value_type(superidx, newsuperidx));
		}
		for (sv_t::const_iterator svi(starv.begin()), sve(starv.end()); svi != sve; ++svi) {
			unsigned int superidx(*svi);
			if (nodemap.find(superidx) != nodemap.end())
				continue;
			const LSuperVertex& super(m_lsupervertices[superidx]);
			unsigned int newsuperidx(lgraphaddvertex(super.get_object(), true, false));
			nodemap.insert(nodemap_t::value_type(superidx, newsuperidx));
		}
		for (nodemap_t::const_iterator ni1(nodemap.begin()), ne(nodemap.end()); ni1 != ne; ++ni1) {
			const LSuperVertex& super(m_lsupervertices[ni1->first]);
			const LSuperVertex& super0(m_lsupervertices[ni1->second]);
			for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
				LGraph::out_edge_iterator e0, e1;
				for (boost::tie(e0, e1) = boost::out_edges(super.get_v0() + pi, m_lgraph); e0 != e1; ++e0) {
					const LEdge& e(m_lgraph[*e0]);
					if (e.is_match(airwaymatchlevel))
						continue;
					const LVertex& vv1(m_lgraph[boost::target(*e0, m_lgraph)]);
					if (vv1.get_superindex() == ni1->second)
						continue;
					nodemap_t::const_iterator ni2(nodemap.find(vv1.get_superindex()));
					if (ni2 == ne || ni2 == ni1)
						continue;
					const LSuperVertex& super1(m_lsupervertices[ni2->second]);
					if (super1.get_v0() == super1.get_v1())
						continue;
					boost::add_edge(super0.get_v0() + pi, super1.get_v0() + pi, e, m_lgraph);
					if (true) {
						std::ostringstream oss;
						oss << "SID/STAR: add direct route " << lgraphvertexname(super0.get_v0() + pi) << ' '
						    << lgraphindex2airway(e.get_ident(), true) << ' ' << lgraphvertexname(super1.get_v0() + pi);
						m_signal_log(log_debug0, oss.str());
					}
				}
			}
		}
	}
	for (sv_t::const_iterator svi(sidv.begin()), sve(sidv.end()); svi != sve; ++svi) {
		unsigned int superidx(*svi);
		{
			nodemap_t::const_iterator i(nodemap.find(superidx));
			if (i != nodemap.end()) {
				const LSuperVertex& super(m_lsupervertices[superidx]);
				unsigned int newsuperidx(i->second);
				const LSuperVertex& newsuper(m_lsupervertices[newsuperidx]);
				for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
					LGraph::out_edge_iterator e0, e1;
					for (boost::tie(e0, e1) = boost::out_edges(super.get_v0() + pi, m_lgraph); e0 != e1; ++e0) {
						const LEdge& e(m_lgraph[*e0]);
						if (e.is_match(airwaymatchlevel))
							continue;
						boost::add_edge(newsuper.get_v0() + pi, boost::target(*e0, m_lgraph), e, m_lgraph);
					}
				}
				superidx = newsuperidx;
			}
		}
		const LSuperVertex& super(m_lsupervertices[superidx]);
		const Performance& perf(m_performance[0]);
		double x(super.get_coord().spheric_distance_nmi_dbl(sdep.get_coord()) * perf.get_secpernmi());
		if (is_optfuel())
			x *= perf.get_fuelpersec();
		for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
			const Performance& perf(m_performance[pi]);
			boost::add_edge(sdep.get_v0(), super.get_v0() + pi, LEdge(airwaysid, x), m_lgraph);
			if (is_optfuel())
				x += perf.get_fuelpenalty();
			else
				x += perf.get_timepenalty();
		}
		lgraphkilllevelchg(super);
	}
	for (sv_t::const_iterator svi(starv.begin()), sve(starv.end()); svi != sve; ++svi) {
		unsigned int superidx(*svi);
		{
			nodemap_t::const_iterator i(nodemap.find(superidx));
			if (i != nodemap.end()) {
				const LSuperVertex& super(m_lsupervertices[superidx]);
				unsigned int newsuperidx(i->second);
				const LSuperVertex& newsuper(m_lsupervertices[newsuperidx]);
				for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
					LGraph::in_edge_iterator e0, e1;
					for (boost::tie(e0, e1) = boost::in_edges(super.get_v0() + pi, m_lgraph); e0 != e1; ++e0) {
						const LEdge& e(m_lgraph[*e0]);
						if (e.is_match(airwaymatchlevel))
							continue;
						boost::add_edge(boost::source(*e0, m_lgraph), newsuper.get_v0() + pi, e, m_lgraph);
					}
				}
				superidx = newsuperidx;
			}
		}
		const LSuperVertex& super(m_lsupervertices[superidx]);
		const Performance& perf(m_performance[0]);
		double x(super.get_coord().spheric_distance_nmi_dbl(sdest.get_coord()) * perf.get_secpernmi());
		if (is_optfuel())
			x *= perf.get_fuelpersec();
		for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
			const Performance& perf(m_performance[pi]);
			boost::add_edge(super.get_v0() + pi, sdest.get_v0(), LEdge(airwaystar, x), m_lgraph);
			if (is_optfuel())
				x += perf.get_fuelpenalty();
			else
				x += perf.get_timepenalty();
		}
		lgraphkilllevelchg(super);
	}
	{
		std::ostringstream oss;
		oss << "Routing Graph after SID/STAR population: " << boost::num_vertices(m_lgraph) << " V, " << boost::num_edges(m_lgraph) << " E";
		m_signal_log(log_debug0, oss.str());
	}
	lgraphmodified();
	return true;
}

void CFMUAutoroute::lgraphupdatemetric(LRoute& r)
{
	if (r.size() < 2) {
		r.set_metric(std::numeric_limits<double>::quiet_NaN());
		return;
	}
	double m(0);
	for (LRoute::size_type i(1); i < r.size(); ++i) {
		LGraph::edge_descriptor e;
		bool ok;
		tie(e, ok) = r.get_edge(i, m_lgraph);
		if (!ok) {
			r.set_metric(std::numeric_limits<double>::infinity());
			const LVertex& v0(m_lgraph[r[i-1].get_vertex()]);
			const LVertex& v1(m_lgraph[r[i].get_vertex()]);
			const LSuperVertex& sv0(m_lsupervertices[v0.get_superindex()]);
			const LSuperVertex& sv1(m_lsupervertices[v1.get_superindex()]);
			std::ostringstream oss;
			oss << "Route Update Metric: Leg " << sv0.get_ident()
			    << "--" << lgraphindex2airway(r[i].get_edge(), true) << "->"
			    << sv1.get_ident() << " does not exist";
			m_signal_log(log_debug0, oss.str());
			return;
		}
		const LEdge& ee(m_lgraph[e]);
		m += ee.get_metric();
	}
	r.set_metric(m);
}

void CFMUAutoroute::lgraphclearcurrent(void)
{
	while (m_route.get_nrwpt())
		m_route.delete_wpt(0);
	m_routetime = 0;
	m_routefuel = 0;
	m_validationresponse.clear();
	m_solutionvertices.clear();
	{
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1; ++e0) {
			LEdge& ee(m_lgraph[*e0]);
			ee.set_solution(false);
		}
	}
}

bool CFMUAutoroute::lgraphsetcurrent(const LRoute& r)
{
	lgraphclearcurrent();
	if (r.size() < 2)
		return false;
	{
		LGraph::vertex_descriptor v(r.back().get_vertex());
		m_solutionvertices.insert(v);
		{
			const LVertex& vv(m_lgraph[v]);
			const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
			if (!sv.insert(m_route, 0)) {
				if (true) {
					std::ostringstream oss;
					oss << "Cannot insert " << sv.get_ident() << " into fpl";
					m_signal_log(log_debug0, oss.str());
				}
				lgraphclearcurrent();
				return false;
			}
			if (get_destination_ifr())
				m_route[0].set_flags(m_route[0].get_flags() | FPlanWaypoint::altflag_ifr);
			else
				m_route[0].set_flags(m_route[0].get_flags() & ~FPlanWaypoint::altflag_ifr);
		}
	}
	for (LRoute::size_type i(r.size()); i > 1; ) {
		--i;
		LGraph::vertex_descriptor u(r[i-1].get_vertex()), v(r[i].get_vertex());
		m_solutionvertices.insert(u);
		const LVertex& uu(m_lgraph[u]);
		const LVertex& vv(m_lgraph[v]);
		LGraph::edge_descriptor e;
		{
			bool ok;
			tie(e, ok) = r.get_edge(i, m_lgraph);
			if (!ok) {
				if (true) {
					std::ostringstream oss;
					oss << "k shortest path: Solution Edge " << lgraphvertexname(u)
					    << "--" << lgraphindex2airway(r[i].get_edge(), true) << "->"
					    << lgraphvertexname(v) << " not found";
					m_signal_log(log_debug0, oss.str());
				}
				lgraphclearcurrent();
				return false;
			}
		}
		LEdge& ee(m_lgraph[e]);
		ee.set_solution(true);
		if (false) {
			const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
			const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
			const LEdge& ee(m_lgraph[e]);
			std::ostringstream oss;
			oss << "Solution Edge: " << su.get_ident() << " (" << m_performance[u - su.get_v0()].get_altitude() << ") "
			    << su.get_coord().get_lat_str2() << ' ' << su.get_coord().get_lon_str2()
			    << " --" << lgraphindex2airway(ee.get_ident()) << "-> "
			    << sv.get_ident() << " (" << m_performance[v - sv.get_v0()].get_altitude() << ") "
			    << sv.get_coord().get_lat_str2() << ' ' << sv.get_coord().get_lon_str2();
			m_signal_log(log_debug0, oss.str());
		}
		if (uu.get_superindex() == vv.get_superindex())
			continue;
		const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
		const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
		if (!su.insert(m_route, 0)) {
			if (true) {
				std::ostringstream oss;
				oss << "Cannot insert " << su.get_ident() << " into fpl";
				m_signal_log(log_debug0, oss.str());
			}
			lgraphclearcurrent();
			return false;
		}
		m_route[0].set_pathcode(FPlanWaypoint::pathcode_none);
		if (i <= 1)
			break;
		m_route[0].set_flags(m_route[0].get_flags() | (FPlanWaypoint::altflag_ifr | FPlanWaypoint::altflag_standard));
		unsigned int pi(u - su.get_v0());
		if (pi >= m_performance.size()) {
			if (true)
				m_signal_log(log_debug0, "Level graph structure error");
			lgraphclearcurrent();
			return false;
		}
		m_route[0].set_altitude(m_performance[pi].get_altitude());
		if (ee.is_match(airwaydct)) {
			m_route[0].set_pathcode(FPlanWaypoint::pathcode_directto);
		} else if (ee.is_match(airwaymatchawy)) {
			m_route[0].set_pathname(lgraphindex2airway(ee.get_ident()));
			m_route[0].set_pathcode(FPlanWaypoint::pathcode_airway);
		}
	}
	if (get_departure_ifr())
		m_route[0].set_flags(m_route[0].get_flags() | FPlanWaypoint::altflag_ifr);
	else
		m_route[0].set_flags(m_route[0].get_flags() & ~FPlanWaypoint::altflag_ifr);
	if (false && m_route.get_nrwpt() >= 4) {
		FPlanWaypoint& wptstar(m_route[m_route.get_nrwpt() - 2]);
		const FPlanWaypoint& wpt(m_route[m_route.get_nrwpt() - 3]);
		wptstar.set_altitude(wpt.get_altitude());
		uint16_t flags(wptstar.get_flags());
		flags ^= (flags ^ wpt.get_flags()) & FPlanWaypoint::altflag_standard;
		wptstar.set_flags(flags);
	}
	updatefpl();
	return true;
}

void CFMUAutoroute::lgraphmodified(void)
{
	m_solutiontree.clear();
	m_solutionpool.clear();
}

bool CFMUAutoroute::lgraphisinsolutiontree(const LRoute& r) const
{
	const LGraphSolutionTreeNode *tree(&m_solutiontree);
	for (LRoute::size_type i(0); i < r.size(); ++i) {
		LGraphSolutionTreeNode::const_iterator ti(tree->find(LVertexEdge(r[i])));
		if (ti == tree->end())
			return false;
		tree = &ti->second;
	}
	return true;
}

bool CFMUAutoroute::lgraphinsertsolutiontree(const LRoute& r)
{
	bool intree(true);
	LGraphSolutionTreeNode *tree(&m_solutiontree);
	for (LRoute::const_iterator ri(r.begin()), re(r.end()); ri != re; ++ri) {
		std::pair<LGraphSolutionTreeNode::iterator,bool> ins(tree->insert(LGraphSolutionTreeNode::value_type(LVertexEdge(*ri), LGraphSolutionTreeNode())));
		tree = &ins.first->second;
		intree = intree && !ins.second;
	}
	return intree;
}

std::pair<CFMUAutoroute::LGraph::edge_descriptor,bool> CFMUAutoroute::lgraphfindbestedge(LGraph::vertex_descriptor u, LGraph::vertex_descriptor v, bool solutiononly) const
{
	std::pair<LGraph::edge_descriptor,bool> r;
	r.second = false;
	double bestm(std::numeric_limits<double>::max());
	LGraph::out_edge_iterator e0, e1;
	for (boost::tie(e0, e1) = boost::out_edges(u, m_lgraph); e0 != e1; ++e0) {
		if (boost::target(*e0, m_lgraph) != v)
			continue;
		const LEdge& ee(m_lgraph[*e0]);
		if (solutiononly && !ee.is_solution())
			continue;
		if (std::isnan(ee.get_metric()) || std::isinf(ee.get_metric()))
			continue;
		if (ee.get_metric() >= bestm)
			continue;
		bestm = ee.get_metric();
		r.second = true;
		r.first = *e0;
	}
	return r;
}

CFMUAutoroute::LRoute CFMUAutoroute::lgraphdijkstraroute(LGraph::vertex_descriptor u, LGraph::vertex_descriptor v, const std::vector<LGraph::vertex_descriptor>& predecessors) const
{
	LRoute r;
	if (u == v)
		return r;
	while (v != u) {
		LGraph::vertex_descriptor uu(predecessors[v]);
		if (uu == v) {
			r.clear();
			return r;
		}
		std::pair<LGraph::edge_descriptor,bool> e(lgraphfindbestedge(uu, v, false));
		if (!e.second) {
			r.clear();
			return r;
		}
		r.insert(r.begin(), LVertexEdge(e.first, m_lgraph));
		v = uu;
	}
	r.insert(r.begin(), LVertexEdge(u));
	return r;	
}

bool CFMUAutoroute::is_metric_invalid(double m)
{
	return std::isnan(m) || std::isinf(m) || m == std::numeric_limits<double>::max();
}

struct CFMUAutoroute::Combine {
        double operator()(double a, const CFMUAutoroute::LEdge& b) const
	{
                return a + b.m_metric;
        }
};

#if 0

CFMUAutoroute::LRoute CFMUAutoroute::lgraphnextsolution(void)
{
	LRoute r;
	for (;;) {
		if (m_solutionpool.empty()) {
			const LSuperVertex& sdep(m_lsupervertices[m_superdep]);
			const LSuperVertex& sdest(m_lsupervertices[m_superdest]);
			std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph), 0);
			dijkstra_shortest_paths(m_lgraph, sdep.get_v0(), boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)).
						predecessor_map(&predecessors[0]));
			r = lgraphdijkstraroute(sdep.get_v0(), sdest.get_v0(), predecessors);
			lgraphupdatemetric(r);
			if (is_metric_invalid(r.get_metric())) {
				r.clear();
				return r;
			}
			m_solutionpool.insert(r);
			// check if already in solution tree; return if not, otherwise use as solution to base other solutions off
			if (lgraphinsertsolutiontree(r))
				return r;
		}
		r = *m_solutionpool.begin();
		m_solutionpool.erase(m_solutionpool.begin());
		// verify if solution still exists in graph
		bool ok(r.is_existing(m_lgraph));
		if (true) {
			std::ostringstream oss;
			oss << "k shortest path: Route";
			for (LRoute::const_iterator ri(r.begin()), re(r.end()); ri != re; ++ri) {
				if (ri->get_edge() != airwaymatchnone)
					oss << ' ' << lgraphindex2airway(ri->get_edge(), true);
				oss << ' ' << lgraphvertexname(ri->get_vertex());
			}
			oss << " metric " << r.get_metric();
			if (ok)
				oss << " still exists";
			else
				oss << " no longer exists";
			m_signal_log(log_debug0, oss.str());
		}
		if (ok)
			break;
	}
	lgraphinsertsolutiontree(r);
	// generate new solutions
	typedef std::map<LGraph::edge_descriptor,double> wmap_t;
	wmap_t wmap;	
	{
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1; ++e0) {
			const LEdge& ee(m_lgraph[*e0]);
			wmap.insert(wmap_t::value_type(*e0, ee.get_metric()));
		}
	}
	boost::associative_property_map<wmap_t> pmwmap(wmap); 
	LGraphSolutionTreeNode *tree(&m_solutiontree);
	const LSuperVertex& sdest(m_lsupervertices[m_superdest]);
	for (LRoute::size_type i(0); i + 1 < r.size(); ++i) {
		if (tree) {
			LGraphSolutionTreeNode::iterator ti(tree->find(LVertexEdge(r[i])));
			if (ti == tree->end())
				tree = 0;
			else
				tree = &ti->second;
		}
		if (tree) {
			for (LGraphSolutionTreeNode::iterator ti(tree->begin()), te(tree->end()); ti != te;) {
				LGraphSolutionTreeNode::iterator tix(ti);
				++ti;
				LGraph::edge_descriptor e;
                                bool ok;
				boost::tie(e, ok) = LRoute::get_edge(r[i], tix->first, m_lgraph);
				if (!ok) {
					if (true) {
						std::ostringstream oss;
						oss << "k shortest path: Leg " << lgraphvertexname(r[i].get_vertex())
						    << "--" << lgraphindex2airway(tix->first.get_edge(), true) << "->"
						    << lgraphvertexname(tix->first.get_vertex()) << " does not exist in graph";
						m_signal_log(log_debug0, oss.str());
					}
					tree->erase(tix);
					continue;
				}
				wmap_t::iterator wi(wmap.find(e));
				if (wi == wmap.end()) {
					if (true) {
						std::ostringstream oss;
						oss << "k shortest path: Leg " << lgraphvertexname(r[i].get_vertex())
						    << "--" << lgraphindex2airway(tix->first.get_edge(), true) << "->"
						    << lgraphvertexname(tix->first.get_vertex()) << " does not exist in weight map";
						m_signal_log(log_debug0, oss.str());
					}
					continue;
				}
				wi->second = std::numeric_limits<double>::max();
			}
		}
		// kill level changes immediately after SID endpoint
		if (i == 1) {
			LGraph::out_edge_iterator e0, e1;
                        for (boost::tie(e0, e1) = boost::out_edges(r[i].get_vertex(), m_lgraph); e0 != e1; ++e0) {
				const LEdge& ee(m_lgraph[*e0]);
				if (!ee.is_match(airwaymatchlevel))
					continue;
				wmap_t::iterator wi(wmap.find(*e0));
				if (wi == wmap.end())
					continue;
				wi->second = std::numeric_limits<double>::max();
			}
		}
		std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph), 0);
		dijkstra_shortest_paths(m_lgraph, r[i].get_vertex(), boost::weight_map(pmwmap).predecessor_map(&predecessors[0]));
		//std::vector<double> dists(boost::num_vertices(m_lgraph), 0);
		//dijkstra_shortest_paths(m_lgraph, r[i].get_vertex(), boost::weight_map(boost::get(boost::edge_bundle, m_lgraph)).predecessor_map(&predecessors[0]).distance_map(&dists[0]).distance_combine(Combine()));
		//dijkstra_shortest_paths(m_lgraph, r[i].get_vertex(), boost::weight_map(pmwmap).predecessor_map(&predecessors[0]).distance_map(&dists[0]));
		LRoute r1;
		{
			LGraph::vertex_descriptor u(r[i].get_vertex());
			LGraph::vertex_descriptor v(sdest.get_v0());
			while (v != u) {
				LGraph::vertex_descriptor uu(predecessors[v]);
				if (uu == v) {
					r1.clear();
					break;
				}
				std::pair<LGraph::edge_descriptor,bool> e(lgraphfindbestedge(uu, v));
				if (!e.second) {
					r1.clear();
					break;
				}
				r1.insert(r1.begin(), LVertexEdge(e.first, m_lgraph));
				// check that weight map exists and is not infinity
				wmap_t::iterator wi(wmap.find(e.first));
				if (wi == wmap.end()) {
					if (true) {
						std::ostringstream oss;
						oss << "k shortest path: Leg " << lgraphvertexname(boost::source(e.first, m_lgraph))
						    << "--" << lgraphindex2airway(m_lgraph[e.first].get_ident(), true) << "->"
						    << lgraphvertexname(boost::target(e.first, m_lgraph)) << " does not exist in weight map";
						m_signal_log(log_debug0, oss.str());
					}
					r1.clear();
					break;
				}
				if (is_metric_invalid(wi->second)) {
					if (true) {
						std::ostringstream oss;
						oss << "k shortest path: Leg " << lgraphvertexname(boost::source(e.first, m_lgraph))
						    << "--" << lgraphindex2airway(m_lgraph[e.first].get_ident(), true) << "->"
						    << lgraphvertexname(boost::target(e.first, m_lgraph)) << " has invalid metric";
						m_signal_log(log_debug0, oss.str());
					}
					r1.clear();
					break;
				}
				v = uu;
			}
		}
		if (false && tree) {
			LGraph::out_edge_iterator e0, e1;
			for (boost::tie(e0, e1) = boost::out_edges(r[i].get_vertex(), m_lgraph); e0 != e1; ++e0) {
				const LEdge& ee(m_lgraph[*e0]);
				if (boost::target(*e0, m_lgraph) == r[i+1].get_vertex() && 
				    ee.get_ident() == r[i+1].get_edge())
					continue;
				wmap_t::iterator wi(wmap.find(*e0));
				if (wi == wmap.end())
					continue;
				wi->second = ee.get_metric();
			}
		}
		// kill all out-edges to avoid revisiting a node already in the solution
		{
			const LVertex& vv(m_lgraph[r[i].get_vertex()]);
			const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
			for (LGraph::vertex_descriptor v0(sv.get_v0()); v0 <= sv.get_v1(); ++v0) {
				LGraph::out_edge_iterator e0, e1;
				for (boost::tie(e0, e1) = boost::out_edges(v0, m_lgraph); e0 != e1; ++e0) {
					wmap_t::iterator wi(wmap.find(*e0));
					if (wi == wmap.end())
						continue;
					wi->second = std::numeric_limits<double>::max();
					if (true) {
						std::ostringstream oss;
						oss << "k shortest path: Killing edge " << lgraphvertexname(boost::source(*e0, m_lgraph))
						    << "--" << lgraphindex2airway(m_lgraph[*e0].get_ident(), true) << "->"
						    << lgraphvertexname(boost::target(*e0, m_lgraph));
						m_signal_log(log_debug0, oss.str());
					}
				}
			}
		}
		if (r1.size() < 2)
			continue;
		r1.insert(r1.begin(), r.begin(), r.begin() + (i+1));
		lgraphupdatemetric(r1);
		if (is_metric_invalid(r1.get_metric())) {
			if (true) {
				std::ostringstream oss;
				oss << "k shortest path: Route";
		        	for (LRoute::const_iterator ri(r1.begin()), re(r1.end()); ri != re; ++ri) {
					if (ri->get_edge() != airwaymatchnone)
						oss << ' ' << lgraphindex2airway(ri->get_edge(), true);
					oss << ' ' << lgraphvertexname(ri->get_vertex());
				}
				oss << " has no metric";
				m_signal_log(log_debug0, oss.str());
			}
			continue;
		}
		if (true) {
			std::ostringstream oss;
			oss << "k shortest path: Route";
			for (LRoute::const_iterator ri(r1.begin()), re(r1.end()); ri != re; ++ri) {
				if (ri->get_edge() != airwaymatchnone)
					oss << ' ' << lgraphindex2airway(ri->get_edge(), true);
				oss << ' ' << lgraphvertexname(ri->get_vertex());
			}
			oss << " metric " << r1.get_metric() << " added to solution pool";
			m_signal_log(log_debug0, oss.str());
		}
		m_solutionpool.insert(r1);
	}
	// check for too many 
	while (!m_solutionpool.empty() && !m_solutionpool.begin()->is_existing(m_lgraph))
		m_solutionpool.erase(m_solutionpool.begin());
	{
		static const lgraphsolutionpool_t::size_type maxsz(16384);
		lgraphsolutionpool_t::size_type sz(m_solutionpool.size());
		if (sz > maxsz) {
			lgraphsolutionpool_t::size_type n(0);
			for (lgraphsolutionpool_t::iterator i(m_solutionpool.begin()), e(m_solutionpool.end()); i != e; ) {
				lgraphsolutionpool_t::iterator i2(i);
				++i;
				if (n >= maxsz || !i2->is_existing(m_lgraph)) {
					m_solutionpool.erase(i2);
					continue;
				}
				++n;
			}
		}
	}
	if (m_solutionpool.empty())
		return LRoute();
	return *m_solutionpool.begin();
}

#else

CFMUAutoroute::LRoute CFMUAutoroute::lgraphnextsolution(void)
{
	LRoute r;
	for (;;) {
		if (m_solutionpool.empty()) {
			const LSuperVertex& sdep(m_lsupervertices[m_superdep]);
			const LSuperVertex& sdest(m_lsupervertices[m_superdest]);
			std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph), 0);
			dijkstra_shortest_paths(m_lgraph, sdep.get_v0(), boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)).
						predecessor_map(&predecessors[0]));
			r = lgraphdijkstraroute(sdep.get_v0(), sdest.get_v0(), predecessors);
			lgraphupdatemetric(r);
			if (is_metric_invalid(r.get_metric())) {
				r.clear();
				return r;
			}
			m_solutionpool.insert(r);
			// check if already in solution tree; return if not, otherwise use as solution to base other solutions off
			if (lgraphinsertsolutiontree(r))
				return r;
		}
		r = *m_solutionpool.begin();
		m_solutionpool.erase(m_solutionpool.begin());
		// verify if solution still exists in graph
		bool ok(r.is_existing(m_lgraph));
		if (true) {
			std::ostringstream oss;
			oss << "k shortest path: Route";
			for (LRoute::const_iterator ri(r.begin()), re(r.end()); ri != re; ++ri) {
				if (ri->get_edge() != airwaymatchnone)
					oss << ' ' << lgraphindex2airway(ri->get_edge(), true);
				oss << ' ' << lgraphvertexname(ri->get_vertex());
			}
			oss << " metric " << r.get_metric();
			if (ok)
				oss << " still exists";
			else
				oss << " no longer exists";
			m_signal_log(log_debug0, oss.str());
		}
		if (ok)
			break;
	}
	lgraphinsertsolutiontree(r);
	// generate new solutions
	// "abuse" solution flag to enable/disable edges
	{
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1; ++e0) {
			LEdge& ee(m_lgraph[*e0]);
			ee.set_solution(true);
		}
	}
        LGraphSolutionEdgeFilter filter(m_lgraph);
        typedef boost::filtered_graph<LGraph, LGraphSolutionEdgeFilter> fg_t;
        fg_t fg(m_lgraph, filter);
	LGraphSolutionTreeNode *tree(&m_solutiontree);
	const LSuperVertex& sdest(m_lsupervertices[m_superdest]);
	for (LRoute::size_type i(0); i + 1 < r.size(); ++i) {
		if (tree) {
			LGraphSolutionTreeNode::iterator ti(tree->find(LVertexEdge(r[i])));
			if (ti == tree->end())
				tree = 0;
			else
				tree = &ti->second;
		}
		if (tree) {
			for (LGraphSolutionTreeNode::iterator ti(tree->begin()), te(tree->end()); ti != te;) {
				LGraphSolutionTreeNode::iterator tix(ti);
				++ti;
				LGraph::edge_descriptor e;
                                bool ok;
				boost::tie(e, ok) = LRoute::get_edge(r[i], tix->first, m_lgraph);
				if (!ok) {
					if (true) {
						std::ostringstream oss;
						oss << "k shortest path: Leg " << lgraphvertexname(r[i].get_vertex())
						    << "--" << lgraphindex2airway(tix->first.get_edge(), true) << "->"
						    << lgraphvertexname(tix->first.get_vertex()) << " does not exist in graph";
						m_signal_log(log_debug0, oss.str());
					}
					tree->erase(tix);
					continue;
				}
				m_lgraph[e].set_solution(false);
			}
		}
		// kill level changes immediately after SID endpoint
		if (i == 1) {
			LGraph::out_edge_iterator e0, e1;
                        for (boost::tie(e0, e1) = boost::out_edges(r[i].get_vertex(), m_lgraph); e0 != e1; ++e0) {
				LEdge& ee(m_lgraph[*e0]);
				if (!ee.is_match(airwaymatchlevel))
					continue;
				ee.set_solution(false);
			}
		}
		// kill other vertices with the same ident
		{
			unsigned int superidx(m_lgraph[r[i].get_vertex()].get_superindex());
			const Glib::ustring& ident(m_lsupervertices[superidx].get_ident());
			for (unsigned int sidx(0); sidx < m_lsupervertices.size(); ++sidx) {
				if (sidx == superidx)
					continue;
				const LSuperVertex& sv(m_lsupervertices[sidx]);
				if (sv.get_ident() != ident)
					continue;
				for (LGraph::vertex_descriptor v0(sv.get_v0()); v0 <= sv.get_v1(); ++v0) {
					LGraph::out_edge_iterator e0, e1;
					for (boost::tie(e0, e1) = boost::out_edges(v0, m_lgraph); e0 != e1; ++e0) {
						LEdge& ee(m_lgraph[*e0]);
						ee.set_solution(false);
						if (false) {
							std::ostringstream oss;
							oss << "k shortest path: SV " << sidx
							    << " Killing edge " << lgraphvertexname(boost::source(*e0, m_lgraph))
							    << "--" << lgraphindex2airway(ee.get_ident(), true) << "->"
							    << lgraphvertexname(boost::target(*e0, m_lgraph));
							m_signal_log(log_debug0, oss.str());
						}
					}
				}
			}
		}
		std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph), 0);
		dijkstra_shortest_paths(fg, r[i].get_vertex(), boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)).predecessor_map(&predecessors[0]));
		LRoute r1;
		{
			LGraph::vertex_descriptor u(r[i].get_vertex());
			LGraph::vertex_descriptor v(sdest.get_v0());
			while (v != u) {
				LGraph::vertex_descriptor uu(predecessors[v]);
				if (uu == v) {
					r1.clear();
					break;
				}
				std::pair<LGraph::edge_descriptor,bool> e(lgraphfindbestedge(uu, v, true));
				if (!e.second) {
					r1.clear();
					break;
				}
				r1.insert(r1.begin(), LVertexEdge(e.first, m_lgraph));
				// check that weight map exists and is not infinity
				LEdge& ee(m_lgraph[e.first]);
				if (!ee.is_solution()) {
					if (true) {
						std::ostringstream oss;
						oss << "k shortest path: Leg " << lgraphvertexname(boost::source(e.first, m_lgraph))
						    << "--" << lgraphindex2airway(ee.get_ident(), true) << "->"
						    << lgraphvertexname(boost::target(e.first, m_lgraph)) << " does not exist in filtered graph";
						m_signal_log(log_debug0, oss.str());
					}
					r1.clear();
					break;
				}
				v = uu;
			}
		}
		if (false && tree) {
			LGraph::out_edge_iterator e0, e1;
			for (boost::tie(e0, e1) = boost::out_edges(r[i].get_vertex(), m_lgraph); e0 != e1; ++e0) {
				LEdge& ee(m_lgraph[*e0]);
				if (boost::target(*e0, m_lgraph) == r[i+1].get_vertex() &&
				    ee.get_ident() == r[i+1].get_edge())
					continue;
				ee.set_solution(false);
			}
		}
		// kill all out-edges to avoid revisiting a node already in the solution
		{
			unsigned int superidx(m_lgraph[r[i].get_vertex()].get_superindex());
			const LSuperVertex& sv(m_lsupervertices[superidx]);
			for (LGraph::vertex_descriptor v0(sv.get_v0()); v0 <= sv.get_v1(); ++v0) {
				LGraph::out_edge_iterator e0, e1;
				for (boost::tie(e0, e1) = boost::out_edges(v0, m_lgraph); e0 != e1; ++e0) {
					LEdge& ee(m_lgraph[*e0]);
					ee.set_solution(false);
					if (false) {
						std::ostringstream oss;
						oss << "k shortest path: SV " << superidx
						    << " Killing edge " << lgraphvertexname(boost::source(*e0, m_lgraph))
						    << "--" << lgraphindex2airway(ee.get_ident(), true) << "->"
						    << lgraphvertexname(boost::target(*e0, m_lgraph));
						m_signal_log(log_debug0, oss.str());
					}
				}
			}
		}
		if (r1.size() < 2)
			continue;
		r1.insert(r1.begin(), r.begin(), r.begin() + (i+1));
		lgraphupdatemetric(r1);
		if (is_metric_invalid(r1.get_metric())) {
			if (true) {
				std::ostringstream oss;
				oss << "k shortest path: Route";
		        	for (LRoute::const_iterator ri(r1.begin()), re(r1.end()); ri != re; ++ri) {
					if (ri->get_edge() != airwaymatchnone)
						oss << ' ' << lgraphindex2airway(ri->get_edge(), true);
					oss << ' ' << lgraphvertexname(ri->get_vertex());
				}
				oss << " has no metric";
				m_signal_log(log_debug0, oss.str());
			}
			continue;
		}
		if (true) {
			std::ostringstream oss;
			oss << "k shortest path: Route";
			for (LRoute::const_iterator ri(r1.begin()), re(r1.end()); ri != re; ++ri) {
				if (ri->get_edge() != airwaymatchnone)
					oss << ' ' << lgraphindex2airway(ri->get_edge(), true);
				oss << ' ' << lgraphvertexname(ri->get_vertex());
			}
			oss << " metric " << r1.get_metric() << " added to solution pool";
			m_signal_log(log_debug0, oss.str());
		}
		m_solutionpool.insert(r1);
	}
	// check for too many 
	while (!m_solutionpool.empty() && !m_solutionpool.begin()->is_existing(m_lgraph))
		m_solutionpool.erase(m_solutionpool.begin());
	{
		static const lgraphsolutionpool_t::size_type maxsz(16384);
		lgraphsolutionpool_t::size_type sz(m_solutionpool.size());
		if (sz > maxsz) {
			lgraphsolutionpool_t::size_type n(0);
			for (lgraphsolutionpool_t::iterator i(m_solutionpool.begin()), e(m_solutionpool.end()); i != e; ) {
				lgraphsolutionpool_t::iterator i2(i);
				++i;
				if (n >= maxsz || !i2->is_existing(m_lgraph)) {
					m_solutionpool.erase(i2);
					continue;
				}
				++n;
			}
		}
	}
	if (m_solutionpool.empty())
		return LRoute();
	return *m_solutionpool.begin();
}

#endif

void CFMUAutoroute::lgraphroute(void)
{
	for (;;) {
		if (m_superdep >= m_lsupervertices.size() || m_superdest >= m_lsupervertices.size() ||
		    m_superdep == m_superdest) {
			m_signal_log(log_debug0, "Invalid departure and/or destination, stopping...");
			stop();
			return;
		}
		// dispatch events, check for stop command
		{
			
			Glib::RefPtr<Glib::MainContext> ctx(Glib::MainContext::get_default());
			while (ctx->iteration(false));
			if (!m_running) {
				m_signal_log(log_debug0, "Interrupted by user, stopping...");
				return;
			}
		}
		lgraphclearcurrent();
		LRoute r;
		{
			Glib::TimeVal tv;
			tv.assign_current_time();
			r = lgraphnextsolution();
			{
				Glib::TimeVal tv1;
				tv1.assign_current_time();
				tv = tv1 - tv;
			}
			if (true) {
				std::ostringstream oss;
				oss << "Next route computation: " << std::fixed << std::setprecision(3) << tv.as_double()
				    << "s, solution pool size " << m_solutionpool.size();
				m_signal_log(log_debug0, oss.str());
			}
		}
		if (r.size() < 2) {
			if (true) {
				const LSuperVertex& sdep(m_lsupervertices[m_superdep]);
				const LSuperVertex& sdest(m_lsupervertices[m_superdest]);
				std::ostringstream oss;
				oss << "No route found; "
				    << boost::out_degree(sdep.get_v0(), m_lgraph) << " SID edges, "
				    << boost::in_degree(sdest.get_v0(), m_lgraph) << " STAR edges";
				m_signal_log(log_debug0, oss.str());
			}
			stop();
			return;
		}
		if (!lgraphsetcurrent(r)) {
			stop();
			return;
		}
		if (r.is_repeated_nodes(m_lgraph, m_lsupervertices)) {
			if (true) {
				std::ostringstream oss;
				oss << "Flight Plan has duplicate points";
		        	for (LRoute::const_iterator ri(r.begin()), re(r.end()); ri != re; ++ri) {
					if (ri->get_edge() != airwaymatchnone)
						oss << ' ' << lgraphindex2airway(ri->get_edge(), true);
					oss << ' ' << lgraphvertexname(ri->get_vertex());
				}
				m_signal_log(log_debug0, oss.str());
			}
			continue;
		}
		if (r.size() >= 4 && r[2].is_match(airwaymatchlevel)) {
			if (true) {
				std::ostringstream oss;
				oss << "Flight Plan contains level change on SID endpoint";
		        	for (LRoute::const_iterator ri(r.begin()), re(r.end()); ri != re; ++ri) {
					if (ri->get_edge() != airwaymatchnone)
						oss << ' ' << lgraphindex2airway(ri->get_edge(), true);
					oss << ' ' << lgraphvertexname(ri->get_vertex());
				}
				m_signal_log(log_debug0, oss.str());
			}
			continue;
		}
		m_signal_log(log_fplproposal, get_simplefpl());
		bool nochange(true);
		if (get_tfr_enabled()) {
			TrafficFlowRestrictions::Result res;
			{
				Glib::TimeVal tv;
				tv.assign_current_time();
				res = m_tfr.check_fplan(m_route, 'G', m_aircraft.get_aircrafttype(), m_aircraft.get_equipment(),
							m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb);
				{
					Glib::TimeVal tv1;
					tv1.assign_current_time();
					tv = tv1 - tv;
				}
				if (true) {
					std::ostringstream oss;
					oss << "Local TFR Check: " << std::fixed << std::setprecision(3) << tv.as_double()
					    << 's';
					m_signal_log(log_debug0, oss.str());
				}
			}
			// Log Messages
			for (unsigned int i = 0; i < res.messages_size(); ++i)
				logmessage(res.get_message(i));
			nochange = res.empty();
			typedef TrafficFlowRestrictions::RuleResult::set_t set_t;
			for (unsigned int i = 0; i < res.size(); ++i) {
				const TrafficFlowRestrictions::RuleResult& rule(res[i]);
				{
					std::ostringstream oss;
					oss << rule.get_rule() << ' ';
					switch (rule.get_codetype()) {
					case TrafficFlowRestrictions::RuleResult::codetype_mandatory:
						oss << 'M';
						break;

					case TrafficFlowRestrictions::RuleResult::codetype_forbidden:
						oss << 'F';
						break;

					case TrafficFlowRestrictions::RuleResult::codetype_closed:
						oss << 'C';
						break;

					default:
						oss << '?';
						break;
					}
					if (!rule.get_vertexset().empty()) {
						oss << " V:";
						bool subseq(false);
						for (set_t::const_iterator i(rule.get_vertexset().begin()), e(rule.get_vertexset().end()); i != e; ++i) {
							if (subseq)
								oss << ',';
							subseq = true;
							oss << *i;
						}
					}
					if (!rule.get_edgeset().empty()) {
						oss << " E:";
						bool subseq(false);
						for (set_t::const_iterator i(rule.get_edgeset().begin()), e(rule.get_edgeset().end()); i != e; ++i) {
							if (subseq)
								oss << ',';
							subseq = true;
							oss << *i;
						}
					}
					for (unsigned int i = 0; i < rule.size(); ++i) {
						const TrafficFlowRestrictions::RuleResult::Alternative& alt(rule[i]);
						if (i)
							oss << " /";
						if (!alt.get_vertexset().empty()) {
							oss << " V:";
							bool subseq(false);
							for (set_t::const_iterator i(alt.get_vertexset().begin()), e(alt.get_vertexset().end()); i != e; ++i) {
								if (subseq)
									oss << ',';
								subseq = true;
								oss << *i;
							}
						}
						if (!alt.get_edgeset().empty()) {
							oss << " E:";
							bool subseq(false);
							for (set_t::const_iterator i(alt.get_edgeset().begin()), e(alt.get_edgeset().end()); i != e; ++i) {
								if (subseq)
									oss << ',';
								subseq = true;
								oss << *i;
							}
						}
						for (unsigned int j = 0; j < alt.size(); ++j) {
							const TrafficFlowRestrictions::RuleResult::Alternative::Sequence& el(alt[j]);
							oss << ' ' << el.get_type_string();
							switch (el.get_type()) {
							case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airway:
								oss << ' ' << el.get_ident();
								break;

							case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_sid:
								oss << ' ' << el.get_ident() << ' ' << el.get_airport();
								break;

							case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_star:
								oss << ' ' << el.get_ident() << ' ' << el.get_airport();
								break;

							case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airspace:
								oss << ' ' << el.get_ident() << '/'
								    << AirspacesDb::Airspace::get_class_string(el.get_bdryclass(), AirspacesDb::Airspace::typecode_ead);
								break;

							case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct:
							case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_point:
							default:
								break;
							}
							oss << " F" << std::setw(3) << std::setfill('0') << ((el.get_lower_alt() + 50) / 100) << "...F"
							    << std::setw(3) << std::setfill('0') << ((std::min((int32_t)el.get_upper_alt(), (int32_t)99900) + 50) / 100);
							if (el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airway ||
							    el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct ||
							    el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_point) {
								const TrafficFlowRestrictions::RuleResult::Alternative::Sequence::RulePoint& pt(el.get_point0());
								oss << ' ' << pt.get_ident();
							}
							if (el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airway ||
							    el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct) {
								const TrafficFlowRestrictions::RuleResult::Alternative::Sequence::RulePoint& pt(el.get_point1());
								oss << ' ' << pt.get_ident();
							}
						}
					}
					m_signal_log(log_graphrule, oss.str());
					m_signal_log(log_fplvalidation, oss.str());
				}
				lgraphtfrruleresult(rule, res.get_fplan());
			}
		}
		if (false) {
			Glib::RefPtr<Glib::MainContext> ctx(Glib::MainContext::get_default());
			ctx->iteration(false);
		}
		if (!m_running) {
			m_signal_log(log_debug0, "Interrupted by user, stopping...");
			return;
		}
		if (nochange)
			break;
		m_signal_log(log_fplvalidation, "");
		++m_iterationcount[0];
		if (true) {
			Glib::signal_idle().connect_once(sigc::mem_fun(*this, &CFMUAutoroute::lgraphroute), Glib::PRIORITY_HIGH_IDLE);
			return;
		}
	}
	lgraphsubmitroute();
}

void CFMUAutoroute::lgraphsubmitroute(void)
{
	++m_iterationcount[1];
	if (!m_childrun) {
		child_close();
		child_run();
	}
	if (!m_childrun) {
		m_signal_log(log_debug0, "Cannot restart child, stopping...");
		stop();
		return;
	}
	m_childchanstdin->write(get_simplefpl() + "\n");
	m_childchanstdin->flush();
	m_childtimeoutcount = 0;
	m_connchildtimeout.disconnect();
	m_connchildtimeout = Glib::signal_timeout().connect(sigc::mem_fun(*this, &CFMUAutoroute::child_timeout), 60000);
}

bool CFMUAutoroute::child_timeout(void)
{
	m_connchildtimeout.disconnect();
	if (!m_running) {
		m_signal_log(log_debug0, "Interrupted by user, stopping...");
		return false;
	}
	++m_childtimeoutcount;
	if (m_childtimeoutcount > 5) {
		m_signal_log(log_debug0, "Child timed out, stopping...");
		stop();
		child_close();
		return false;
	}
	child_close();
	child_run();
	if (!m_childrun) {
		m_signal_log(log_debug0, "Cannot restart child after timeout, stopping...");
		stop();
		return false;
	}		
	m_childchanstdin->write(get_simplefpl() + "\n");
	m_childchanstdin->flush();
	m_connchildtimeout = Glib::signal_timeout().connect(sigc::mem_fun(*this, &CFMUAutoroute::child_timeout), 60000);
	return true;
}

void CFMUAutoroute::lgraphtfrruleresult(const TrafficFlowRestrictions::RuleResult& rule, const TrafficFlowRestrictions::Fpl& fpl)
{
	// actually modify the graph
	typedef TrafficFlowRestrictions::RuleResult::set_t set_t;
	switch (rule.get_codetype()) {
	case TrafficFlowRestrictions::RuleResult::codetype_forbidden:
		{
			bool work(false);
			for (unsigned int j = 0; j < rule.size(); ++j) {
				const TrafficFlowRestrictions::RuleResult::Alternative& alt(rule[j]);
				if (alt.get_edgeset().empty()) {
					unsigned int i;
					{
						set_t::const_iterator it(alt.get_vertexset().begin()), e(alt.get_vertexset().end());
						if (it == e)
							continue;
						i = *it;
						++it;
						if (it != e)
							continue;
					}
					if (i >= fpl.size())
						continue;
					if (!rule.get_edgeset().empty())
						continue;
					{
						set_t vset, vset1;
						vset1.insert(i);
						std::set_difference(alt.get_vertexset().begin(), alt.get_vertexset().end(), vset1.begin(), vset1.end(),
								    std::insert_iterator<set_t>(vset, vset.begin()));
						if (!vset.empty())
							continue;
						vset.clear();
						vset1.insert(0U);
						vset1.insert(fpl.size() - 1U);
						std::set_difference(rule.get_vertexset().begin(), rule.get_vertexset().end(), vset1.begin(), vset1.end(),
								    std::insert_iterator<set_t>(vset, vset.begin()));
						if (!vset.empty())
							continue;
					}
					const TrafficFlowRestrictions::FplWpt& wpt(fpl[i]);
					work = lgraphdisconnectsolutionvertex(wpt.get_ident()) || work;
					continue;
				}
				unsigned int i;
				{
					set_t::const_iterator it(alt.get_edgeset().begin()), e(alt.get_edgeset().end());
					if (it == e)
						continue;
					i = *it;
					++it;
					if (it != e)
						continue;
				}
				if (i >= fpl.size() + 1U)
					continue;
				{
					set_t vset, vset1;
					vset1.insert(i);
					vset1.insert(i + 1U);
					std::set_difference(alt.get_vertexset().begin(), alt.get_vertexset().end(), vset1.begin(), vset1.end(),
							    std::insert_iterator<set_t>(vset, vset.begin()));
					if (!vset.empty())
						break;
					vset.clear();
					vset1.insert(0U);
					vset1.insert(fpl.size() - 1U);
					std::set_difference(rule.get_vertexset().begin(), rule.get_vertexset().end(), vset1.begin(), vset1.end(),
							    std::insert_iterator<set_t>(vset, vset.begin()));
					if (!vset.empty())
						continue;
					vset.clear();
					std::set_difference(rule.get_edgeset().begin(), rule.get_edgeset().end(),
							    alt.get_edgeset().begin(), alt.get_edgeset().end(),
							    std::insert_iterator<set_t>(vset, vset.begin()));
					if (!vset.empty())
						continue;
				}
				const TrafficFlowRestrictions::FplWpt& wpt0(fpl[i]);
				const TrafficFlowRestrictions::FplWpt& wpt1(fpl[i + 1U]);
				work = lgraphdisconnectsolutionedge(wpt0.get_ident(), wpt1.get_ident()) || work;
			}
			if (work)
				return;
			break;
		}

	case TrafficFlowRestrictions::RuleResult::codetype_closed:
		{
			if (rule.get_edgeset().empty()) {
				unsigned int i;
				{
					set_t::const_iterator it(rule.get_vertexset().begin()), e(rule.get_vertexset().end());
					if (it == e)
						break;
					i = *it;
					++it;
					if (it != e)
						break;
				}
				if (i >= fpl.size())
					break;
				const TrafficFlowRestrictions::FplWpt& wpt(fpl[i]);
				if (lgraphdisconnectsolutionvertex(wpt.get_ident()))
					return;
				break;
			}
			unsigned int i;
			{
				set_t::const_iterator it(rule.get_edgeset().begin()), e(rule.get_edgeset().end());
				if (it == e)
					break;
				i = *it;
				++it;
				if (it != e)
					break;
			}
			if (i >= fpl.size() + 1U)
				return;
			{
				set_t vset, vset1;
				vset1.insert(i);
				vset1.insert(i + 1U);
				vset1.insert(0U);
				vset1.insert(fpl.size() - 1U);
				std::set_difference(rule.get_vertexset().begin(), rule.get_vertexset().end(), vset1.begin(), vset1.end(),
						    std::insert_iterator<set_t>(vset, vset.begin()));
				if (!vset.empty())
					break;
			}
			const TrafficFlowRestrictions::FplWpt& wpt0(fpl[i]);
			const TrafficFlowRestrictions::FplWpt& wpt1(fpl[i + 1U]);
			if (lgraphdisconnectsolutionedge(wpt0.get_ident(), wpt1.get_ident()))
				return;
			break;
		}

	default:
	case TrafficFlowRestrictions::RuleResult::codetype_mandatory:
		break;
	}
	double scale(1.05);
	bool work(false);
	switch (rule.get_codetype()) {
	case TrafficFlowRestrictions::RuleResult::codetype_mandatory:
		scale = 0.95;
		// fall through
		
	case TrafficFlowRestrictions::RuleResult::codetype_forbidden:
		for (unsigned int j = 0; j < rule.size(); ++j) {
			const TrafficFlowRestrictions::RuleResult::Alternative& alt(rule[j]);
			if (alt.get_edgeset().empty()) {
				for (set_t::const_iterator i(alt.get_vertexset().begin()), e(alt.get_vertexset().end()); i != e; ++i) {
					if (*i >= fpl.size())
						continue;
					const TrafficFlowRestrictions::FplWpt& wpt(fpl[*i]);
					work = lgraphscalesolutionvertex(wpt.get_ident(), scale) || work;
				}
				continue;
			}
			for (set_t::const_iterator i(alt.get_edgeset().begin()), e(alt.get_edgeset().end()); i != e; ++i) {
				if (*i >= fpl.size() + 1U)
					continue;
				const TrafficFlowRestrictions::FplWpt& wpt0(fpl[*i]);
				const TrafficFlowRestrictions::FplWpt& wpt1(fpl[*i + 1U]);
				work = lgraphscalesolutionedge(wpt0.get_ident(), wpt1.get_ident(), scale) || work;
			}
		}
		break;

	case TrafficFlowRestrictions::RuleResult::codetype_closed:
		if (rule.get_edgeset().empty()) {
			for (set_t::const_iterator i(rule.get_vertexset().begin()), e(rule.get_vertexset().end()); i != e; ++i) {
				if (*i >= fpl.size())
					continue;
				const TrafficFlowRestrictions::FplWpt& wpt(fpl[*i]);
				work = lgraphscalesolutionvertex(wpt.get_ident(), scale) || work;
			}
			break;
		}
		for (set_t::const_iterator i(rule.get_edgeset().begin()), e(rule.get_edgeset().end()); i != e; ++i) {
			if (*i >= fpl.size() + 1U)
				continue;
			const TrafficFlowRestrictions::FplWpt& wpt0(fpl[*i]);
			const TrafficFlowRestrictions::FplWpt& wpt1(fpl[*i + 1U]);
			work = lgraphscalesolutionedge(wpt0.get_ident(), wpt1.get_ident(), scale) || work;
		}
		break;

	default:
		break;
	}
	if (work)
		lgraphmodified();
}

bool CFMUAutoroute::lgraphkilllevelchg(unsigned int superindex)
{
	if (superindex >= m_lsupervertices.size())
		return false;
	return lgraphkilllevelchg(m_lsupervertices[superindex]);
}

bool CFMUAutoroute::lgraphkilllevelchg(const LSuperVertex& super)
{
	if (super.get_v0() == super.get_v1())
		return false;
	bool work(false);
	for (LGraph::vertex_descriptor v0(super.get_v0()); v0 < super.get_v1(); ++v0) {
		LGraph::vertex_descriptor v1(v0);
		for (;;) {
			bool done(true);
			LGraph::out_edge_iterator e0, e1;
			for (tie(e0, e1) = out_edges(v0, m_lgraph); e0 != e1; ) {
				LGraph::out_edge_iterator ex(e0);
				++e0;
				if (target(*ex, m_lgraph) != v1)
					continue;
				remove_edge(ex, m_lgraph);
				work = true;
				if (lgraphrestartedgescan) {
					done = false;
					break;
				}
			}
			if (done)
				break;
		}
		for (;;) {
			bool done(true);
			LGraph::out_edge_iterator e0, e1;
			for (tie(e0, e1) = out_edges(v1, m_lgraph); e0 != e1; ) {
				LGraph::out_edge_iterator ex(e0);
				++e0;
				if (target(*ex, m_lgraph) != v0)
					continue;
				remove_edge(ex, m_lgraph);
				work = true;
				if (lgraphrestartedgescan) {
					done = false;
					break;
				}
			}
			if (done)
				break;
		}
	}
	return work;
}

Glib::ustring CFMUAutoroute::lgraphvertexname(LGraph::vertex_descriptor v)
{
	const LVertex& vv(m_lgraph[v]);
	const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
	std::ostringstream oss;
	oss << sv.get_ident();
	if (sv.get_v0() == sv.get_v1())
		return oss.str();
	const Performance& perf(m_performance[v - sv.get_v0()]);
	oss << ":F" << std::setw(3) << std::setfill('0') << ((perf.get_altitude() + 50) / 100);
	return oss.str();
}

void CFMUAutoroute::lgraphlogkilledge(LGraph::edge_descriptor e)
{
	const LEdge& ee(m_lgraph[e]);
	std::ostringstream oss;
	oss << "- " << lgraphvertexname(source(e, m_lgraph))
	    << ' ' << lgraphindex2airway(ee.get_ident(), true) << ' '
	    << lgraphvertexname(target(e, m_lgraph));
	m_signal_log(log_graphchange, oss.str());
}

void CFMUAutoroute::lgraphlogaddedge(LGraph::edge_descriptor e)
{
	const LEdge& ee(m_lgraph[e]);
	std::ostringstream oss;
	oss << "+ " << lgraphvertexname(source(e, m_lgraph))
	    << ' ' << lgraphindex2airway(ee.get_ident(), true) << ' '
	    << lgraphvertexname(target(e, m_lgraph));
	m_signal_log(log_graphchange, oss.str());
}

void CFMUAutoroute::lgraphlogrenameedge(LGraph::edge_descriptor e, lgraphairwayindex_t awyname)
{
	const LEdge& ee(m_lgraph[e]);
	std::ostringstream oss;
	oss << "R " << lgraphvertexname(source(e, m_lgraph))
	    << ' ' << lgraphindex2airway(ee.get_ident(), true) << ' '
	    << lgraphvertexname(target(e, m_lgraph)) << " -> " << lgraphindex2airway(awyname, true);
	m_signal_log(log_graphchange, oss.str());
}

void CFMUAutoroute::lgraphlogscaleedge(LGraph::edge_descriptor e, double scale)
{
	const LEdge& ee(m_lgraph[e]);
	std::ostringstream oss;
	oss << "S " << lgraphvertexname(source(e, m_lgraph))
	    << ' ' << lgraphindex2airway(ee.get_ident(), true) << ' '
	    << lgraphvertexname(target(e, m_lgraph)) << " : "
	    << ee.get_metric() << " * " << scale << " -> " << (ee.get_metric() * scale);
	m_signal_log(log_graphchange, oss.str());
}

bool CFMUAutoroute::lgraphdisconnectvertex(const Glib::ustring& ident, int lvlfrom, int lvlto)
{
	bool work(false);
	Glib::ustring identu(ident.uppercase());
	lvlfrom *= 100;
	lvlto *= 100;
	for (solutionvertices_t::const_iterator svi(m_solutionvertices.begin()), sve(m_solutionvertices.end()); svi != sve; ++svi) {
		const LVertex& vv(m_lgraph[*svi]);
		const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
		if (sv.get_ident() != identu)
			continue;
		// prevent departure or destination from being disconnected if an intersection has the same ident
		if ((lvlfrom == 0 && lvlto == 60000) && (vv.get_superindex() == m_superdep || vv.get_superindex() == m_superdest))
			continue;
		const Performance& perf(m_performance[*svi - sv.get_v0()]);
		if (perf.get_altitude() < lvlfrom || perf.get_altitude() > lvlto)
			continue;
		for (LGraph::vertex_descriptor v0(sv.get_v0()), v1(sv.get_v1()); v0 <= v1; ++v0) {
			for (;;) {
				bool done(true);
				LGraph::out_edge_iterator e0, e1;
				for (tie(e0, e1) = out_edges(v0, m_lgraph); e0 != e1; ) {
					LGraph::out_edge_iterator ex(e0);
					++e0;
					LGraph::vertex_descriptor u(target(*ex, m_lgraph));
					const LVertex& uu(m_lgraph[u]);
					if (vv.get_superindex() == uu.get_superindex())
						continue;
					lgraphlogkilledge(*ex);
					remove_edge(ex, m_lgraph);
					work = true;
					if (lgraphrestartedgescan) {
						done = false;
						break;
					}
				}
				if (done)
					break;
			}
			for (;;) {
				bool done(true);
				LGraph::in_edge_iterator e0, e1;
				for (tie(e0, e1) = in_edges(v0, m_lgraph); e0 != e1; ) {
					LGraph::in_edge_iterator ex(e0);
					++e0;
					LGraph::vertex_descriptor u(source(*ex, m_lgraph));
					const LVertex& uu(m_lgraph[u]);
					if (vv.get_superindex() == uu.get_superindex())
						continue;
					lgraphlogkilledge(*ex);
					remove_edge(*ex, m_lgraph);
					work = true;
					if (lgraphrestartedgescan) {
						done = false;
						break;
					}
				}
				if (done)
					break;
			}
		}
	}
	return work;
}

bool CFMUAutoroute::lgraphdisconnectsolutionvertex(const Glib::ustring& ident)
{
	bool work(false);
	Glib::ustring identu(ident.uppercase());
	for (;;) {
		bool done(true);
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1;) {
			LGraph::edge_iterator ex(e0);
			++e0;
			LEdge& ee(m_lgraph[*ex]);
			if (!ee.is_solution())
				continue;
			const LVertex& uu(m_lgraph[source(*ex, m_lgraph)]);
			const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
			if (su.get_ident() != identu) {
				const LVertex& vv(m_lgraph[target(*ex, m_lgraph)]);
				const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
				if (sv.get_ident() != identu)
					continue;
			}
			lgraphlogkilledge(*ex);
			remove_edge(*ex, m_lgraph);
			work = true;
			if (lgraphrestartedgescan) {
				done = false;
				break;
			}
		}
		if (done)
			break;
	}
	return work;
}

bool CFMUAutoroute::lgraphdisconnectsolutionedge(const Glib::ustring& id0, const Glib::ustring& id1)
{
	bool work(false);
	Glib::ustring id0u(id0.uppercase());
	Glib::ustring id1u(id1.uppercase());
	for (;;) {
		bool done(true);
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1;) {
			LGraph::edge_iterator ex(e0);
			++e0;
			LEdge& ee(m_lgraph[*ex]);
			if (!ee.is_solution())
				continue;
			LGraph::vertex_descriptor u(source(*ex, m_lgraph));
			const LVertex& uu(m_lgraph[u]);
			const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
			if (su.get_ident() != id0u)
				continue;
			LGraph::vertex_descriptor v(target(*ex, m_lgraph));
			const LVertex& vv(m_lgraph[v]);
			const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
			if (sv.get_ident() != id1u)
				continue;
			bool convdct(ee.is_match(airwaymatchawy));
			// check max distance
			if (convdct) {
				double dist(su.get_coord().spheric_distance_nmi_dbl(sv.get_coord()));
				convdct = dist <= get_dctlimit();
			}
			// check whether there already is a DCT
			if (convdct) {
				LGraph::in_edge_iterator e0, e1;
				for (tie(e0, e1) = in_edges(v, m_lgraph); e0 != e1; ++e0) {
					if (*e0 == *ex)
						continue;
					if (source(*e0, m_lgraph) == u)
						continue;
					if (m_lgraph[*e0].is_dct()) {
						convdct = false;
						break;
					}
				}
			}
			// check TFR rules
			if (convdct && get_tfr_enabled()) {
				unsigned int pi(u - su.get_v0());
				if (pi >= m_performance.size()) {
					convdct = false;
				} else {
					TrafficFlowRestrictions::DctParameters dctrule(su.get_ident(), su.get_coord(),
										       sv.get_ident(), sv.get_coord(),
										       m_performance[pi].get_altitude(), m_performance[pi].get_altitude()+1);
					std::vector<TrafficFlowRestrictions::Message> msg;
                                        m_tfr.check_dct(msg, dctrule);
                                        for (unsigned int i = 0; i < msg.size(); ++i)
                                                logmessage(msg[i]);
					if (dctrule.get_alt(0).is_empty())
						convdct = false;
				}
			}
			if (convdct) {
				lgraphlogrenameedge(*ex, airwaydct);
				ee.set_ident(airwaydct);
				ee.set_metric(ee.get_metric() * get_dctpenalty());
				work = true;
				continue;
			}
			lgraphlogkilledge(*ex);
			remove_edge(*ex, m_lgraph);
			work = true;
			if (lgraphrestartedgescan) {
				done = false;
				break;
			}
		}
		if (done)
			break;
	}
	return work;
}

bool CFMUAutoroute::lgraphscalesolutionvertex(const Glib::ustring& ident, double scale)
{
	bool work(false);
	Glib::ustring identu(ident.uppercase());
	LGraph::edge_iterator e0, e1;
	for (tie(e0, e1) = edges(m_lgraph); e0 != e1;) {
		LGraph::edge_iterator ex(e0);
		++e0;
		LEdge& ee(m_lgraph[*ex]);
		if (!ee.is_solution())
			continue;
		const LVertex& uu(m_lgraph[source(*ex, m_lgraph)]);
		const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
		if (su.get_ident() != identu) {
			const LVertex& vv(m_lgraph[target(*ex, m_lgraph)]);
			const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
			if (sv.get_ident() != identu)
				continue;
		}
		double newmetric(ee.get_metric() * scale);
		if (newmetric != ee.get_metric()) {
			lgraphlogscaleedge(*ex, scale);
			work = true;
		}
		ee.set_metric(newmetric);
	}
	return work;
}

bool CFMUAutoroute::lgraphscalesolutionedge(const Glib::ustring& id0, const Glib::ustring& id1, double scale)
{
	bool work(false);
	Glib::ustring id0u(id0.uppercase());
	Glib::ustring id1u(id1.uppercase());
	LGraph::edge_iterator e0, e1;
	for (tie(e0, e1) = edges(m_lgraph); e0 != e1;) {
		LGraph::edge_iterator ex(e0);
		++e0;
		LEdge& ee(m_lgraph[*ex]);
		if (!ee.is_solution())
			continue;
		const LVertex& uu(m_lgraph[source(*ex, m_lgraph)]);
		const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
		if (su.get_ident() != id0u)
			continue;
		const LVertex& vv(m_lgraph[target(*ex, m_lgraph)]);
		const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
		if (sv.get_ident() != id1u)
			continue;
		double newmetric(ee.get_metric() * scale);
		if (newmetric != ee.get_metric()) {
			lgraphlogscaleedge(*ex, scale);
			work = true;
		}
		ee.set_metric(newmetric);
	}
	return work;
}

bool CFMUAutoroute::lgraphmodifyedge(const Glib::ustring& id0, const Glib::ustring& id1, lgraphairwayindex_t awyremove,
				     lgraphairwayindex_t awyadd, int lvlfrom, int lvlto, bool bidir)
{
	bool work(false);
	Glib::ustring id0u(id0.uppercase());
	Glib::ustring id1u(id1.uppercase());
	lvlfrom *= 100;
	lvlto *= 100;
	for (solutionvertices_t::const_iterator sui(m_solutionvertices.begin()), sue(m_solutionvertices.end()); sui != sue; ++sui) {
		const LVertex& uu(m_lgraph[*sui]);
		const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
		if (su.get_ident() != id0u)
			continue;
		for (solutionvertices_t::const_iterator svi(m_solutionvertices.begin()), sve(m_solutionvertices.end()); svi != sve; ++svi) {
			if (sui == svi)
				continue;
			const LVertex& vv(m_lgraph[*svi]);
			const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
			if (sv.get_ident() != id1u)
				continue;
			double dist(su.get_coord().spheric_distance_nmi_dbl(sv.get_coord()));
			for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
				const Performance& perf(m_performance[pi]);
				int alt(perf.get_altitude());
				if (alt < lvlfrom || alt > lvlto)
					continue;
				LGraph::vertex_descriptor u(su.get_v0() + pi), v(sv.get_v0() + pi);
				if (u > su.get_v1() || v > sv.get_v1())
					continue;
				bool doadd(awyadd != airwaymatchnone);
				for (;;) {
					bool done(true);
					LGraph::out_edge_iterator e0, e1;
					for (tie(e0, e1) = out_edges(u, m_lgraph); e0 != e1; ) {
						LGraph::out_edge_iterator ex(e0);
						++e0;
						if (target(*ex, m_lgraph) != v)
							continue;
						const LEdge& ee(m_lgraph[*ex]);
						if (ee.is_match(awyadd)) {
							doadd = false;
							continue;
						}
						if (!ee.is_match(awyremove))
							continue;
						lgraphlogkilledge(*ex);
						remove_edge(ex, m_lgraph);
						work = true;
						if (lgraphrestartedgescan) {
							done = false;
							break;
						}
					}
					if (done)
						break;
				}
				if (doadd) {
					LGraph::edge_descriptor e;
					bool added;
					double x(dist * perf.get_secpernmi());
					if (is_optfuel())
						x *= perf.get_fuelpersec();
					if (awyadd == airwaydct)
						x *= get_dctpenalty();
					tie(e, added) = boost::add_edge(u, v, LEdge(awyadd, x), m_lgraph);
					if (added)
						lgraphlogaddedge(e);
					work = work || added;
				}
				if (!bidir)
					continue;
				doadd = awyadd != airwaymatchnone;
				for (;;) {
					bool done(true);
					LGraph::out_edge_iterator e0, e1;
					for (tie(e0, e1) = out_edges(v, m_lgraph); e0 != e1; ) {
						LGraph::out_edge_iterator ex(e0);
						++e0;
						if (target(*ex, m_lgraph) != u)
							continue;
						const LEdge& ee(m_lgraph[*ex]);
						if (ee.is_match(awyadd)) {
							doadd = false;
							continue;
						}
						if (!ee.is_match(awyremove))
							continue;
						lgraphlogkilledge(*ex);
						remove_edge(ex, m_lgraph);
						work = true;
						if (lgraphrestartedgescan) {
							done = false;
							break;
						}
					}
					if (done)
						break;
				}
				if (doadd) {
					LGraph::edge_descriptor e;
					bool added;
					double x(dist * perf.get_secpernmi());
					if (is_optfuel())
						x *= perf.get_fuelpersec();
					if (awyadd == airwaydct)
						x *= get_dctpenalty();
					tie(e, added) = boost::add_edge(v, u, LEdge(awyadd, x), m_lgraph);
					if (added)
						lgraphlogaddedge(e);
					work = work || added;
				}
			}
		}
	}
	return work;
}

bool CFMUAutoroute::lgraphmodifyawysegment(const Glib::ustring& id0, const Glib::ustring& id1, lgraphairwayindex_t awymatch,
					   lgraphairwayindex_t awyrename, int lvlfrom, int lvlto, bool bidir, bool solutiononly)
{
	if (awymatch == awyrename)
		return false;
	bool work(false);
	Glib::ustring id0u(id0.uppercase());
	Glib::ustring id1u(id1.uppercase());
	lvlfrom *= 100;
	lvlto *= 100;
#if 0
	// construct weight map
	typedef std::map<LGraph::edge_descriptor, double> weights_t;
	weights_t weights;
	weights.clear();
	{
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1; ++e0) {
			LEdge& ee(m_lgraph[*e0]);
			int w(1);
			if (!ee.is_match(awymatch))
				//w = std::numeric_limits<double>::max();
				w = 1e200;
			weights[*e0] = w;
		}
	}
	boost::associative_property_map<weights_t> weightsmap(weights);
#else
	LGraphAwyEdgeFilter filter(m_lgraph, awymatch);
        typedef boost::filtered_graph<LGraph, LGraphAwyEdgeFilter> fg_t;
        fg_t fg(m_lgraph, filter);
#endif
	for (solutionvertices_t::const_iterator sui(m_solutionvertices.begin()), sue(m_solutionvertices.end()); sui != sue; ++sui) {
		const LVertex& uu(m_lgraph[*sui]);
		const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
		if (su.get_ident() != id0u)
			continue;
		for (solutionvertices_t::const_iterator svi(m_solutionvertices.begin()), sve(m_solutionvertices.end()); svi != sve; ++svi) {
			if (sui == svi)
				continue;
			const LVertex& vv(m_lgraph[*svi]);
			const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
			if (sv.get_ident() != id1u)
				continue;
			for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
				const Performance& perf(m_performance[pi]);
				int alt(perf.get_altitude());
				if (alt < lvlfrom || alt > lvlto)
					continue;
				LGraph::vertex_descriptor u(su.get_v0() + pi), v(sv.get_v0() + pi);
				if (u > su.get_v1() || v > sv.get_v1())
					continue;
				std::vector<LGraph::vertex_descriptor> predecessors(boost::num_vertices(m_lgraph), 0);
#if 0
				dijkstra_shortest_paths(m_lgraph, u, boost::weight_map(weightsmap).
							predecessor_map(&predecessors[0]));
#else
				dijkstra_shortest_paths(fg, u, boost::weight_map(boost::get(&LEdge::m_metric, m_lgraph)).
							predecessor_map(&predecessors[0]));
#endif
				LGraph::vertex_descriptor ve(v);
				while (ve != u) {
					LGraph::vertex_descriptor ue(predecessors[ve]);
					for (;;) {
						bool done(true);
						unsigned int renedges(0);
						LGraph::out_edge_iterator e0, e1;
						for (tie(e0, e1) = out_edges(ue, m_lgraph); e0 != e1; ) {
							LGraph::out_edge_iterator ex(e0);
							++e0;
							if (target(*ex, m_lgraph) != ve)
								continue;
							LEdge& ee(m_lgraph[*ex]);
							if (ee.is_match(awyrename)) {
								++renedges;
								if (renedges >= 2) {
									lgraphlogkilledge(*ex);
									remove_edge(ex, m_lgraph);
									work = true;
									if (lgraphrestartedgescan) {
										done = false;
										break;
									}
								}
							}
							if (!ee.is_match(awymatch))
								continue;
							if (solutiononly && !ee.is_solution())
								continue;
							if (renedges < 1 && awyrename != airwaymatchnone) {
								ee.set_ident(awyrename);
								lgraphlogrenameedge(*ex, awyrename);
								++renedges;
								continue;
							}
							lgraphlogkilledge(*ex);
							remove_edge(ex, m_lgraph);
							work = true;
							if (lgraphrestartedgescan) {
								done = false;
								break;
							}
						}
						if (done)
							break;
					}
					if (!bidir) {
						if (ue == ve)
							break;
						ve = ue;
						continue;
					}
					for (;;) {
						bool done(true);
						unsigned int renedges(0);
						LGraph::out_edge_iterator e0, e1;
						for (tie(e0, e1) = out_edges(ve, m_lgraph); e0 != e1; ) {
							LGraph::out_edge_iterator ex(e0);
							++e0;
							if (target(*ex, m_lgraph) != ue)
								continue;
							LEdge& ee(m_lgraph[*ex]);
							if (ee.is_match(awyrename)) {
								++renedges;
								if (renedges >= 2) {
									lgraphlogkilledge(*ex);
									remove_edge(ex, m_lgraph);
									work = true;
									if (lgraphrestartedgescan) {
										done = false;
										break;
									}
									continue;
								}
							}
							if (!ee.is_match(awymatch))
								continue;
							if (solutiononly && !ee.is_solution())
								continue;
							if (renedges < 1 && awyrename != airwaymatchnone) {
								ee.set_ident(awyrename);
								lgraphlogrenameedge(*ex, awyrename);
								++renedges;
								continue;
							}
							lgraphlogkilledge(*ex);
							remove_edge(ex, m_lgraph);
							work = true;
							if (lgraphrestartedgescan) {
								done = false;
								break;
							}
						}
						if (done)
							break;
					}
					if (ue == ve)
						break;
					ve = ue;
				}
			}
		}
	}
	return work;
}

bool CFMUAutoroute::lgraphchangeoutgoing(const Glib::ustring& id, lgraphairwayindex_t awyname)
{
	bool work(false);
	Glib::ustring identu(id.uppercase());
	for (solutionvertices_t::const_iterator svi(m_solutionvertices.begin()), sve(m_solutionvertices.end()); svi != sve; ++svi) {
		const LVertex& vv(m_lgraph[*svi]);
		const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
		if (sv.get_ident() != identu)
			continue;
		for (LGraph::vertex_descriptor v0(sv.get_v0()), v1(sv.get_v1()); v0 <= v1; ++v0) {
			for (;;) {
				bool done(true);
				LGraph::out_edge_iterator e0, e1;
				for (tie(e0, e1) = out_edges(v0, m_lgraph); e0 != e1; ) {
					LGraph::out_edge_iterator ex(e0);
					++e0;
					LEdge& ee(m_lgraph[*ex]);
					if (!ee.is_match(awyname))
						continue;
					work = true;
					if (awyname != airwaydct) {
						lgraphlogrenameedge(*ex, airwaydct);
						ee.set_ident(airwaydct);
						ee.set_metric(ee.get_metric() * get_dctpenalty());
						continue;
					}
					lgraphlogkilledge(*ex);
					remove_edge(ex, m_lgraph);
					if (lgraphrestartedgescan) {
						done = false;
						break;
					}
				}
				if (done)
					break;
			}
		}
	}
	return work;
}

bool CFMUAutoroute::lgraphchangeincoming(const Glib::ustring& id, lgraphairwayindex_t awyname)
{
	bool work(false);
	Glib::ustring identu(id.uppercase());
	for (solutionvertices_t::const_iterator svi(m_solutionvertices.begin()), sve(m_solutionvertices.end()); svi != sve; ++svi) {
		const LVertex& vv(m_lgraph[*svi]);
		const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
		if (sv.get_ident() != identu)
			continue;
		for (LGraph::vertex_descriptor v0(sv.get_v0()), v1(sv.get_v1()); v0 <= v1; ++v0) {
			for (;;) {
				bool done(true);
				LGraph::in_edge_iterator e0, e1;
				for (tie(e0, e1) = in_edges(v0, m_lgraph); e0 != e1; ) {
					LGraph::in_edge_iterator ex(e0);
					++e0;
					LEdge& ee(m_lgraph[*ex]);
					if (!ee.is_match(awyname))
						continue;
					work = true;
					if (awyname != airwaydct) {
						lgraphlogrenameedge(*ex, airwaydct);
						ee.set_ident(airwaydct);
						ee.set_metric(ee.get_metric() * get_dctpenalty());
						continue;
					}
					lgraphlogkilledge(*ex);
					remove_edge(*ex, m_lgraph);
					if (lgraphrestartedgescan) {
						done = false;
						break;
					}
				}
				if (done)
					break;
			}
		}
	}
	return work;
}

bool CFMUAutoroute::lgraphedgetodct(lgraphairwayindex_t awy)
{
	if (awy == airwaydct)
		return false;
	bool work(false);
	{
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1; ++e0) {
			LEdge& ee(m_lgraph[*e0]);
			if (!ee.is_solution())
				continue;
			if (!ee.is_match(awy))
				continue;
			lgraphlogrenameedge(*e0, airwaydct);
			ee.set_ident(airwaydct);
			ee.set_metric(ee.get_metric() * get_dctpenalty());
			work = true;
		}
	}
	return work;
}

bool CFMUAutoroute::lgraphkilloutgoing(const Glib::ustring& id)
{
	Glib::ustring identu(id.uppercase());
	bool work(false);
	for (;;) {
		bool done(true);
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1; ) {
			LGraph::edge_iterator ex(e0);
			++e0;
			LEdge& ee(m_lgraph[*ex]);
			if (!ee.is_solution())
				continue;
			LGraph::vertex_descriptor v(source(*ex, m_lgraph));
			const LVertex& vv(m_lgraph[v]);
			const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
			if (sv.get_ident() != identu)
				continue;
			lgraphlogkilledge(*ex);
			remove_edge(*ex, m_lgraph);
			work = true;
			if (lgraphrestartedgescan) {
				done = false;
				break;
			}
		}
		if (done)
			break;
	}
	return work;
}

bool CFMUAutoroute::lgraphkillincoming(const Glib::ustring& id)
{
	Glib::ustring identu(id.uppercase());
	bool work(false);
	for (;;) {
		bool done(true);
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1; ) {
			LGraph::edge_iterator ex(e0);
			++e0;
			LEdge& ee(m_lgraph[*ex]);
			if (!ee.is_solution())
				continue;
			LGraph::vertex_descriptor v(target(*ex, m_lgraph));
			const LVertex& vv(m_lgraph[v]);
			const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
			if (sv.get_ident() != identu)
				continue;
			lgraphlogkilledge(*ex);
			remove_edge(*ex, m_lgraph);
			work = true;
			if (lgraphrestartedgescan) {
				done = false;
				break;
			}
		}
		if (done)
			break;
	}
	return work;
}

bool CFMUAutoroute::lgraphsolutiongroundclearance(void)
{
	m_signal_log(log_graphrule, "Ground Clearance");
	bool work(false);
	for (;;) {
		bool done(true);
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1;) {
			LGraph::edge_iterator ex(e0);
			++e0;
			LEdge& ee(m_lgraph[*ex]);
			if (!ee.is_solution())
				continue;
			LGraph::vertex_descriptor u(source(*ex, m_lgraph));
			const LVertex& uu(m_lgraph[u]);
			const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
			if (su.get_v0() == su.get_v1())
				continue;
			LGraph::vertex_descriptor v(target(*ex, m_lgraph));
			const LVertex& vv(m_lgraph[v]);
			const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
			if (sv.get_v0() == sv.get_v1())
				continue;
			if (uu.get_superindex() == vv.get_superindex())
				continue;
			int32_t minalt(0);
			{
				TopoDb30::Profile prof(m_topodb.get_profile(su.get_coord(), sv.get_coord(), 5));
				if (prof.empty()) {
					std::ostringstream oss;
					oss << "Error computing ground clearance: " << lgraphvertexname(source(*ex, m_lgraph))
					    << ' ' << lgraphindex2airway(ee.get_ident(), true) << ' '
					    << lgraphvertexname(target(*ex, m_lgraph));
					m_signal_log(log_normal, oss.str());
					continue;
				}
				int32_t alt(prof.get_maxelev() * Point::m_to_ft);
				minalt = alt + 1000;
				if (alt >= 5000)
					minalt += 1000;
				{
					std::ostringstream oss;
					oss << "Leg: " << lgraphvertexname(source(*ex, m_lgraph))
					    << ' ' << lgraphindex2airway(ee.get_ident(), true) << ' '
					    << lgraphvertexname(target(*ex, m_lgraph))
					    << " ground elevation " << alt << "ft minimum level FL" << ((minalt + 99) / 100);
					m_signal_log(log_normal, oss.str());
				}
			}
			for (unsigned int i = 0; i < m_performance.size(); ++i) {
				const Performance& perf(m_performance[i]);
				if (perf.get_altitude() >= minalt)
					break;
				if (su.get_v0() + i > su.get_v1())
					break;
				if (sv.get_v0() + i > sv.get_v1())
					break;
				LGraph::edge_descriptor e;
                                bool ok;
				tie(e, ok) = boost::edge(su.get_v0() + i, sv.get_v0() + i, m_lgraph);
				if (ok) {
					LEdge& ee(m_lgraph[e]);
					if (ee.is_solution())
						work = true;
					done = false;
					lgraphlogkilledge(e);
					remove_edge(e, m_lgraph);
				}
				tie(e, ok) = boost::edge(sv.get_v0() + i, su.get_v0() + i, m_lgraph);
				if (ok) {
					LEdge& ee(m_lgraph[e]);
					if (ee.is_solution())
						work = true;
					done = false;
					lgraphlogkilledge(e);
					remove_edge(e, m_lgraph);
				}
			}
			if (!done)
				break;
		}
		if (done)
			break;
	}
	return work;
}

void CFMUAutoroute::parse_response(void)
{
	m_connchildtimeout.disconnect();
	if (!m_running) {
		m_signal_log(log_debug0, "Interrupted by user, stopping...");
		return;
	}
	m_signal_statuschange(statusmask_newvalidateresponse);
	bool work(m_validationresponse.size() == 1 && m_validationresponse[0] == "NO ERRORS");
	if (!work) {
		work = true;
		for (validationresponse_t::const_iterator i(m_validationresponse.begin()), e(m_validationresponse.end()); i != e; ++i) {
			if (i->empty())
				continue;
			const char * const *ignrx(ignoreregex);
			work = false;
			while (*ignrx) {
				Glib::RefPtr<Glib::Regex> rx(Glib::Regex::create(*ignrx++));
				Glib::MatchInfo minfo;
				if (rx->match(*i, minfo)) {
					work = true;
					break;
				}
			}
			if (!work)
				break;
		}
	}
	if (work) {
		if (m_pathprobe < 0) {
			if (lgraphsolutiongroundclearance()) {
				lgraphroute();
				return;
			}
			m_done = true;
			m_signal_log(log_fplproposal, get_simplefpl());
			stop();
			return;
		}
		for (;;) {
			++m_pathprobe;
			if (m_pathprobe + 1 >= m_route.get_nrwpt())
				break;
			const FPlanWaypoint& wpt(m_route[m_pathprobe]);
			if (wpt.get_flags() & FPlanWaypoint::altflag_ifr &&
			    wpt.get_pathcode() != FPlanWaypoint::pathcode_sid &&
			    wpt.get_pathcode() != FPlanWaypoint::pathcode_star)
				break;
		}
		if (m_pathprobe + 1 >= m_route.get_nrwpt()) {
			m_pathprobe = -1;
			m_signal_log(log_fplproposal, get_simplefpl());
			m_signal_log(log_normal, "No progress made; stopping...");
			stop();
			return;
		}
		lgraphroute();
		return;
	}
	work = false;
	for (validationresponse_t::const_iterator i(m_validationresponse.begin()), e(m_validationresponse.end()); i != e; ++i) {
		if (i->empty())
			continue;
		m_signal_log(log_graphrule, *i);
		if (false)
			std::cerr << "Message: " << *i << std::endl;
		const struct parsefunctions *pf(parsefunctions);
		for (; pf->regex; ++pf) {
			Glib::RefPtr<Glib::Regex> rx(Glib::Regex::create(pf->regex));
			Glib::MatchInfo minfo;
			if (!rx->match(*i, minfo))
				continue;
			if (pf->func)
				work = (this->*(pf->func))(minfo) || work;
			break;
		}
		if (!pf->regex) {
			const char * const *ignrx(ignoreregex);
			while (*ignrx) {
				Glib::RefPtr<Glib::Regex> rx(Glib::Regex::create(*ignrx++));
				Glib::MatchInfo minfo;
				if (!rx->match(*i, minfo))
					continue;
				break;
			}
			if (!*ignrx)
				m_signal_log(log_normal, "Error Message not understood: \"" + *i + "\"");
		}
	}
	if (!work && m_pathprobe >= 0 && m_pathprobe + 1 < m_route.get_nrwpt()) {
		Glib::ustring id0(m_route[m_pathprobe].get_icao());
		if (id0.empty())
			id0 = m_route[m_pathprobe].get_name();
		Glib::ustring id1(m_route[m_pathprobe + 1].get_icao());
		if (id1.empty())
			id1 = m_route[m_pathprobe + 1].get_name();
		work = lgraphmodifyawysegment(id0, id1, airwaymatchall, airwaymatchnone, 0, 600, false, true);
	}
	// disable path probing
	if (work || true) {
		if (m_pathprobe >= 0)
			m_signal_log(log_normal, "Resuming normal operation...");
		m_pathprobe = -1;
		lgraphroute();
		return;
	}
	for (;;) {
		++m_pathprobe;
		if (m_pathprobe + 1 >= m_route.get_nrwpt())
			break;
		const FPlanWaypoint& wpt(m_route[m_pathprobe]);
		if (wpt.get_flags() & FPlanWaypoint::altflag_ifr &&
		    wpt.get_pathcode() != FPlanWaypoint::pathcode_sid &&
		    wpt.get_pathcode() != FPlanWaypoint::pathcode_star)
			break;
	}
	if (m_pathprobe + 1 >= m_route.get_nrwpt()) {
		m_pathprobe = -1;
		m_signal_log(log_fplproposal, get_simplefpl());
		m_signal_log(log_normal, "No progress made; stopping...");
		stop();
		return;
	}
	if (m_pathprobe == 0)
		m_signal_log(log_normal, "No progress made; starting path probing...");
	lgraphroute();
}

const struct CFMUAutoroute::parsefunctions CFMUAutoroute::parsefunctions[] = {
	{ "^ROUTE49: THE POINT ([[:alnum:]]+) IS UNKNOWN IN THE CONTEXT OF THE ROUTE", &CFMUAutoroute::parse_response_route49 },
	{ "^ROUTE52: THE DCT SEGMENT ([[:alnum:]]+)..([[:alnum:]]+) IS FORBIDDEN", &CFMUAutoroute::parse_response_route52 },
	{ "^ROUTE130: UNKNOWN DESIGNATOR ([[:alnum:]]+)", &CFMUAutoroute::parse_response_route130 },
	{ "^ROUTE134: THE STAR LIMIT IS EXCEEDED FOR AERODROME .*? CONNECTING TO ([[:alnum:]]+).", &CFMUAutoroute::parse_response_route134 },
	{ "^ROUTE135: THE SID LIMIT IS EXCEEDED FOR AERODROME .*? CONNECTING TO ([[:alnum:]]+).", &CFMUAutoroute::parse_response_route135 },
	{ "^ROUTE139: ([[:alnum:]]+) IS PRECEDED BY ([[:alnum:]]+) WHICH IS NOT ONE OF ITS POINTS", &CFMUAutoroute::parse_response_route139 },
	{ "^ROUTE140: ([[:alnum:]]+) IS FOLLOWED BY ([[:alnum:]]+) WHICH IS NOT ONE OF ITS POINTS", &CFMUAutoroute::parse_response_route140 },
	{ "^ROUTE165: THE DCT SEGMENT ([[:alnum:]]+)..([[:alnum:]]+)", &CFMUAutoroute::parse_response_route165 },
	{ "^ROUTE168: INVALID DCT ([[:alnum:]]+)..([[:alnum:]]+).", &CFMUAutoroute::parse_response_route168 },
	{ "^ROUTE172: MULTIPLE ROUTES BETWEEN ([[:alnum:]]+) AND ([[:alnum:]]+). ([[:alnum:]]+) IS SUGGESTED.", &CFMUAutoroute::parse_response_route172 },
	{ "^PROF50: CLIMBING/DESCENDING OUTSIDE THE VERTICAL LIMITS OF SEGMENT ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+)", &CFMUAutoroute::parse_response_prof50 },
	{ "^PROF194: ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+) IS NOT AVAILABLE IN FL RANGE F([[:digit:]]+)..F([[:digit:]]+)", &CFMUAutoroute::parse_response_prof194 },
	{ "^PROF195: ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+) DOES NOT EXIST IN FL RANGE F([[:digit:]]+)..F([[:digit:]]+)", &CFMUAutoroute::parse_response_prof195 },
	{ "^PROF197: RS: ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+):F([[:digit:]]+)..F([[:digit:]]+) IS CLOSED FOR CRUISING", &CFMUAutoroute::parse_response_prof197 },
	{ "^PROF198: ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+) IS A CDR 3 IN FL RANGE F([[:digit:]]+)..F([[:digit:]]+)", &CFMUAutoroute::parse_response_prof198 },
	{ "^PROF199: ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+) IS A CLOSED CDR 2 IN FL RANGE F([[:digit:]]+)..F([[:digit:]]+)", &CFMUAutoroute::parse_response_prof199 },
	{ "^PROF201: CANNOT CLIMB OR DESCEND ON ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+) IN FL RANGE CLOSED", &CFMUAutoroute::parse_response_prof201 },
	{ "^PROF201: CANNOT CLIMB OR DESCEND ON ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+) IN FL RANGE F([[:digit:]]+)..F([[:digit:]]+)", &CFMUAutoroute::parse_response_prof201b },
	{ "^PROF204: RS: TRAFFIC VIA ((?:[[:alnum:]]+)(?: [[:alnum:]]+)*?) IS ON FORBIDDEN ROUTE", &CFMUAutoroute::parse_response_prof204 },
	{ "^PROF204: RS: TRAFFIC VIA ((?:[[:alnum:]]+)(?: [[:alnum:]]+)*?):F([[:digit:]]+)..F([[:digit:]]+) IS ON FORBIDDEN ROUTE", &CFMUAutoroute::parse_response_prof204b },
	{ "^PROF204: RS: TRAFFIC VIA ([[:alnum:]]+) ([[:alnum:]]+):F([[:digit:]]+)..F([[:digit:]]+) IS ON FORBIDDEN ROUTE", &CFMUAutoroute::parse_response_prof204c },
	{ "^PROF204: RS: TRAFFIC VIA ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+):F([[:digit:]]+)..F([[:digit:]]+) IS ON FORBIDDEN ROUTE", &CFMUAutoroute::parse_response_prof204d },
	{ "^PROF205: RS: TRAFFIC VIA ([[:alnum:]]+) IS OFF MANDATORY ROUTE", &CFMUAutoroute::parse_response_prof205 },
	{ "^PROF205: RS: TRAFFIC VIA ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+) IS OFF MANDATORY ROUTE", &CFMUAutoroute::parse_response_prof205b },
	{ "^PROF205: RS: TRAFFIC VIA ([[:alnum:]]+) ([[:alnum:]]+) ([[:alnum:]]+):F([[:digit:]]+)..F([[:digit:]]+) IS OFF MANDATORY ROUTE", &CFMUAutoroute::parse_response_prof205c },
	{ "^PROF206: THE DCT SEGMENT ([[:alnum:]]+) .. ([[:alnum:]]+) IS NOT AVAILABLE IN FL RANGE F([[:digit:]]+)..F([[:digit:]]+)", &CFMUAutoroute::parse_response_prof206 },
	{ 0, 0 }
};

const char *CFMUAutoroute::ignoreregex[] = {
	// ignore 8.33 carriage errors
	"^PROF188:",
	"^PROF189:",
	"^PROF190:",
	0
};

bool CFMUAutoroute::parse_response_route49(Glib::MatchInfo& minfo)
{
	return lgraphdisconnectvertex(minfo.fetch(1).uppercase());
}

bool CFMUAutoroute::parse_response_route52(Glib::MatchInfo& minfo)
{
	return lgraphmodifyedge(minfo.fetch(1), minfo.fetch(2), airwaydct, airwaymatchnone, 0, 600, true);
}

bool CFMUAutoroute::parse_response_route130(Glib::MatchInfo& minfo)
{
	Glib::ustring idu(minfo.fetch(1).uppercase());
	bool work(lgraphdisconnectvertex(idu));
	lgraphairwayindex_t awy(lgraphairway2index(idu));
	if (awy != airwaydct && awy != airwaymatchnone)
		work = lgraphedgetodct(awy) || work;
	return work;
}

bool CFMUAutoroute::parse_response_route134(Glib::MatchInfo& minfo)
{
	bool work(false);
	for (;;) {
		bool done(true);
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1; ) {
			LGraph::edge_iterator ex(e0);
			++e0;
			LEdge& ee(m_lgraph[*ex]);
			if (!ee.is_solution())
				continue;
			if (!ee.is_match(airwaystar))
				continue;
			lgraphlogkilledge(*ex);
			remove_edge(*ex, m_lgraph);
			work = true;
			if (lgraphrestartedgescan) {
				done = false;
				break;
			}
		}
		if (done)
			break;
	}
	return work;
}

bool CFMUAutoroute::parse_response_route135(Glib::MatchInfo& minfo)
{
	bool work(false);
	for (;;) {
		bool done(true);
		LGraph::edge_iterator e0, e1;
		for (tie(e0, e1) = edges(m_lgraph); e0 != e1; ) {
			LGraph::edge_iterator ex(e0);
			++e0;
			LEdge& ee(m_lgraph[*ex]);
			if (!ee.is_solution())
				continue;
			if (!ee.is_match(airwaysid))
				continue;
			lgraphlogkilledge(*ex);
			remove_edge(*ex, m_lgraph);
			work = true;
			if (lgraphrestartedgescan) {
				done = false;
				break;
			}
		}
		if (done)
			break;
	}
	return work;
}

bool CFMUAutoroute::parse_response_route139(Glib::MatchInfo& minfo)
{
	lgraphairwayindex_t awy(lgraphairway2index(minfo.fetch(1)));
	if (awy == airwaymatchnone || awy == airwaydct)
		return false;
	return lgraphchangeoutgoing(minfo.fetch(2), awy);
}

bool CFMUAutoroute::parse_response_route140(Glib::MatchInfo& minfo)
{
	lgraphairwayindex_t awy(lgraphairway2index(minfo.fetch(1)));
	if (awy == airwaymatchnone || awy == airwaydct)
		return false;
	return lgraphchangeincoming(minfo.fetch(2), awy);
}

bool CFMUAutoroute::parse_response_route165(Glib::MatchInfo& minfo)
{
	return lgraphmodifyedge(minfo.fetch(1), minfo.fetch(2), airwaydct, airwaymatchnone, 0, 600, true);
}

bool CFMUAutoroute::parse_response_route168(Glib::MatchInfo& minfo)
{
	return lgraphmodifyedge(minfo.fetch(1), minfo.fetch(2), airwaydct, airwaymatchnone, 0, 600, true);
}

bool CFMUAutoroute::parse_response_route172(Glib::MatchInfo& minfo)
{
	lgraphairwayindex_t awy(lgraphairway2index(minfo.fetch(3), true));
	if (awy == airwaydct)
		return false;
	return lgraphmodifyedge(minfo.fetch(1), minfo.fetch(2), airwaydct, awy, 0, 600, false);
}

bool CFMUAutoroute::parse_response_prof50(Glib::MatchInfo& minfo)
{
	lgraphairwayindex_t awy(lgraphairway2index(minfo.fetch(2)));
	if (awy != airwaymatchnone && lgraphmodifyawysegment(minfo.fetch(1), minfo.fetch(3), awy, airwaymatchnone, 0, 600, true, true))
		return true;
	return lgraphmodifyawysegment(minfo.fetch(1), minfo.fetch(3), airwaydct, airwaymatchnone, 0, 600, true, true);
}

bool CFMUAutoroute::parse_response_prof194(Glib::MatchInfo& minfo)
{
	return parse_response_prof195(minfo);
}

bool CFMUAutoroute::parse_response_prof195(Glib::MatchInfo& minfo)
{
	int altfrom(strtol(minfo.fetch(4).c_str(), 0, 10));
	int altto(strtol(minfo.fetch(5).c_str(), 0, 10));
	if (altfrom > altto)
		return false;
	lgraphairwayindex_t awy(lgraphairway2index(minfo.fetch(2)));
	if (awy != airwaymatchnone && lgraphmodifyawysegment(minfo.fetch(1), minfo.fetch(3), awy, airwaymatchnone, altfrom, altto, true, false))
		return true;
	return lgraphmodifyawysegment(minfo.fetch(1), minfo.fetch(3), airwaydct, airwaymatchnone, altfrom, altto, true, false);
}

bool CFMUAutoroute::parse_response_prof197(Glib::MatchInfo& minfo)
{
	int altfrom(strtol(minfo.fetch(4).c_str(), 0, 10));
	int altto(strtol(minfo.fetch(5).c_str(), 0, 10));
	if (altfrom > altto)
		return false;
	lgraphairwayindex_t awy(lgraphairway2index(minfo.fetch(2)));
	if (awy != airwaymatchnone && lgraphmodifyawysegment(minfo.fetch(1), minfo.fetch(3), awy, airwaymatchnone, altfrom, altto, true, true))
		return true;
	return lgraphmodifyawysegment(minfo.fetch(1), minfo.fetch(3), airwaydct, airwaymatchnone, altfrom, altto, true, true);
}

bool CFMUAutoroute::parse_response_prof198(Glib::MatchInfo& minfo)
{
	return parse_response_prof195(minfo);
}

bool CFMUAutoroute::parse_response_prof199(Glib::MatchInfo& minfo)
{
	return parse_response_prof195(minfo);
}

bool CFMUAutoroute::parse_response_prof201(Glib::MatchInfo& minfo)
{
	return parse_response_prof50(minfo);
}

bool CFMUAutoroute::parse_response_prof201b(Glib::MatchInfo& minfo)
{
	int altfrom(strtol(minfo.fetch(4).c_str(), 0, 10));
	int altto(strtol(minfo.fetch(5).c_str(), 0, 10));
	if (altfrom > altto)
		return false;
	lgraphairwayindex_t awy(lgraphairway2index(minfo.fetch(2)));
	if (awy != airwaymatchnone && lgraphmodifyawysegment(minfo.fetch(1), minfo.fetch(3), awy, airwaymatchnone, altfrom, altto, true, true))
		return true;
	return lgraphmodifyawysegment(minfo.fetch(1), minfo.fetch(3), airwaydct, airwaymatchnone, altfrom, altto, true, true);
}

bool CFMUAutoroute::parse_response_prof204(Glib::MatchInfo& minfo)
{
	bool work(false);
	if (false)
		std::cerr << "PROF204: \"" << minfo.fetch(1) << "\" \"" << minfo.fetch(0) << "\"" << std::endl;
	typedef std::vector<Glib::ustring> token_t;
	token_t token(Glib::Regex::split_simple(" +", minfo.fetch(1)));
 	for (int i = 0; i < token.size(); ++i) {
 	 	if (i + 2 < token.size() &&
		    !Engine::AirwayGraphResult::Vertex::is_ident_numeric(token[i]) &&
		    Engine::AirwayGraphResult::Vertex::is_ident_numeric(token[i + 1]) &&
		    !Engine::AirwayGraphResult::Vertex::is_ident_numeric(token[i + 2])) {
			if (true)
				std::cerr << "PROF204: kill " << token[i] << "--" << token[i + 1] << "->" << token[i + 2] << std::endl;
			lgraphairwayindex_t awy(lgraphairway2index(token[i + 1]));
			if (awy != airwaymatchnone && awy != airwaydct) {
				bool work2(lgraphmodifyawysegment(token[i], token[i + 2], awy, airwaymatchnone, 0, 600, true, true));
				if (!work2) {
					if (true)
						std::cerr << "PROF204: " << token[i] << "--" << lgraphindex2airway(awy) << "->" << token[i + 2]
							  << " did not match, trying DCT" << std::endl;
					work2 = lgraphmodifyawysegment(token[i], token[i + 2], airwaydct, airwaymatchnone, 0, 600, true, true);
					if (!work2) {
						if (true)
							std::cerr << "PROF204: " << token[i] << "--->" << token[i + 2] << " did not match, trying airway convert to DCT" << std::endl;
						work2 = lgraphedgetodct(awy);
						if (!work2) {
							std::cerr << "PROF204: could not convert " << lgraphindex2airway(awy)
								  << " to DCT, removing incoming/outgoing edges" << std::endl;
							work2 = lgraphkilloutgoing(token[i]);
							work2 = lgraphkillincoming(token[i + 2]) || work2;
						}
					}
				}
				work = work2 || work;
			}
			i += 2;
			continue;
		}
		if (true)
			std::cerr << "PROF204: kill " << token[i] << std::endl;
		work = lgraphdisconnectsolutionvertex(token[i]) || work;
		if (!work && token[i].size() >= 4) {
			opendb();
			AirspacesDb::elementvector_t ev(m_airspacedb.find_by_icao(token[i], 0, /* AirspacesDb::comp_startswith */ AirspacesDb::comp_exact,
										  0, AirspacesDb::Airspace::subtables_all));
			for (AirspacesDb::elementvector_t::const_iterator ei(ev.begin()), ee(ev.end()); ei != ee; ++ei) {
				const AirspacesDb::Airspace& aspc(*ei);
				if (!aspc.is_valid() || aspc.get_typecode() != AirspacesDb::Airspace::typecode_ead)
					continue;
				{
					std::ostringstream oss;
					oss << "A " << aspc.get_icao_name();
					m_signal_log(log_graphchange, oss.str());
				}
				const MultiPolygonHole& poly(aspc.get_polygon());
				Rect bbox(poly.get_bbox());
				for (;;) {
					bool done(true);
					LGraph::edge_iterator e0, e1;
					for (tie(e0, e1) = edges(m_lgraph); e0 != e1; ) {
						LGraph::edge_iterator ex(e0);
						++e0;
						LEdge& ee(m_lgraph[*ex]);
						if (!ee.is_solution())
							continue;
						LGraph::vertex_descriptor u(source(*ex, m_lgraph));
						const LVertex& uu(m_lgraph[u]);
						const LSuperVertex& su(m_lsupervertices[uu.get_superindex()]);
						LGraph::vertex_descriptor v(target(*ex, m_lgraph));
						const LVertex& vv(m_lgraph[v]);
						const LSuperVertex& sv(m_lsupervertices[vv.get_superindex()]);
						bool insideu(bbox.is_inside(su.get_coord()) && poly.windingnumber(su.get_coord()));
						bool insidev(bbox.is_inside(sv.get_coord()) && poly.windingnumber(sv.get_coord()));
						bool deledge(insideu || insidev || poly.is_intersection(su.get_coord(), sv.get_coord()));
						if (!deledge)
							continue;
						lgraphlogkilledge(*ex);
						remove_edge(*ex, m_lgraph);
						work = true;
						if (lgraphrestartedgescan) {
							done = false;
							break;
						}
					}
					if (done)
						break;
				}
			}
		}
	}
	return work;
}

bool CFMUAutoroute::parse_response_prof204b(Glib::MatchInfo& minfo)
{
	int altfrom(strtol(minfo.fetch(2).c_str(), 0, 10));
	int altto(strtol(minfo.fetch(3).c_str(), 0, 10));
	if (altfrom > altto)
		return false;
	bool work(false);
	if (false)
		std::cerr << "PROF204: \"" << minfo.fetch(1) << "\" \"" << minfo.fetch(0) << "\"" << std::endl;
	typedef std::vector<Glib::ustring> token_t;
	token_t token(Glib::Regex::split_simple(" +", minfo.fetch(1)));
 	for (int i = 0; i < token.size(); ++i) {
 	 	if (i + 2 < token.size() &&
		    !Engine::AirwayGraphResult::Vertex::is_ident_numeric(token[i]) &&
		    Engine::AirwayGraphResult::Vertex::is_ident_numeric(token[i + 1]) &&
		    !Engine::AirwayGraphResult::Vertex::is_ident_numeric(token[i + 2])) {
			if (true)
				std::cerr << "PROF204: kill " << token[i] << "--" << token[i + 1] << "->" << token[i + 2] << std::endl;
			lgraphairwayindex_t awy(lgraphairway2index(token[i + 1]));
			if (awy != airwaymatchnone && awy != airwaydct) {
				bool work2(lgraphmodifyawysegment(token[i], token[i + 2], awy, airwaymatchnone, altfrom, altto, true, true));
				if (!work2) {
					if (true)
						std::cerr << "PROF204: " << token[i] << "--" << lgraphindex2airway(awy) << "->" << token[i + 2]
							  << " did not match, trying DCT" << std::endl;
					work2 = lgraphmodifyawysegment(token[i], token[i + 2], airwaydct, airwaymatchnone, altfrom, altto, true, true);
					if (!work2) {
						if (true)
							std::cerr << "PROF204: " << token[i] << "--->" << token[i + 2] << " did not match, trying airway convert to DCT" << std::endl;
						work2 = lgraphedgetodct(awy);
						if (!work2) {
							std::cerr << "PROF204: could not convert " << lgraphindex2airway(awy)
								  << " to DCT, removing incoming/outgoing edges" << std::endl;
							work2 = lgraphkilloutgoing(token[i]);
							work2 = lgraphkillincoming(token[i + 2]) || work2;
						}
					}
				}
				work = work2 || work;
			}
			i += 2;
			continue;
		}
		if (true)
			std::cerr << "PROF204: kill " << token[i] << std::endl;
		work = lgraphdisconnectvertex(token[i], altfrom, altto) || work;
	}
	return work;
}

bool CFMUAutoroute::parse_response_prof204c(Glib::MatchInfo& minfo)
{
	int altfrom(strtol(minfo.fetch(3).c_str(), 0, 10));
	int altto(strtol(minfo.fetch(4).c_str(), 0, 10));
	if (altfrom > altto)
		return false;
	return lgraphmodifyedge(minfo.fetch(1), minfo.fetch(2), airwaymatchdctawy, airwaymatchnone, altfrom, altto, true);
}

bool CFMUAutoroute::parse_response_prof204d(Glib::MatchInfo& minfo)
{
	int altfrom(strtol(minfo.fetch(4).c_str(), 0, 10));
	int altto(strtol(minfo.fetch(5).c_str(), 0, 10));
	if (altfrom > altto)
		return false;
	lgraphairwayindex_t awy(lgraphairway2index(minfo.fetch(2)));
	if (awy != airwaymatchnone && lgraphmodifyawysegment(minfo.fetch(1), minfo.fetch(3), awy, airwaymatchnone, altfrom, altto, true, true))
		return true;
	return lgraphmodifyawysegment(minfo.fetch(1), minfo.fetch(3), airwaydct, airwaymatchnone, altfrom, altto, true, true);
}

bool CFMUAutoroute::parse_response_prof205(Glib::MatchInfo& minfo)
{
	return lgraphdisconnectsolutionvertex(minfo.fetch(1));
}

bool CFMUAutoroute::parse_response_prof205b(Glib::MatchInfo& minfo)
{
	return parse_response_prof50(minfo);
}

bool CFMUAutoroute::parse_response_prof205c(Glib::MatchInfo& minfo)
{
	return parse_response_prof204d(minfo);
}

bool CFMUAutoroute::parse_response_prof206(Glib::MatchInfo& minfo)
{
	int altfrom(strtol(minfo.fetch(3).c_str(), 0, 10));
	int altto(strtol(minfo.fetch(4).c_str(), 0, 10));
	if (altfrom > altto)
		return false;
	return lgraphmodifyedge(minfo.fetch(1), minfo.fetch(2), airwaydct, airwaymatchnone, altfrom, altto, true);
}
//PROF201: CANNOT CLIMB OR DESCEND ON SOPER N851 PIXOS IN FL RANGE CLOSED, BECAUSE OF UNAVAILABLE LEVELS ON N851
//PROF50: CLIMBING/DESCENDING OUTSIDE THE VERTICAL LIMITS OF SEGMENT NUNRI T103 KUDIS
//PROF205: RS: TRAFFIC VIA EGHL+4 IS OFF MANDATORY ROUTE REF:[EG2380D] GWC COMP FOR TRAFFIC 1.DEP FARNBOROUGH G
//PROF205: RS: TRAFFIC VIA EGHH EGHI EGHO EGHL+4 IS OFF MANDATORY ROUTE REF:[EGOG5112A] GWC DCT SFD APP.4
//PROF204: RS: TRAFFIC VIA EDGGADF IS ON FORBIDDEN ROUTE REF:[ED3115A] EDGGADF

// Airspace BG has 49 navaids, 108 intersections
// Airspace BI has 82 navaids, 77 intersections
// Airspace CFMU1 CFMU1 has 3893 navaids, 14369 intersections
// Airspace CY-AERODB has 1294 navaids, 14666 intersections
// Airspace CY has 987 navaids, 10145 intersections
// Airspace DA has 89 navaids, 128 intersections
// Airspace DT has 25 navaids, 154 intersections
// Airspace EB MIL EB MIL DOWNLOAD has 114 navaids, 322 intersections
// Airspace EB-EAD-1 has 85 navaids, 194 intersections
// Airspace EB-EAD-2 has 0 navaids, 3 intersections
// Airspace EB-EAD-3 has 1 navaids, 3 intersections
// Airspace EB-EAD-5 has 3 navaids, 18 intersections
// Airspace EBDWNLOAD1 EBDOWNLOAD1 has 681 navaids, 3320 intersections
// Airspace EBDWNLOAD2 EBDOWNLOAD2 has 231 navaids, 1035 intersections
// Airspace ED has 355 navaids, 3156 intersections
// Airspace EE has 20 navaids, 94 intersections
// Airspace EF has 104 navaids, 385 intersections
// Airspace EG2 has 56 navaids, 143 intersections
// Airspace EG has 362 navaids, 1126 intersections
// Airspace EH-EAD-1 EH-EAD-1 has 101 navaids, 605 intersections
// Airspace EH-EAD-2 EH-EAD-2 has 0 navaids, 33 intersections
// Airspace EI-EAD-1 EI-EAD-1 has 60 navaids, 217 intersections
// Airspace EI-EAD-2 EI-EAD-2 has 0 navaids, 6 intersections
// Airspace EI-EAD-3 EI-EAD-3 has 0 navaids, 1 intersections
// Airspace EI-EAD-4 EI-EAD-4 has 0 navaids, 22 intersections
// Airspace EI-EAD-5 EI-EAD-5 has 0 navaids, 19 intersections
// Airspace EK-EAD-1 EK-EAD-1 has 74 navaids, 403 intersections
// Airspace EK-EAD-2 EK-EAD-2 has 0 navaids, 51 intersections
// Airspace EK-EAD-3 EK-EAD-3 has 5 navaids, 26 intersections
// Airspace EK-EAD-4 EK-EAD-4 has 6 navaids, 20 intersections
// Airspace EN-EAD-1 EN EAD FIR PLUS DELEGATED AREAS has 270 navaids, 1185 intersections
// Airspace EN-EAD-2 NO_FIR AREA NEAR NORWAY has 0 navaids, 10 intersections
// Airspace EN-EAD-3 JAN MAYEN ISLAND (JAN NDB LOCATION) has 1 navaids, 0 intersections
// Airspace EP has 146 navaids, 584 intersections
// Airspace ES has 241 navaids, 1028 intersections
// Airspace EV-EAD-1 EV-EAD-1 has 14 navaids, 137 intersections
// Airspace EV-EAD-3 EV-EAD-3 has 3 navaids, 17 intersections
// Airspace EV-EAD-4 EV-EAD-4 has 2 navaids, 9 intersections
// Airspace EV-EAD-5 EV-EAD-5 has 0 navaids, 1 intersections
// Airspace EY has 17 navaids, 159 intersections
// Airspace FA has 187 navaids, 256 intersections
// Airspace FABEC1N FABEC PROJECT PHASE 1 NORTH has 57 navaids, 531 intersections
// Airspace FABEC1W FABEC PROJECT PHASE 1 WEST has 102 navaids, 305 intersections
// Airspace FABEC2N FABEC PROJECT PHASE 2 NORTH has 159 navaids, 814 intersections
// Airspace FABECSOUTH FABEC PROJECT PHASE SOUTH has 184 navaids, 852 intersections
// Airspace FD_EAD has 4 navaids, 6 intersections
// Airspace FX_EAD has 4 navaids, 3 intersections
// Airspace FY_EAD has 4 navaids, 3 intersections
// Airspace G1AA AREA OF RESPONSIBILITY 1 has 5384 navaids, 18698 intersections
// Airspace G1BB AREA OF RESPONSIBILITY 2 has 453 navaids, 6887 intersections
// Airspace G1CC AREA OF RESPONSIBILITY 3 has 2333 navaids, 9117 intersections
// Airspace GM has 59 navaids, 158 intersections
// Airspace GO has 97 navaids, 334 intersections
// Airspace GV has 7 navaids, 39 intersections
// Airspace HE has 56 navaids, 93 intersections
// Airspace HL has 61 navaids, 123 intersections
// Airspace KCHART1 COVERAGE OF CHARTS U.S.A. has 969 navaids, 13909 intersections
// Airspace LA has 5 navaids, 30 intersections
// Airspace LB has 36 navaids, 118 intersections
// Airspace LC has 10 navaids, 67 intersections
// Airspace LD-EAD-1 EAD AIRSPACE CROATIA has 43 navaids, 131 intersections
// Airspace LD-EAD-2 EAD AIRSPACE DELEGATED TO CROATIA has 1 navaids, 4 intersections
// Airspace LD-EAD-3 AIRSPACE DELEGATION TO CROATIA has 0 navaids, 10 intersections
// Airspace LE_1 has 265 navaids, 826 intersections
// Airspace LE_2 has 50 navaids, 115 intersections
// Airspace LF has 482 navaids, 1549 intersections
// Airspace LG has 119 navaids, 225 intersections
// Airspace LH has 79 navaids, 252 intersections
// Airspace LI has 269 navaids, 766 intersections
// Airspace LICD LAMPEDUSA has 4 navaids, 5 intersections
// Airspace LJ has 17 navaids, 79 intersections
// Airspace LK has 97 navaids, 443 intersections
// Airspace LL has 28 navaids, 51 intersections
// Airspace LM has 14 navaids, 93 intersections
// Airspace LO has 50 navaids, 214 intersections
// Airspace LP has 65 navaids, 351 intersections
// Airspace LQ has 29 navaids, 65 intersections
// Airspace LR has 72 navaids, 360 intersections
// Airspace LS has 42 navaids, 204 intersections
// Airspace LT has 207 navaids, 403 intersections
// Airspace LU has 11 navaids, 49 intersections
// Airspace LW has 12 navaids, 29 intersections
// Airspace LY has 61 navaids, 172 intersections
// Airspace LZ-EAD-1 LZ-EAD-1 has 32 navaids, 138 intersections
// Airspace LZ-EAD-2 LZ-EAD-2 has 0 navaids, 0 intersections
// Airspace NZ-LNIA LOWER NORTH ISLAND ARC has 40 navaids, 434 intersections
// Airspace NZ-NOEN NORTH ENRC has 64 navaids, 885 intersections
// Airspace NZ-SOEN SOUTH ENRC has 74 navaids, 939 intersections
// Airspace NZ-UNIA UPPER NORTH ISLAND ARC has 33 navaids, 500 intersections
// Airspace NZ-USIA UPPER SOUTH ISLAND ARC has 36 navaids, 366 intersections
// Airspace NZ has 557 navaids, 4093 intersections
// Airspace OI has 180 navaids, 305 intersections
// Airspace OJ has 22 navaids, 59 intersections
// Airspace OL has 7 navaids, 30 intersections
// Airspace OS has 24 navaids, 40 intersections
// Airspace RC has 71 navaids, 97 intersections
// Airspace RP has 92 navaids, 295 intersections
// Airspace UB-EAD-1 AZERBAIJAN EAD AREA EAST has 30 navaids, 80 intersections
// Airspace UB-EAD-2 AZERBAIJAN EAD AREA WEST has 5 navaids, 3 intersections
// Airspace UD has 13 navaids, 36 intersections
// Airspace UG has 20 navaids, 35 intersections
// Airspace UK has 149 navaids, 401 intersections
// Airspace UM BELARUS has 32 navaids, 153 intersections
// Airspace UT-2 TAJIKISTAN has 9 navaids, 26 intersections
// Airspace U_ADR-1 has 187 navaids, 594 intersections
// Airspace U_ADR-2 has 2 navaids, 24 intersections
// Airspace WS SINGAPORE has 26 navaids, 164 intersections
// Airspace XX has 8 navaids, 3 intersections
